/*****************************************************************************/
/* A callback object for use in scheduling completion of asynchronous        */
/* requests.                                                                 */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

#ifndef USBBACK_CALLBACK_H
#define USBBACK_CALLBACK_H

#include <linux/spinlock.h>
#include "usbback_work.h"

/* Service parameter blocks contain callbacks for asynchronous completions.  */

struct usbback_callback
{
    struct usbback_work work;
    int                 error;
};

/* Client of service initialises callback with its callback function.        */

static inline void usbback_callback_init
(
    struct usbback_callback * callback,
    void ( * function )( struct usbback_callback * callback )
)
{
    usbback_work_init( &callback->work, (void(*)(void*))function, callback );

    callback->error = 0;
}

/* Client may use link whilst it owns parameter block. Service may use link  */
/* whilst it owns parameter block.  Link is reserved whilst callback is      */
/* scheduled for completion.                                                 */

static inline struct list_head * usbback_callback_to_link
  ( struct usbback_callback * callback )
{
    return usbback_work_to_link( &callback->work );
}

static inline struct usbback_callback * usbback_callback_link_to
  ( struct list_head * link )
{
    return container_of
      ( usbback_work_link_to( link ), struct usbback_callback, work );
}

/* Service which completes requests concurrently may call                    */
/* usbback_callback_complete or usbback_callback_success to complete the     */
/* callback.                                                                 */

static inline void usbback_callback_complete
  ( struct usbback_callback * callback, int error )
{
    callback->error = error;

    usbback_work_schedule( &callback->work );
}

static inline void usbback_callback_success
  ( struct usbback_callback * callback )
{
    usbback_callback_complete( callback, 0 );
}

/* These functions used by serialiser below. */

static inline void usbback_callback_set_error
  ( struct usbback_callback * callback, int error )
{
    callback->error = error;
}

static inline void usbback_callback_complete_synchronously
  ( struct usbback_callback * callback )
{
    usbback_work_perform_synchronously( &callback->work );
}

/* When callback completes, client may call usbback_callback_query_error to  */
/* get the error code.                                                       */

static inline int usbback_callback_query_error
  ( struct usbback_callback * callback )
{
    return callback->error;
}

/* Services which must serialise completions to preserve submission order    */
/* can use one of these.                                                     */

struct usbback_callback_serialiser
{
    spinlock_t          lock;
    struct list_head    list;
    struct usbback_work work;
    int                 running;
};

void usbback_callback_serialiser_function( void * context );

#define USBBACK_CALLBACK_SERIALISER_INIT( name )                  \
{                                                                 \
    SPIN_LOCK_UNLOCKED,                                           \
    LIST_HEAD_INIT( name.list ),                                  \
    USBBACK_WORK_INIT                                             \
      ( name.work, usbback_callback_serialiser_function, &name ), \
    0                                                             \
}

#define USBBACK_CALLBACK_SERIALISER( name ) \
struct usbback_callback_serialiser name =   \
  USBBACK_CALLBACK_SERIALISER_INIT( name )

/* The service completes the callback to the serialiser which serialises the */
/* completions to the client, performing them in submission order.           */

static inline void usbback_callback_serialiser_complete_callback
(
    struct usbback_callback_serialiser * serialiser,
    struct usbback_callback            * callback,
    int                                  error
)
{
    usbback_callback_set_error( callback, error );

    unsigned long flags;

    spin_lock_irqsave( &serialiser->lock, flags );

    list_add_tail( usbback_callback_to_link( callback ), &serialiser->list );

    spin_unlock_irqrestore( &serialiser->lock, flags );

    usbback_work_schedule( &serialiser->work );
}

#endif
