/*****************************************************************************/
/* A device object representing the low-level connection to a single         */
/* front-end.                                                                */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on                                                                  */
/*                                                                           */
/* arch/xen/drivers/usbif/backend/control.c                                  */
/* arch/xen/drivers/usbif/backend/interface.c                                */
/* arch/xen/drivers/usbif/backend/main.c                                     */
/*                                                                           */
/* original copyright notices follow...                                      */
/*****************************************************************************/

/******************************************************************************
 * arch/xen/drivers/usbif/backend/control.c
 * 
 * Routines for interfacing with the control plane.
 * 
 * Copyright (c) 2004, Keir Fraser
 */

/******************************************************************************
 * arch/xen/drivers/usbif/backend/interface.c
 * 
 * USB device interface management.
 * 
 * by Mark Williamson, Copyright (c) 2004
 */

/******************************************************************************
 * arch/xen/drivers/blkif/backend/interface.c
 * 
 * Block-device interface management.
 * 
 * Copyright (c) 2004, Keir Fraser
 */

/******************************************************************************
 * arch/xen/drivers/usbif/backend/main.c
 * 
 * Backend for the Xen virtual USB driver - provides an abstraction of a
 * USB host controller to the corresponding frontend driver.
 *
 * by Mark Williamson
 * Copyright (c) 2004 Intel Research Cambridge
 * Copyright (c) 2004, 2005 Mark Williamson
 *
 * Based on arch/xen/drivers/blkif/backend/main.c
 * Copyright (c) 2003-2004, Keir Fraser & Steve Hand
 */


#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <asm/pgalloc.h>
#include <asm-xen/evtchn.h>
#include <asm-xen/hypervisor.h>
#include <asm-xen/xen-public/io/ring.h>
#include "usbback_assert.h"
#include "usbback_bus.h"
#include "usbback_device.h"
#include "usbback_trace.h"
#include "usbback_work.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define VMALLOC_VMADDR(x) ((unsigned long)(x))
#endif

static DECLARE_RWSEM( usbback_device_list_sem );

typedef enum
{
    usbback_device_state_i,
    usbback_device_state_i_cr,
    usbback_device_state_i_sd,
    usbback_device_state_i_cr_sd,
    usbback_device_state_i_cr_rs,
    usbback_device_state_i_sd_sd,
    usbback_device_state_i_cr_rs_dt,
    usbback_device_state_i_cr_rs_cn,
    usbback_device_state_i_sd_sd_ud,
    usbback_device_state_i_cr_rs_dt_sd,
    usbback_device_state_i_cr_rs_dt_ud,
    usbback_device_state_i_cr_rs_cn_sd,
    usbback_device_state_i_cr_rs_cn_rs,
    usbback_device_state_i_cr_rs_dt_sd_ud,
    usbback_device_state_i_cr_rs_cn_sd_rs,
    usbback_device_state_i_cr_rs_cn_rs_dn
}
usbback_device_state;

typedef enum
{
    usbback_device_stimulus_cr, /* Create                */
    usbback_device_stimulus_dt, /* Destroy               */
    usbback_device_stimulus_cn, /* Connect               */
    usbback_device_stimulus_dn, /* Disconnect            */
    usbback_device_stimulus_cp, /* Claim port            */
    usbback_device_stimulus_rp, /* Release port          */
    usbback_device_stimulus_sd, /* Shutdown 1 or 2       */
    usbback_device_stimulus_rs, /* Response success sent */
    usbback_device_stimulus_rf, /* Response failure sent */
    usbback_device_stimulus_dd, /* Disconnect done       */
    usbback_device_stimulus_ud, /* Unregister done       */
    usbback_device_stimulus_invalid
}
usbback_device_stimulus;

struct usbback_device;

struct usbback_device_resource
{
    struct list_head                link;
    struct usbback_device         * device;
    struct usbback_device_request   request;
};

#define USBBACK_DEVICE_RESOURCE_COUNT 16

struct usbback_device
{
    struct device                  dev;
    struct list_head               link;
    u32                            handle;
    spinlock_t                     lock;
    usbback_device_state           state;
    int                            received_msg_free;
    ctrl_msg_t                     received_msg;
    struct usbback_work            connect_1_work;
    struct usbback_work            interrupt_1_work;
    struct usbback_work            disconnect_1_work;
    struct usbback_work            disconnect_2_work;
    struct usbback_work            respond_1_work;
    struct usbback_work            unregister_1_work;
    void                       ( * request_handler )
      ( struct usbback_device_request * request, void * context );
    void                       ( * claim_port      )
      ( u32 port, char * path, void * context );
    void                       ( * release_port    )
      ( u32 port, void * context );
    void                       ( * flush_requests )( void * context );
    void                         * driver_context;
    int                            driver_attached;
    usbif_back_ring_t              back_ring;
    unsigned int                   evtchn;
    int                            irq;
    domid_t                        remote_domain_id;
    int                            shutdown;
    struct list_head               free_resource_list;
    struct usbback_device_resource resources[ USBBACK_DEVICE_RESOURCE_COUNT ];
    atomic_t                       connect_reference_count;
    int                            interrupt_1_running;
};

struct usbback_device * usbback_device_dev_to( struct device * dev )
{
    return container_of( dev, struct usbback_device, dev );
}

static inline void usbback_device_get_connect_reference
  ( struct usbback_device * device )
{
  /*    trace(); */

    atomic_inc( &device->connect_reference_count );
}

static inline void usbback_device_put_connect_reference
  ( struct usbback_device * device )
{
  /*    trace(); */

    if( atomic_dec_and_test( &device->connect_reference_count ) )
    {
        int scheduled = usbback_work_schedule( &device->disconnect_2_work );

        ASSERT( scheduled );
    }
}

static void usbback_device_handle_stimulus
(
    struct usbback_device   * device,
    usbback_device_stimulus   stimulus,
    unsigned long           * flags
);

static void usbback_device_release( struct device * dev );

static void usbback_device_connect_1( void * data );

static void usbback_device_interrupt_1( void * data );

static void usbback_device_disconnect_1( void * data );

static void usbback_device_disconnect_2( void * data );

static void usbback_device_respond_1( void * data );

static void usbback_device_unregister_1( void * data );

static int usbback_device_init_or_exit
  ( struct usbback_device * device, u32 handle, int exit )
{
    trace();

    int return_value = 0;

    if( exit )
    {
        goto EXIT;
    }

    memset( device, 0, sizeof( *device ) );

    INIT_LIST_HEAD( &device->link );

    device->handle = handle;

    device->dev.bus    = &usbback_bus_type;
    device->dev.parent = &usbback_bus;

    snprintf
      ( device->dev.bus_id, BUS_ID_SIZE, "%08lx", (unsigned long)handle );

    device->dev.release = usbback_device_release;

    spin_lock_init( &device->lock );

    device->state = usbback_device_state_i;

    device->received_msg_free = 1;

    usbback_work_init
      ( &device->connect_1_work, usbback_device_connect_1, device );

    usbback_work_init
      ( &device->interrupt_1_work, usbback_device_interrupt_1, device );

    usbback_work_init
      ( &device->disconnect_1_work, usbback_device_disconnect_1, device );

    usbback_work_init
      ( &device->disconnect_2_work, usbback_device_disconnect_2, device );

    usbback_work_init
      ( &device->respond_1_work, usbback_device_respond_1, device );

    usbback_work_init
      ( &device->unregister_1_work, usbback_device_unregister_1, device );

    INIT_LIST_HEAD( &device->free_resource_list );

    {
        int i;

        for( i = 0; i < USBBACK_DEVICE_RESOURCE_COUNT; i++ )
        {
            struct usbback_device_resource * resource =
              &device->resources[ i ];

            resource->device = device;

            INIT_LIST_HEAD( &resource->link );

            list_add( &resource->link, &device->free_resource_list );
        }
    }

    atomic_set( &device->connect_reference_count, 0 );

    if( ( return_value = device_register( &device->dev ) ) != 0 )
    {
        goto EXIT_NO_DEVICE_REGISTER;
    }

    return 0;

  EXIT:

  EXIT_NO_DEVICE_REGISTER:

    return return_value;
}

static int usbback_device_init( struct usbback_device * device, u32 handle )
{
    trace();

    return usbback_device_init_or_exit( device, handle, 0 );
}

static void usbback_device_exit( struct usbback_device * device )
{
    trace();

    (void)usbback_device_init_or_exit( device, 0, 1 );
}

void usbback_device_get( struct usbback_device * device )
{
    trace();

    get_device( &device->dev );
}

void usbback_device_put( struct usbback_device * device )
{
    trace();

    put_device( &device->dev );
}

static void usbback_device_release( struct device * dev )
{
    trace();

    struct usbback_device * device = usbback_device_dev_to( dev );

    usbback_device_exit( device );

    kfree( device );
}

static int usbback_device_shutdown( struct device * dev, void * ignored );

struct usbback_device * usbback_device_find_or_allocate
  ( u32 handle, int shutting_down )
{
    trace();

    static LIST_HEAD( usbback_device_list );

    struct usbback_device * device = NULL;

    /* Try to reference an already created device. */

    down_write( &usbback_device_list_sem );

    {
        struct usbback_device * cur = NULL;

        list_for_each_entry( cur, &usbback_device_list, link )
        {
            if( handle == cur->handle )
            {
                device = cur;

                usbback_device_get( device );

                break;
            }
        }
    }

    if( device == NULL )
    {
        /* No existing device so try to create one. */

        device = (struct usbback_device *)kmalloc
          ( sizeof( struct usbback_device ), GFP_KERNEL );

        if( device != NULL )
        {
            /* Managed to allocate memory. */

            if( usbback_device_init( device, handle ) == 0 )
            {
                usbback_device_get( device );

                list_add_tail( &device->link, &usbback_device_list );

                if( shutting_down )
                {
                    usbback_device_shutdown( &device->dev, NULL );
                }
            }
            else
            {
                kfree( device );

                device = NULL;
            }
        }
    }

    up_write( &usbback_device_list_sem );

    return device;
}

struct message_info
{
    size_t                  length;
    usbback_device_stimulus stimulus;
};

static struct message_info message_subtype_info[] =
{
    [ CMSG_USBIF_BE_CREATE       ] =
      { sizeof( usbif_be_create_t       ), usbback_device_stimulus_cr },
    [ CMSG_USBIF_BE_DESTROY      ] =
      { sizeof( usbif_be_destroy_t      ), usbback_device_stimulus_dt },
    [ CMSG_USBIF_BE_CONNECT      ] =
      { sizeof( usbif_be_connect_t      ), usbback_device_stimulus_cn },
    [ CMSG_USBIF_BE_DISCONNECT   ] =
      { sizeof( usbif_be_disconnect_t   ), usbback_device_stimulus_dn },
    [ CMSG_USBIF_BE_CLAIM_PORT   ] =
      { sizeof( usbif_be_claim_port_t   ), usbback_device_stimulus_cp },
    [ CMSG_USBIF_BE_RELEASE_PORT ] =
      { sizeof( usbif_be_release_port_t ), usbback_device_stimulus_rp },
};

void usbback_device_handle_control_message
(
    struct usbback_device * device,
    ctrl_msg_t            * msg,
    unsigned long           id
)
{
    trace();

    /* Message wouldn't have made it to us unless it contained a header with */
    /* our handle in it.                                                     */

    ASSERT( msg->length >= sizeof( usbif_be_header_t ) );

    /* Ensure we see latest state of received_msg_free and receive exactly   */
    /* one message at a time.                                                */

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    if
    (
        (
            msg->subtype
            <
            ( sizeof( message_subtype_info ) / sizeof( struct message_info ) )
        )
        &&
        ( msg->length == message_subtype_info[ msg->subtype ].length )
        &&
        ( device->received_msg_free )
    )
    {
        device->received_msg_free = 0;

        device->received_msg = *msg;

        usbback_device_handle_stimulus
        (
            device,
            message_subtype_info[ msg->subtype ].stimulus,
            &flags
        );

        spin_unlock_irqrestore( &device->lock, flags );
    }
    else
    {
        spin_unlock_irqrestore( &device->lock, flags );

        msg->length = 0;

        ctrl_if_send_response( msg );
    }
}

static irqreturn_t usbback_device_interrupt
  ( int irq, void * context, struct pt_regs * regs )
{
  /*    trace(); */

    struct usbback_device * device = (struct usbback_device *)context;

    (void)usbback_work_schedule( &device->interrupt_1_work );

    return IRQ_HANDLED;
}

static void usbback_device_interrupt_1( void * data )
{
  /*    trace(); */

    struct usbback_device * device = (struct usbback_device *)data;

    struct list_head issue_list;

    INIT_LIST_HEAD( &issue_list );

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    if( !device->interrupt_1_running )
    {
        int found_some;

        do
        {
            found_some = 0;

            RING_IDX request_producer = device->back_ring.sring->req_prod;

            rmb();

            {
                RING_IDX i;

                for
                (
                    i = device->back_ring.req_cons;
                    ( device->driver_attached )
                    &&
                    ( !list_empty( &device->free_resource_list ) )
                    &&
                    /* FIXME --- Why 2 RING conditions here?  */
                    ( i != request_producer )
                    &&
                    ( !RING_REQUEST_CONS_OVERFLOW( &device->back_ring, i ) );
                    i++
                )
                {
                    found_some = 1;

                    usbif_request_t * usbif_request =
                      RING_GET_REQUEST( &device->back_ring, i );

                    struct usbback_device_resource * resource = list_entry
                    (
                        device->free_resource_list.next,
                        struct usbback_device_resource,
                        link
                    );

                    list_del_init( &resource->link );

                    memcpy
                    (
                        &resource->request.usbif_request,
                        usbif_request,
                        sizeof( resource->request.usbif_request )
                    );

                    list_add_tail( &resource->link, &issue_list );

                    usbback_device_get_connect_reference( device );
                }

                device->back_ring.req_cons = i;
            }

            if( found_some )
            {
                device->interrupt_1_running = 1;

                spin_unlock_irqrestore( &device->lock, flags );

                {
                    struct usbback_device_resource * resource;
                    struct usbback_device_resource * temp;

                    list_for_each_entry_safe
                      ( resource, temp, &issue_list, link )
                    {
                        list_del_init( &resource->link );

                        device->request_handler
                          ( &resource->request, device->driver_context );
                    }
                }

                spin_lock_irqsave( &device->lock, flags );

                device->interrupt_1_running = 0;
            }
        }
        while( found_some );
    }

    spin_unlock_irqrestore( &device->lock, flags );
}

void usbback_device_complete_request
(
    struct usbback_device_request * request,
    s16                             status,
    unsigned long                   length
)
{
  /*    trace(); */

    request->usbif_response.id        = request->usbif_request.id;
    request->usbif_response.operation = request->usbif_request.operation;
    request->usbif_response.status    = status;
    request->usbif_response.data      = 0;
    request->usbif_response.length    = length;

    struct usbback_device_resource * resource =
      container_of( request, struct usbback_device_resource, request );

    struct usbback_device * device = resource->device;

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    usbif_response_t * response =
      RING_GET_RESPONSE( &device->back_ring, device->back_ring.rsp_prod_pvt );

    memcpy
      ( response, &resource->request.usbif_response, sizeof( *response ) );

    wmb();

    device->back_ring.rsp_prod_pvt++;

    RING_PUSH_RESPONSES( &device->back_ring );

    list_add( &resource->link, &device->free_resource_list );

    spin_unlock_irqrestore( &device->lock, flags );

    notify_via_evtchn( device->evtchn );

    usbback_device_interrupt( 0, device, 0 );

    usbback_device_put_connect_reference( device );
}

static int usbback_device_shutdown( struct device * dev, void * ignored )
{
    trace();

    struct usbback_device * device = usbback_device_dev_to( dev );

    device->shutdown = 0;

    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        usbback_device_handle_stimulus
          ( device, usbback_device_stimulus_sd, &flags );

        spin_unlock_irqrestore( &device->lock, flags );
    }

    usbback_work_until( device->shutdown );

    return 0;
}

static void usbback_device_shutdown_all( void )
{
    trace();

    (void)bus_for_each_dev
      ( &usbback_bus_type, NULL, NULL, usbback_device_shutdown );
}

void usbback_device_shutdown_all_phase_1( void )
{
    trace();

    usbback_device_shutdown_all();
}

void usbback_device_shutdown_all_phase_2( void )
{
    trace();

    usbback_device_shutdown_all();
}

void usbback_device_attach_backend
(
    struct usbback_device * device,
    void                ( * request_handler )
      ( struct usbback_device_request * request, void * context ),
    void                ( * claim_port      )
      ( u32 port, char * path, void * context ),
    void                ( * release_port    )( u32 port, void * context ),
    void                ( * flush_requests )( void * context ),
    void                  * context
)
{
    trace();

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    device->request_handler = request_handler;
    device->claim_port      = claim_port;
    device->release_port    = release_port;
    device->flush_requests  = flush_requests;
    device->driver_context  = context;

    device->driver_attached = 1;

    spin_unlock_irqrestore( &device->lock, flags );
}

void usbback_device_detach_backend( struct usbback_device * device )
{
    trace();

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    device->driver_attached = 0;

    spin_unlock_irqrestore( &device->lock, flags );

    flush_scheduled_work();
}

domid_t usbback_device_query_remote_domain_id( struct usbback_device * device )
{
    trace();

    return device->remote_domain_id;
}

static void usbback_device_invalid_stimulus
  ( struct usbback_device * device, usbback_device_stimulus stimulus );

static void usbback_device_reject_message
  ( struct usbback_device * device, unsigned long * flags );

static void usbback_device_connect( struct usbback_device * device );

static void usbback_device_disconnect( struct usbback_device * device );

static void usbback_device_respond
  ( struct usbback_device * device, u32 status );

static void usbback_device_complete_shutdown( struct usbback_device * device );

static void usbback_device_unregister( struct usbback_device * device );

static void usbback_device_claim_port
  ( struct usbback_device * device, unsigned long * flags);

static void usbback_device_release_port
  ( struct usbback_device * device, unsigned long * flags );

static void usbback_device_handle_stimulus
(
    struct usbback_device   * device,
    usbback_device_stimulus   stimulus,
    unsigned long           * flags
)
{
    trace3
    (
        "device %p in state %d received stimulus %d",
        device,
        device->state,
        stimulus
    );

    switch( device->state )
    {
      case usbback_device_state_i:
        /* Uncreated */
        switch( stimulus )
        {
          case usbback_device_stimulus_cr:
            device->state = usbback_device_state_i_cr;
            usbback_device_respond( device, USBIF_BE_STATUS_OKAY );
            break;
          case usbback_device_stimulus_dt: /* FALLS THROUGH */
          case usbback_device_stimulus_cn: /* FALLS THROUGH */
          case usbback_device_stimulus_dn: /* FALLS THROUGH */
          case usbback_device_stimulus_cp: /* FALLS THROUGH */
          case usbback_device_stimulus_rp:
            usbback_device_reject_message( device, flags );
            break;
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_sd;
            usbback_device_complete_shutdown( device );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr:
        /* Creating/Responding to complete Disconnect */
        switch( stimulus )
        {
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_cr_sd;
            break;
          case usbback_device_stimulus_rs:
            device->state = usbback_device_state_i_cr_rs;
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_sd:
        /* Shutdown 1 completed */
        switch( stimulus )
        {
          case usbback_device_stimulus_cr: /* FALLS THROUGH */
          case usbback_device_stimulus_dt: /* FALLS THROUGH */
          case usbback_device_stimulus_cn: /* FALLS THROUGH */
          case usbback_device_stimulus_dn: /* FALLS THROUGH */
          case usbback_device_stimulus_cp: /* FALLS THROUGH */
          case usbback_device_stimulus_rp:
            break;
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_sd_sd;
            usbback_device_unregister( device );
            usbback_device_complete_shutdown( device );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_sd:
        /* Creating/Responding to complete Disconnect */
        /* Shutdown 1 outstanding */
        switch( stimulus )
        {
          case usbback_device_stimulus_rs:
            device->state = usbback_device_state_i_sd;
            usbback_device_complete_shutdown( device );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs:
        /* Created */
        switch( stimulus )
        {
          case usbback_device_stimulus_cr:
            usbback_device_reject_message( device, flags );
            break;
          case usbback_device_stimulus_dt:
            device->state = usbback_device_state_i_cr_rs_dt;
            usbback_device_unregister( device );
            break;
          case usbback_device_stimulus_cn:
            device->state = usbback_device_state_i_cr_rs_cn;
            usbback_device_connect( device );
            break;
          case usbback_device_stimulus_dn:
            usbback_device_reject_message( device, flags );
            break;
          case usbback_device_stimulus_cp:
            usbback_device_claim_port( device, flags );
            break;
          case usbback_device_stimulus_rp:
            usbback_device_release_port( device, flags );
            break;
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_sd;
            usbback_device_complete_shutdown( device );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_sd_sd:
        /* Shutdown 2 completed   */
        /* Unregister outstanding */
        switch( stimulus )
        {
          case usbback_device_stimulus_ud:
            device->state = usbback_device_state_i_sd_sd_ud;
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs_dt:
        /* Destroying             */
        /* Unregister outstanding */
        switch( stimulus )
        {
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_cr_rs_dt_sd;
            usbback_device_complete_shutdown( device );
            break;
          case usbback_device_stimulus_ud:
            device->state = usbback_device_state_i_cr_rs_dt_ud;
            usbback_device_respond( device, USBIF_BE_STATUS_OKAY );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs_cn:
        /* Connecting             */
        switch( stimulus )
        {
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_cr_rs_cn_sd;
            break;
          case usbback_device_stimulus_rs:
            device->state = usbback_device_state_i_cr_rs_cn_rs;
            break;
          case usbback_device_stimulus_rf:
            device->state = usbback_device_state_i_cr_rs;
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_sd_sd_ud:
        /* Unregistered */
        /* Shutdown/Destroyed */
        switch( stimulus )
        {
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs_dt_sd:
        /* Shutdown 1 completed   */
        /* Unregister outstanding */
        switch( stimulus )
        {
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_sd_sd;
            usbback_device_complete_shutdown( device );
            break;
          case usbback_device_stimulus_ud:
            device->state = usbback_device_state_i_cr_rs_dt_sd_ud;
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs_dt_ud:
        /* Destroying   */
        /* Unregistered */
        switch( stimulus )
        {
          case usbback_device_stimulus_rs:
            device->state = usbback_device_state_i_sd_sd_ud;
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs_cn_sd:
        /* Connecting             */
        /* Shutdown 1 outstanding */
        switch( stimulus )
        {
          case usbback_device_stimulus_rs:
            device->state = usbback_device_state_i_cr_rs_cn_sd_rs;
            usbback_device_disconnect( device );
            break;
          case usbback_device_stimulus_rf:
            device->state = usbback_device_state_i_sd;
            usbback_device_complete_shutdown( device );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs_cn_rs:
        /* Connected */
        switch( stimulus )
        {
          case usbback_device_stimulus_cr:
            usbback_device_reject_message( device, flags );
            break;
          case usbback_device_stimulus_dt:
            usbback_device_reject_message( device, flags );
            break;
          case usbback_device_stimulus_cn:
            usbback_device_reject_message( device, flags );
            break;
          case usbback_device_stimulus_dn:
            device->state = usbback_device_state_i_cr_rs_cn_rs_dn;
            usbback_device_disconnect( device );
            break;
          case usbback_device_stimulus_cp:
            usbback_device_claim_port( device, flags );
            break;
          case usbback_device_stimulus_rp:
            usbback_device_release_port( device, flags );
            break;
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_cr_rs_cn_sd_rs;
            usbback_device_disconnect( device );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs_dt_sd_ud:
        /* Destroying   */
        /* Unregistered */
        /* Shutdown 1 completed */
        switch( stimulus )
        {
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_sd_sd_ud;
            usbback_device_complete_shutdown( device );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs_cn_sd_rs:
        /* Disconnecting for shutdown */
        /* Shutdown 1 outstanding     */
        switch( stimulus )
        {
          case usbback_device_stimulus_cr: /* FALLS THROUGH */
          case usbback_device_stimulus_dt: /* FALLS THROUGH */
          case usbback_device_stimulus_cn: /* FALLS THROUGH */
          case usbback_device_stimulus_dn: /* FALLS THROUGH */
          case usbback_device_stimulus_cp: /* FALLS THROUGH */
          case usbback_device_stimulus_rp: /* FALLS THROUGH */
            break;
          case usbback_device_stimulus_dd:
            device->state = usbback_device_state_i_sd;
            usbback_device_complete_shutdown( device );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbback_device_state_i_cr_rs_cn_rs_dn:
        /* Disconnecting */
        switch( stimulus )
        {
          case usbback_device_stimulus_sd:
            device->state = usbback_device_state_i_cr_rs_cn_sd_rs;
            break;
          case usbback_device_stimulus_dd:
            device->state = usbback_device_state_i_cr;
            usbback_device_respond( device, USBIF_BE_STATUS_OKAY );
            break;
          default:
            usbback_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      default:
        usbback_device_invalid_stimulus( device, stimulus );
        break;
    }
}

static void usbback_device_invalid_stimulus
  ( struct usbback_device * device, usbback_device_stimulus stimulus )
{
    trace();

    printk
    (
        KERN_ERR "usbback: device %p in state %d"
        "received invalid stimulus %d",
        device,
        device->state,
        stimulus
    );
}

static void usbback_device_reject_message
  ( struct usbback_device * device, unsigned long * flags )
{
    trace();

    ctrl_msg_t msg = device->received_msg;

    device->received_msg_free = 1;

    spin_unlock_irqrestore( &device->lock, *flags );

    msg.length = 0;

    ctrl_if_send_response( &msg );

    spin_lock_irqsave( &device->lock, *flags );
}

static void usbback_device_connect( struct usbback_device * device )
{
    trace();

    int scheduled = usbback_work_schedule( &device->connect_1_work );

    ASSERT( scheduled );
}

static void usbback_device_connect_1( void * data )
{
    trace();

    u32 status = USBIF_BE_STATUS_OKAY;

    struct usbback_device * device = (struct usbback_device *)data;

    usbif_be_connect_t connect;

    memcpy
    (
        &connect,
        device->received_msg.msg,
        sizeof( connect )
    );

    struct vm_struct * vma = get_vm_area( PAGE_SIZE, VM_IOREMAP );

    if( vma == NULL )
    {
        status = USBIF_BE_STATUS_OUT_OF_MEMORY;

        goto EXIT_NO_VMA;
    }

    {
        int error;

        pgprot_t prot = __pgprot
          ( _PAGE_PRESENT | _PAGE_RW | _PAGE_DIRTY | _PAGE_ACCESSED );

        if
        (
            (
                error = direct_remap_area_pages
                (
                    &init_mm,
                    VMALLOC_VMADDR( vma->addr ),
                    connect.shmem_frame << PAGE_SHIFT,
                    PAGE_SIZE,
                    prot,
                    connect.domid
                )
            )
            !=
            0
        )
        {
            switch( error )
            {
              case -ENOMEM:
                status = USBIF_BE_STATUS_OUT_OF_MEMORY;
                break;
              case -EFAULT:
                status = USBIF_BE_STATUS_MAPPING_ERROR;
                break;
              default:
                status = USBIF_BE_STATUS_ERROR;
                break;
            }

            goto EXIT_NO_REMAP;
        }
    }

    usbif_sring_t * shared_ring = (usbif_sring_t *)vma->addr;

    SHARED_RING_INIT( shared_ring );

    BACK_RING_INIT( &device->back_ring, shared_ring );

    device->evtchn = connect.evtchn;

    device->irq = bind_evtchn_to_irq( device->evtchn );

    device->remote_domain_id = connect.domid;

    if
    (
        request_irq
        (
            device->irq,
            usbback_device_interrupt,
            SA_SAMPLE_RANDOM,
            "usbback",
            device
        )
        !=
        0
    )
    {
        goto EXIT_NO_IRQ;
    }

    usbback_device_get_connect_reference( device );

    goto SEND_RESPONSE;

  EXIT_NO_IRQ:

    unbind_evtchn_from_irq( device->evtchn );

  EXIT_NO_REMAP:

    vfree( vma->addr );

  EXIT_NO_VMA:

  SEND_RESPONSE:

    usbback_device_respond( device, status );
}

static void usbback_device_disconnect( struct usbback_device * device )
{
    trace();

    int scheduled = usbback_work_schedule( &device->disconnect_1_work );

    ASSERT( scheduled );
}

static void usbback_device_disconnect_1( void * data )
{
    trace();

    struct usbback_device * device = (struct usbback_device *)data;

    free_irq( device->irq, device );

    device->flush_requests( device->driver_context );

    usbback_device_put_connect_reference( device );
}

static void usbback_device_disconnect_2( void * data )
{
    trace();

    struct usbback_device * device = (struct usbback_device *)data;

    unbind_evtchn_from_irq( device->evtchn );

    vfree( device->back_ring.sring );

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    usbback_device_handle_stimulus
      ( device, usbback_device_stimulus_dd, &flags );

    spin_unlock_irqrestore( &device->lock, flags );
}

static void usbback_device_respond
  ( struct usbback_device * device, u32 status )
{
    trace();

    usbback_device_get( device );

    memcpy
    (
        &( (usbif_be_header_t *)device->received_msg.msg )->status,
        &status,
        sizeof( status )
    );

    int scheduled = usbback_work_schedule( &device->respond_1_work );

    ASSERT( scheduled );
}

static void usbback_device_respond_1( void * data )
{
    trace();

    struct usbback_device * device = (struct usbback_device *)data;

    ctrl_msg_t msg = device->received_msg;

    u32 status;

    memcpy
    (
        &status,
        &( (usbif_be_header_t *)msg.msg )->status,
        sizeof( status )
    );

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    if( status == USBIF_BE_STATUS_OKAY )
    {
        usbback_device_handle_stimulus
          ( device, usbback_device_stimulus_rs, &flags );
    }
    else
    {
        usbback_device_handle_stimulus
          ( device, usbback_device_stimulus_rf, &flags );
    }

    device->received_msg_free = 1;

    spin_unlock_irqrestore( &device->lock, flags );

    ctrl_if_send_response( &msg );

    usbback_device_put( device );
}

static void usbback_device_unregister( struct usbback_device * device )
{
    trace();

    usbback_device_get( device );

    int scheduled = usbback_work_schedule( &device->unregister_1_work );

    ASSERT( scheduled );
}

static void usbback_device_unregister_1( void * data )
{
    trace();

    struct usbback_device * device = (struct usbback_device *)data;

    down_write( &usbback_device_list_sem );

    list_del_init( &device->link );

    up_write( &usbback_device_list_sem );

    device_unregister( &device->dev );

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    usbback_device_handle_stimulus
      ( device, usbback_device_stimulus_ud, &flags );

    spin_unlock_irqrestore( &device->lock, flags );

    usbback_device_put( device );
}

static void usbback_device_complete_shutdown( struct usbback_device * device )
{
    trace();

    device->shutdown = 1;

    usbback_work_wake_up();
}

static void usbback_device_claim_port
  ( struct usbback_device * device, unsigned long * flags )
{
    trace();

    ctrl_msg_t msg = device->received_msg;

    device->received_msg_free = 1;

    spin_unlock_irqrestore( &device->lock, *flags );

    u32 port;

    memcpy
    (
        &port,
        &( (usbif_be_claim_port_t *)msg.msg )->port,
        sizeof( port )
    );

    device->claim_port
    (
        port,
        ( (usbif_be_claim_port_t *)msg.msg )->path,
        device->driver_context
    );

    u32 status = USBIF_BE_STATUS_OKAY;

    memcpy
    (
        &( (usbif_be_header_t *)msg.msg )->status,
        &status,
        sizeof( status )
    );

    ctrl_if_send_response( &msg );

    spin_lock_irqsave( &device->lock, *flags );
}

static void usbback_device_release_port
  ( struct usbback_device * device, unsigned long * flags )
{
    trace();

    ctrl_msg_t msg = device->received_msg;

    device->received_msg_free = 1;

    spin_unlock_irqrestore( &device->lock, *flags );

    u32 port;

    memcpy
    (
        &port,
        &( (usbif_be_release_port_t *)msg.msg )->port,
        sizeof( port )
    );

    device->release_port
    (
        port,
        device->driver_context
    );

    u32 status = USBIF_BE_STATUS_OKAY;

    memcpy
    (
        &( (usbif_be_header_t *)msg.msg )->status,
        &status,
        sizeof( status )
    );

    ctrl_if_send_response( &msg );

    spin_lock_irqsave( &device->lock, *flags );
}
