/*****************************************************************************/
/* The discovery code performs discovery of connections to front-ends and    */
/* allocates usbback_devices to represent them.                              */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on arch/xen/drivers/usbif/backend/control.c, original copyright     */
/* notice follows...                                                         */
/*****************************************************************************/

/******************************************************************************
 * arch/xen/drivers/usbif/backend/control.c
 * 
 * Routines for interfacing with the control plane.
 * 
 * Copyright (c) 2004, Keir Fraser
 */

#include <linux/module.h>
#include <linux/rwsem.h>
#include <asm-xen/ctrl_if.h>
#include "usbback_assert.h"
#include "usbback_device.h"
#include "usbback_discovery.h"
#include "usbback_trace.h"

static DECLARE_RWSEM( usbback_discovery_shutting_down_sem );

static int usbback_discovery_shutting_down = 0;

static void usbback_discovery_ctrl_if_receiver
  ( ctrl_msg_t * msg, unsigned long id )
{
    trace();

    if( msg->length >= sizeof( usbif_be_header_t ) )
    {
        usbif_be_header_t * header = (usbif_be_header_t *)&msg->msg[ 0 ];

        down_read( &usbback_discovery_shutting_down_sem );

        struct usbback_device * device = usbback_device_find_or_allocate
          ( header->domid, usbback_discovery_shutting_down );

        up_read( &usbback_discovery_shutting_down_sem );

        if( device != NULL )
        {
            usbback_device_handle_control_message( device, msg, id );

            usbback_device_put( device );
        }
        else
        {
            header->status = USBIF_BE_STATUS_OUT_OF_MEMORY;

            ctrl_if_send_response( msg );
        }
    }
    else
    {
        msg->length = 0;

        ctrl_if_send_response( msg );
    }
}

static void usbback_discovery_send_driver_status( int status )
{
    trace();

    control_msg_t msg =
    {
        .type    = CMSG_USBIF_BE,
        .subtype = CMSG_USBIF_BE_DRIVER_STATUS_CHANGED,
        .length  = sizeof( usbif_be_driver_status_changed_t ),
    };

    {
        usbif_be_driver_status_changed_t st = { .status = status, };

        memcpy( msg.msg, &st, sizeof( st ) );
    }

    if
    (
        ctrl_if_send_message_and_get_response( &msg, &msg, TASK_INTERRUPTIBLE )
        !=
        0
    )
    {
        ASSERT( 0 );
    }

    trace();
}

static int usbback_discovery_init_or_exit( int exit )
{
    trace();

    int return_value = 0;

    if( exit )
    {
        goto EXIT;
    }

    if
    (
        !ctrl_if_register_receiver
        (
            CMSG_USBIF_BE,
            usbback_discovery_ctrl_if_receiver,
            CALLBACK_IN_BLOCKING_CONTEXT
        )
    )
    {
        return_value = -EBUSY;

        goto EXIT_NO_CTRL_IF;
    }

    usbback_discovery_send_driver_status( USBIF_DRIVER_STATUS_UP );

    return 0;

  EXIT:

    down_write( &usbback_discovery_shutting_down_sem );

    /* Make sure any new devices created whilst we are shutting down get     */
    /* told we are shutting down...                                          */

    usbback_discovery_shutting_down = 1;

    /* ...before we tell all existing devices to stop using front end        */
    /* resources...                                                          */

    usbback_device_shutdown_all_phase_1();

    up_write( &usbback_discovery_shutting_down_sem );

    /* ...now all exisiting devices have stopped using front end resources   */
    /* and all new devices will not start using front end resources, we can  */
    /* disconnect from the control interface...                              */

    usbback_discovery_send_driver_status( USBIF_DRIVER_STATUS_DOWN );

    /* ...now we have received a response from the driver status down        */
    /* message there are no control messages in progress and we will receive */
    /* no more control messages...                                           */

    ctrl_if_unregister_receiver
      ( CMSG_USBIF_BE, usbback_discovery_ctrl_if_receiver );

    /* ...so we will not create any more devices, the existing devices are   */
    /* no longer required to respond to any control messages and we can      */
    /* perform the second phase of device shutdown which will result in the  */
    /* devices being freed once their reference counts fall to zero.         */

    usbback_device_shutdown_all_phase_2();

  EXIT_NO_CTRL_IF:

    return return_value;
}

int usbback_discovery_init( void )
{
    trace();

    return usbback_discovery_init_or_exit( 0 );
}

void usbback_discovery_exit( void )
{
    trace();

    (void)usbback_discovery_init_or_exit( 1 );
}
