/*****************************************************************************/
/* Device driver which drives usbback_device devices.                        */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

#include <linux/device.h>
#include <linux/usb.h>
#include "usbback_assert.h"
#include "usbback_bus.h"
#include "usbback_device.h"
#include "usbback_driver.h"
#include "usbback_driver_backend.h"
#include "usbback_driver_port.h"
#include "usbback_trace.h"

static int usbback_driver_probe_usb
  ( struct usb_interface * intf, const struct usb_device_id * id )
{
    trace();

    return usbback_driver_port_probe_usb( intf );
}

static void usbback_driver_disconnect_usb( struct usb_interface * intf )
{
    trace();

    usbback_driver_port_disconnect_usb( intf );
}

static int usbback_driver_probe_or_remove_backend
  ( struct device * dev, int remove )
{
    trace();

    struct usbback_driver_backend * backend;

    int return_value = 0;

    if( remove )
    {
        goto REMOVE;
    }

    backend = usbback_driver_backend_allocate( usbback_device_dev_to( dev ) );

    if( backend == NULL )
    {
        return_value = -ENOMEM;

        goto EXIT_NO_BACKEND;
    }

    dev_set_drvdata( dev, backend );

    return 0;

  REMOVE:

    backend = dev_get_drvdata( dev );

    usbback_driver_backend_shutdown( backend );

    dev_set_drvdata( dev, NULL );

  EXIT_NO_BACKEND:

    return return_value;
}

static int usbback_driver_probe_backend( struct device * dev )
{
    trace();

    return usbback_driver_probe_or_remove_backend( dev, 0 );
}

static int usbback_driver_remove_backend( struct device * dev )
{
    trace();

    return usbback_driver_probe_or_remove_backend( dev, 1 );
}

static void usbback_driver_shutdown_backend( struct device * dev )
{
    trace();

    /* FIXME implement device shutdown method (for kernel shutdown?) */
}

static struct usb_device_id usbback_driver_usb_id_table[] =
{
    { .driver_info = 1 }, /* Matches all devices. */
    {}
};

MODULE_DEVICE_TABLE( usb, usbback_driver_usb_id_table );

struct usb_driver usbback_driver_usb_driver =
{
    .owner      = THIS_MODULE,
    .name       = "usbback",
    .probe      = usbback_driver_probe_usb,
    .disconnect = usbback_driver_disconnect_usb,
    .id_table   = usbback_driver_usb_id_table,
};

static struct device_driver usbback_driver_backend_driver =
{
    .name     = "usbback",
    .bus      = &usbback_bus_type,
    .owner    = THIS_MODULE,
    .probe    = usbback_driver_probe_backend,
    .remove   = usbback_driver_remove_backend,
    .shutdown = usbback_driver_shutdown_backend,
};

static int usbback_driver_init_or_exit( int exit )
{
    trace();

    int return_value = 0;

    if( exit )
    {
        goto EXIT;
    }

    return_value = driver_register( &usbback_driver_backend_driver );

    if( return_value != 0 )
    {
        goto EXIT_NO_BACKEND_REGISTER;
    }

    return_value = usb_register( &usbback_driver_usb_driver );

    if( return_value != 0 )
    {
        goto EXIT_NO_USB_REGISTER;
    }

    return 0;

  EXIT:

    usb_deregister( &usbback_driver_usb_driver );

  EXIT_NO_USB_REGISTER:

    driver_unregister( &usbback_driver_backend_driver );

  EXIT_NO_BACKEND_REGISTER:

    return return_value;
}

int usbback_driver_init( void )
{
    trace();

    return usbback_driver_init_or_exit( 0 );
}

void usbback_driver_exit( void )
{
    trace();

    (void)usbback_driver_init_or_exit( 1 );
}
