/*****************************************************************************/
/* usbback_driver_backend is the representation in the driver of the         */
/* back-end side of a single front-end to back-end connection.               */
/* One of these objects is used by the usbback_driver to manage each         */
/* usbback_device.                                                           */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on arch/xen/drivers/usbif/backend/main.c, original copyright notice */
/* follows...                                                                */
/*****************************************************************************/

/******************************************************************************
 * arch/xen/drivers/usbif/backend/main.c
 * 
 * Backend for the Xen virtual USB driver - provides an abstraction of a
 * USB host controller to the corresponding frontend driver.
 *
 * by Mark Williamson
 * Copyright (c) 2004 Intel Research Cambridge
 * Copyright (c) 2004, 2005 Mark Williamson
 *
 * Based on arch/xen/drivers/blkif/backend/main.c
 * Copyright (c) 2003-2004, Keir Fraser & Steve Hand
 */

#include "usbback_assert.h"
#include "usbback_device.h"
#include "usbback_driver_backend.h"
#include "usbback_driver_port.h"
#include "usbback_trace.h"

#define USBBACK_DRIVER_PORT_COUNT 7

struct usbback_driver_backend
{
    struct usbback_device * device;
    struct usbback_driver_port port[ USBBACK_DRIVER_PORT_COUNT ];
};

static void usbback_driver_backend_request_handler
  ( struct usbback_device_request * request, void * context );

static void usbback_driver_backend_claim_port
  ( u32 port_index, char * path, void * context );

static void usbback_driver_backend_release_port
  ( u32 port_index, void * context );

static void usbback_driver_backend_flush_requests( void * context );

static int usbback_driver_backend_init_or_exit
(
    struct usbback_driver_backend * backend,
    struct usbback_device         * device,
    int                             exit
)
{
    trace();

    int return_value = 0;

    if( exit )
    {
        goto EXIT;
    }

    memset( backend, 0, sizeof( *backend ) );

    backend->device = device;

    int i;

    for( i = 0; i < USBBACK_DRIVER_PORT_COUNT; i++ )
    {
        return_value =
          usbback_driver_port_init( &backend->port[ i ], backend );

        if( return_value != 0 )
        {
            goto EXIT_NO_PORT;
        }
    }

    usbback_device_attach_backend
    (
        device,
        usbback_driver_backend_request_handler,
        usbback_driver_backend_claim_port,
        usbback_driver_backend_release_port,
        usbback_driver_backend_flush_requests,
        backend
    );

    return 0;

  EXIT:

    usbback_device_detach_backend( backend->device );

    i = USBBACK_DRIVER_PORT_COUNT;

  EXIT_NO_PORT:

    i--;

    for( ; i >= 0; i-- )
    {
        usbback_driver_port_shutdown( &backend->port[ i ] );
    }

    return return_value;
}

struct usbback_driver_backend * usbback_driver_backend_allocate
  ( struct usbback_device * device )
{
    struct usbback_driver_backend * backend =
      (struct usbback_driver_backend *)kmalloc
        ( sizeof( struct usbback_driver_backend ), GFP_KERNEL );

    if
    (
        ( backend != NULL )
        &&
        ( usbback_driver_backend_init_or_exit( backend, device, 0 ) != 0 )
    )
    {
        kfree( backend );

        backend = NULL;
    }

    return backend;
}

void usbback_driver_backend_shutdown
  ( struct usbback_driver_backend * backend )
{
    trace();

    (void)usbback_driver_backend_init_or_exit( backend, NULL, 1 );

    kfree( backend );
}

domid_t usbback_driver_backend_query_remote_domain_id
  ( struct usbback_driver_backend * backend )
{
    trace();

    return usbback_device_query_remote_domain_id( backend->device );
}

static void usbback_driver_backend_handle_probe
(
    struct usbback_driver_backend * backend,
    struct usbback_device_request * request
);

static void usbback_driver_backend_handle_reset
(
    struct usbback_driver_backend * backend,
    struct usbback_device_request * request
);

static void usbback_driver_backend_handle_io
(
    struct usbback_driver_backend * backend,
    struct usbback_device_request * request
);

static void usbback_driver_backend_handle_unlink
(
    struct usbback_driver_backend * backend,
    struct usbback_device_request * request
);

static void usbback_driver_backend_request_handler
  ( struct usbback_device_request * request, void * context )
{
  /*    trace(); */

    struct usbback_driver_backend * backend =
      (struct usbback_driver_backend *)context;

    switch( request->usbif_request.operation )
    {
      case USBIF_OP_PROBE:
        usbback_driver_backend_handle_probe( backend, request );
        break;
      case USBIF_OP_RESET:
        usbback_driver_backend_handle_reset( backend, request );
        break;
      case USBIF_OP_IO:
        usbback_driver_backend_handle_io( backend, request );
        break;
      case USBIF_OP_UNLINK:
        usbback_driver_backend_handle_unlink( backend, request );
        break;
      default:
        usbback_device_complete_request( request, USBIF_RSP_ERROR, 0 );
        break;
    }
}

static void usbback_driver_backend_handle_probe
(
    struct usbback_driver_backend * backend,
    struct usbback_device_request * request
)
{
  /*    trace(); */

    u32 port_index = request->usbif_request.port;

    s16 status = USBIF_RSP_ERROR;

    if( port_index < USBBACK_DRIVER_PORT_COUNT )
    {
        status = usbback_driver_port_usbdev_is_connected
          ( &backend->port[ port_index ] ) ? 1 : 0;
    }

    usbback_device_complete_request( request, status, 0 );
}

static void usbback_driver_backend_handle_reset
(
    struct usbback_driver_backend * backend,
    struct usbback_device_request * request
)
{
    trace();

    u32 port_index = request->usbif_request.port;

    s16 status = USBIF_RSP_ERROR;

    if( port_index < USBBACK_DRIVER_PORT_COUNT )
    {
        status = usbback_driver_port_reset( &backend->port[ port_index ] );
    }

    usbback_device_complete_request( request, status, 0 );
}

static struct usbback_driver_port *
  usbback_driver_backend_find_port_by_guest_address
(
    struct usbback_driver_backend * backend,
    unsigned long                   guest_address
)
{
    trace();

    int i;

    for( i = 0; i < USBBACK_DRIVER_PORT_COUNT; i++ )
    {
        struct usbback_driver_port * port = &backend->port[ i ];

        if( usbback_driver_port_match_guest_address( port, guest_address ) )
        {
            return port;
        }
    }

    return NULL;
}

static void usbback_driver_backend_handle_io
(
    struct usbback_driver_backend * backend,
    struct usbback_device_request * request
)
{
    trace();

    struct usbback_driver_port * port =
      usbback_driver_backend_find_port_by_guest_address
        ( backend, request->usbif_request.devnum );

    if( port != NULL )
    {
        usbback_driver_port_handle_io( port, request );
    }
    else
    {
        usbback_device_complete_request( request, USBIF_RSP_ERROR, 0 );
    }
}

static void usbback_driver_backend_handle_unlink
(
    struct usbback_driver_backend * backend,
    struct usbback_device_request * request
)
{
    trace();

    int i;

    for( i = 0; i < USBBACK_DRIVER_PORT_COUNT; i++ )
    {
        struct usbback_driver_port * port = &backend->port[ i ];

        if
        (
            usbback_driver_port_try_unlink
              ( port, request->usbif_request.unlink_id )
        )
        {
            break;
        }
    }

    usbback_device_complete_request( request, USBIF_RSP_OKAY, 0 );
}

static void usbback_driver_backend_claim_port
  ( u32 port_index, char * path, void * context )
{
    trace2( "port_index: %d, path: %s", port_index, path );

    ASSERT( port_index < USBBACK_DRIVER_PORT_COUNT );

    struct usbback_driver_backend * backend =
      (struct usbback_driver_backend *)context;

    usbback_driver_port_claim( &backend->port[ port_index ], path );

    bus_rescan_devices( &usb_bus_type );
}

static void usbback_driver_backend_release_port
  ( u32 port_index, void * context )
{
    trace1( "port_index: %d", port_index );

    ASSERT( port_index < USBBACK_DRIVER_PORT_COUNT );

    struct usbback_driver_backend * backend =
      (struct usbback_driver_backend *)context;

    usbback_driver_port_release( &backend->port[ port_index ] );

    bus_rescan_devices( &usb_bus_type );
}

static void usbback_driver_backend_flush_requests( void * context )
{
    trace();

    struct usbback_driver_backend * backend =
      (struct usbback_driver_backend *)context;

    int i;

    for( i = 0; i < USBBACK_DRIVER_PORT_COUNT; i++ )
    {
        struct usbback_driver_port * port = &backend->port[ i ];

        usbback_driver_port_flush_requests( port );
    }
}
