/*****************************************************************************/
/* Object which manages a single USB device and handles all the              */
/* usbback_device_requests for that device by assigning                      */
/* usbback_driver_port_resources to process them.                            */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on arch/xen/drivers/usbif/backend/main.c, original copyright notice */
/* follows...                                                                */
/*****************************************************************************/

/******************************************************************************
 * arch/xen/drivers/usbif/backend/main.c
 * 
 * Backend for the Xen virtual USB driver - provides an abstraction of a
 * USB host controller to the corresponding frontend driver.
 *
 * by Mark Williamson
 * Copyright (c) 2004 Intel Research Cambridge
 * Copyright (c) 2004, 2005 Mark Williamson
 *
 * Based on arch/xen/drivers/blkif/backend/main.c
 * Copyright (c) 2003-2004, Keir Fraser & Steve Hand
 */

#include <linux/rwsem.h>

#include "usbback_assert.h"
#include "usbback_driver_port.h"
#include "usbback_trace.h"

extern struct usb_driver usbback_driver_usb_driver;

static DECLARE_RWSEM( usbback_driver_port_list_sem );

static LIST_HEAD( usbback_driver_port_list );

typedef enum
{
    usbback_driver_port_stimulus_pc, /* Port claim                           */
    usbback_driver_port_stimulus_pr, /* Port release                         */
    usbback_driver_port_stimulus_up, /* USB probe                            */
    usbback_driver_port_stimulus_ud, /* USB disconnect                       */
    usbback_driver_port_stimulus_rq, /* Request queued/resource freed        */
    usbback_driver_port_stimulus_sd, /* Shutdown                             */
    usbback_driver_port_stimulus_ri, /* test_io requests idle synchronous    */
    usbback_driver_port_stimulus_rr, /* test_io requests running synchronous */
    usbback_driver_port_stimulus_rp, /* release performed */
}
usbback_driver_port_stimulus;

static void usbback_driver_port_handle_stimulus
(
    struct usbback_driver_port   * port,
    usbback_driver_port_stimulus   stimulus
);

static void usbback_driver_port_purge_io_1( void * data );

static void usbback_driver_port_perform_release_1( void * data );

int usbback_driver_port_init
(
    struct usbback_driver_port    * port,
    struct usbback_driver_backend * backend
)
{
    trace();

    memset( port, 0, sizeof( *port ) );

    port->backend = backend;

    INIT_LIST_HEAD( &port->link );

    spin_lock_init( &port->lock );

    port->state = usbback_driver_port_state_i;

    INIT_LIST_HEAD( &port->request_list );
    INIT_LIST_HEAD( &port->purge_list );
    INIT_LIST_HEAD( &port->free_resource_list );

    {
        int i;

        for( i = 0; i < USBBACK_DRIVER_PORT_RESOURCE_COUNT; i++ )
        {
            struct usbback_driver_port_resource * resource =
              &port->resources[ i ];

            usbback_driver_port_resource_init( resource, port );

            list_add( &resource->link, &port->free_resource_list );
        }
    }

    usbback_work_init
      ( &port->purge_io_1_work, usbback_driver_port_purge_io_1, port );

    usbback_work_init
    (
        &port->perform_release_1_work,
        usbback_driver_port_perform_release_1,
        port
    );

    return 0;
}

void usbback_driver_port_claim
  ( struct usbback_driver_port * port, char * path )
{
    trace();

    down_write( &usbback_driver_port_list_sem );

    strncpy( port->path, path, sizeof( port->path ) );

    list_add_tail( &port->link, &usbback_driver_port_list );

    spin_lock_irq( &port->lock );

    usbback_driver_port_handle_stimulus
      ( port, usbback_driver_port_stimulus_pc );

    spin_unlock_irq( &port->lock );

    up_write( &usbback_driver_port_list_sem );
}

void usbback_driver_port_release( struct usbback_driver_port * port )
{
    trace();

    down_write( &usbback_driver_port_list_sem );

    list_del_init( &port->link );

    up_write( &usbback_driver_port_list_sem );

    port->release = 0;

    spin_lock_irq( &port->lock );

    usbback_driver_port_handle_stimulus
      ( port, usbback_driver_port_stimulus_pr );

    spin_unlock_irq( &port->lock );

    usbback_work_until( port->release );
}

int usbback_driver_port_probe_usb( struct usb_interface * intf )
{
    trace();

    int return_value = 0;

    struct usb_device * usbdev = interface_to_usbdev( intf );

    printk( KERN_INFO "usbback: Probe for usb device %s\n", usbdev->devpath );

    struct usbback_driver_port * port = NULL;

    down_read( &usbback_driver_port_list_sem );

    {
        struct usbback_driver_port * cur = NULL;

        list_for_each_entry( cur, &usbback_driver_port_list, link )
        {
            printk
            (
                KERN_INFO
                "usbback: Testing against configured path %s:",
                cur->path
            );

            if
            (
                strncmp( cur->path, usbdev->devpath, sizeof( cur->path ) )
                ==
                0
            )
            {
                printk( " matches.\n" );

                port = cur;

                break;
            }
            else
            {
                printk( " doesn't match.\n" );
            }
        }
    }

    if( port == NULL )
    {
        printk( KERN_INFO "usbback: No match found.\n" );

        return_value = -ENODEV;

        goto EXIT_NO_PORT;
    }

    {
        int i;

        for( i = 0; i < usbdev->actconfig->desc.bNumInterfaces; i++ )
        {
            struct usb_interface * other_intf = usb_ifnum_to_if( usbdev, i );

            if( other_intf != intf )
            {
                if( usb_interface_claimed( other_intf ) )
                {
                    printk
                    (
                        KERN_INFO
                        "usbback: An interface of the matching device is "
                        "already in use by another driver.\n"
                    );

                    return_value = -EBUSY;

                    goto EXIT_INTERFACE_ALREADY_CLAIMED;
                }
            }
        }
    }

    {
        int i;

        for( i = 0; i < usbdev->actconfig->desc.bNumInterfaces; i++ )
        {
            struct usb_interface * other_intf = usb_ifnum_to_if( usbdev, i );

            if( other_intf != intf )
            {
                int error = usb_driver_claim_interface
                  ( &usbback_driver_usb_driver, other_intf, port );

                /* We already checked all the interfaces were available. */

                ASSERT( error == 0 );
            }
        }
    }

    usbdev = usb_get_dev( usbdev );

    usb_set_intfdata( intf, port );

    spin_lock_irq( &port->lock );

    port->usbdev_is_connected = 1;

    port->usbdev = usbdev;

    usbback_driver_port_handle_stimulus
      ( port, usbback_driver_port_stimulus_up );

    spin_unlock_irq( &port->lock );

  EXIT_INTERFACE_ALREADY_CLAIMED:

  EXIT_NO_PORT:

    up_read( &usbback_driver_port_list_sem );

    return return_value;
}

void usbback_driver_port_disconnect_usb( struct usb_interface * intf )
{
    trace();

    struct usbback_driver_port * port = usb_get_intfdata( intf );

    /* Protect against reentrant call when we release other interfaces. */

    if( port != NULL )
    {
        ASSERT( port->usbdev_is_connected );

        spin_lock_irq( &port->lock );

        port->usbdev_is_connected = 0;

        port->enabled = 0;

        port->disconnect = 0;

        usbback_driver_port_handle_stimulus
          ( port, usbback_driver_port_stimulus_ud );

        spin_unlock_irq( &port->lock );

        usbback_work_until( port->disconnect );

        struct usb_device * usbdev = port->usbdev;

        {
            int i;

            for( i = 0; i < usbdev->actconfig->desc.bNumInterfaces; i++ )
            {
                struct usb_interface * other_intf =
                  usb_ifnum_to_if( usbdev, i );

                if( other_intf != intf )
                {
                    /* Protect against reentrant call when we release other */
                    /* interfaces.                                          */

                    usb_set_intfdata( other_intf, NULL );

                    usb_driver_release_interface
                      ( &usbback_driver_usb_driver, other_intf );
                }
            }
        }

        usb_set_intfdata( intf, NULL );

        usb_put_dev( usbdev );
    }
}

int usbback_driver_port_reset( struct usbback_driver_port * port )
{
    trace();

    int status = USBIF_RSP_ERROR;

    spin_lock_irq( &port->lock );

    port->guest_address = 0;

    if( usbback_driver_port_usbdev_is_connected( port ) )
    {
        if( port->usbdev->speed == USB_SPEED_LOW )
        {
            status = USBIF_RSP_LOW_SPEED;
        }
        else if( port->usbdev->speed == USB_SPEED_HIGH )
        {
            status = USBIF_RSP_HIGH_SPEED;
        }
        else
        {
            status = USBIF_RSP_FULL_SPEED;
        }

        port->enabled = 1;
    }
    else
    {
        port->enabled = 0;
    }

    spin_unlock_irq( &port->lock );

    return status;
}

void usbback_driver_port_set_guest_address
  ( struct usbback_driver_port * port, unsigned long guest_address )
{
    trace();

    port->guest_address = guest_address;
}

int usbback_driver_port_match_guest_address
  ( struct usbback_driver_port * port, unsigned long guest_address )
{
    trace();

    return ( port->enabled && ( guest_address == port->guest_address ) );
}

void usbback_driver_port_handle_io
(
    struct usbback_driver_port    * port,
    struct usbback_device_request * request
)
{
    trace();

    unsigned long flags;

    spin_lock_irqsave( &port->lock, flags );

    list_add_tail( &request->link, &port->request_list );

    usbback_driver_port_handle_stimulus
      ( port, usbback_driver_port_stimulus_rq );

    spin_unlock_irqrestore( &port->lock, flags );
}

int usbback_driver_port_try_unlink
  ( struct usbback_driver_port * port, int unlink_id )
{
    trace();

    int return_value = 0;

    struct usbback_device_request * request = NULL;

    unsigned long flags;

    spin_lock_irqsave( &port->lock, flags );

    {
        struct usbback_device_request * temp;

        list_for_each_entry( temp, &port->request_list, link )
        {
            if( temp->usbif_request.unlink_id == unlink_id )
            {
                request = temp;

                list_del_init( &request->link );

                break;
            }
        }
    }

    if( request == NULL )
    {
        int i;

        for( i = 0; i < USBBACK_DRIVER_PORT_RESOURCE_COUNT; i++ )
        {
            struct usbback_driver_port_resource * resource =
              &port->resources[ i ];

            if
            (
                usbback_driver_port_resource_try_unlink( resource, unlink_id )
            )
            {
                return_value = 1;

                break;
            }
        }
    }

    spin_unlock_irqrestore( &port->lock, flags );

    if( request )
    {
        usbback_device_complete_request( request, USBIF_RSP_ERROR, 0 );

        return_value = 1;
    }

    return return_value;
}

static void usbback_driver_port_purge_io( struct usbback_driver_port * port );

void usbback_driver_port_flush_requests( struct usbback_driver_port * port )
{
    trace();

    unsigned long flags;

    spin_lock_irqsave( &port->lock, flags );

    usbback_driver_port_purge_io( port );

    {
        int i;

        for( i = 0; i < USBBACK_DRIVER_PORT_RESOURCE_COUNT; i++ )
        {
            struct usbback_driver_port_resource * resource =
              &port->resources[ i ];

            usbback_driver_port_resource_flush_request( resource );
        }
    }

    spin_unlock_irqrestore( &port->lock, flags );
}

void usbback_driver_port_shutdown( struct usbback_driver_port * port )
{
    trace();

    down_write( &usbback_driver_port_list_sem );

    list_del_init( &port->link );

    up_write( &usbback_driver_port_list_sem );

    port->shutdown = 0;

    spin_lock_irq( &port->lock );

    usbback_driver_port_handle_stimulus
      ( port, usbback_driver_port_stimulus_sd );

    spin_unlock_irq( &port->lock );

    usbback_work_until( port->shutdown );

    flush_scheduled_work();
}

static void usbback_driver_port_invalid_stimulus
  ( struct usbback_driver_port * port, usbback_driver_port_stimulus stimulus );

static void usbback_driver_port_purge_io( struct usbback_driver_port * port );

static void usbback_driver_port_test_io( struct usbback_driver_port * port );

static void usbback_driver_port_kick_io( struct usbback_driver_port * port );

static void usbback_driver_port_perform_release
  ( struct usbback_driver_port * port );

static void usbback_driver_port_complete_release
  ( struct usbback_driver_port * port );

static void usbback_driver_port_complete_disconnect
  ( struct usbback_driver_port * port );

static void usbback_driver_port_complete_shutdown
  ( struct usbback_driver_port * port );

static void usbback_driver_port_handle_stimulus
(
    struct usbback_driver_port   * port,
    usbback_driver_port_stimulus   stimulus
)
{
    trace3
    (
        "port %p in state %d received stimulus %d",
        port,
        port->state,
        stimulus
    );

    switch( port->state )
    {
      case usbback_driver_port_state_i:
        /* Released     */
        /* Disconnected */
        /* Idle         */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_pc:
            port->state = usbback_driver_port_state_i_pc;
            break;
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            break;
          case usbback_driver_port_stimulus_sd:
            usbback_driver_port_complete_shutdown( port );
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc:
        /* Claimed      */
        /* Disconnected */
        /* Idle         */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_pr:
            port->state = usbback_driver_port_state_i;
            usbback_driver_port_complete_release( port );
            break;
          case usbback_driver_port_stimulus_up:
            port->state = usbback_driver_port_state_i_pc_up;
            break;
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            break;
          case usbback_driver_port_stimulus_sd:
            port->state = usbback_driver_port_state_i;
            usbback_driver_port_complete_shutdown( port );
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up:
        /* Claimed       */
        /* Probed        */
        /* Maybe Running */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_pr:
            port->state = usbback_driver_port_state_i_pc_up_pr;
            usbback_driver_port_purge_io( port );
            usbback_driver_port_test_io( port );
            break;
          case usbback_driver_port_stimulus_ud:
            port->state = usbback_driver_port_state_i_pc_up_ud;
            usbback_driver_port_purge_io( port );
            usbback_driver_port_test_io( port );
            break;
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_kick_io( port );
            break;
          case usbback_driver_port_stimulus_sd:
            port->state = usbback_driver_port_state_i_pc_up_sd;
            usbback_driver_port_purge_io( port );
            usbback_driver_port_test_io( port );
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up_pr:
        /* Releasing       */
        /* Probed          */
        /* Testing/Running */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_ud:
            port->state = usbback_driver_port_state_i_pc_up_pr_ud;
            break;
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            usbback_driver_port_test_io( port );
            break;
          case usbback_driver_port_stimulus_ri:
            port->state = usbback_driver_port_state_i_pc_up_pr_ri;
            usbback_driver_port_perform_release( port );
            break;
          case usbback_driver_port_stimulus_rr:
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up_ud:
        /* Claimed         */
        /* Disconnecting   */
        /* Testing/Running */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_pr:
            port->state = usbback_driver_port_state_i_pc_up_pr_ud;
            break;
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            usbback_driver_port_test_io( port );
            break;
          case usbback_driver_port_stimulus_sd:
            port->state = usbback_driver_port_state_i_pc_up_ud_sd;
            break;
          case usbback_driver_port_stimulus_ri:
            port->state = usbback_driver_port_state_i_pc;
            usbback_driver_port_complete_disconnect( port );
            break;
          case usbback_driver_port_stimulus_rr:
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up_sd:
        /* Released        */
        /* Probed          */
        /* Shutting down   */
        /* Testing/Running */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_ud:
            port->state = usbback_driver_port_state_i_pc_up_ud_sd;
            break;
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            usbback_driver_port_test_io( port );
            break;
          case usbback_driver_port_stimulus_ri:
            port->state = usbback_driver_port_state_i_pc_up_sd_ri;
            usbback_driver_port_perform_release( port );
            break;
          case usbback_driver_port_stimulus_rr:
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up_pr_ud:
        /* Releasing       */
        /* Disconnecting   */
        /* Testing/Running */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            usbback_driver_port_test_io( port );
            break;
          case usbback_driver_port_stimulus_ri:
            port->state = usbback_driver_port_state_i_pc_up_pr_ud_ri;
            usbback_driver_port_perform_release( port );
            usbback_driver_port_complete_disconnect( port );
            break;
          case usbback_driver_port_stimulus_rr:
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up_pr_ri:
        /* Releasing          */
        /* Probed             */
        /* Idle               */
        /* Performing Release */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_ud:
            port->state = usbback_driver_port_state_i_pc_up_pr_ud_ri;
            usbback_driver_port_complete_disconnect( port );
            break;
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up_ud_sd:
        /* Released        */
        /* Disconnecting   */
        /* Shutting down   */
        /* Testing/Running */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            usbback_driver_port_test_io( port );
            break;
          case usbback_driver_port_stimulus_ri:
            port->state = usbback_driver_port_state_i_pc_up_ud_sd_ri;
            usbback_driver_port_perform_release( port );
            usbback_driver_port_complete_disconnect( port );
            break;
          case usbback_driver_port_stimulus_rr:
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up_sd_ri:
        /* Released           */
        /* Probed             */
        /* Shutting down      */
        /* Idle               */
        /* Performing release */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_ud:
            port->state = usbback_driver_port_state_i_pc_up_ud_sd_ri;
            usbback_driver_port_complete_disconnect( port );
            break;
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up_pr_ud_ri:
        /* Releasing          */
        /* Disconnected       */
        /* Idle               */
        /* Performing release */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            break;
          case usbback_driver_port_stimulus_rp:
            port->state = usbback_driver_port_state_i;
            usbback_driver_port_complete_release( port );
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      case usbback_driver_port_state_i_pc_up_ud_sd_ri:
        /* Released           */
        /* Disconnected       */
        /* Shutting down      */
        /* Idle               */
        /* Performing release */
        switch( stimulus )
        {
          case usbback_driver_port_stimulus_rq:
            usbback_driver_port_purge_io( port );
            break;
          case usbback_driver_port_stimulus_rp:
            port->state = usbback_driver_port_state_i;
            usbback_driver_port_complete_shutdown( port );
            break;
          default:
            usbback_driver_port_invalid_stimulus( port, stimulus );
            break;
        }
        break;
      default:
        usbback_driver_port_invalid_stimulus( port, stimulus );
        break;
    }
}

static void usbback_driver_port_invalid_stimulus
  ( struct usbback_driver_port * port, usbback_driver_port_stimulus stimulus )
{
    trace();

    printk
    (
        KERN_ERR "usbback: port %p in state %d"
        "received invalid stimulus %d",
        port,
        port->state,
        stimulus
    );
}

static void usbback_driver_port_purge_io( struct usbback_driver_port * port )
{
    trace();

    while( !list_empty( &port->request_list ) )
    {
        struct usbback_device_request * request = list_entry
          ( port->request_list.next, struct usbback_device_request, link );

        list_del_init( &request->link );

        list_add_tail( &request->link, &port->purge_list );
    }

    (void)usbback_work_schedule( &port->purge_io_1_work );
}

static void usbback_driver_port_purge_io_1( void * data )
{
    trace();

    struct usbback_driver_port * port = (struct usbback_driver_port *)data;

    spin_lock_irq( &port->lock );

    while( !list_empty( &port->purge_list ) )
    {
        struct usbback_device_request * request = list_entry
          ( port->purge_list.next, struct usbback_device_request, link );

        list_del_init( &request->link );

        spin_unlock_irq( &port->lock );

        usbback_device_complete_request( request, USBIF_RSP_ERROR, 0 );

        spin_lock_irq( &port->lock );
    }

    spin_unlock_irq( &port->lock );
}

static void usbback_driver_port_test_io( struct usbback_driver_port * port )
{
    trace();

    if( port->current_requests == 0 )
    {
        usbback_driver_port_handle_stimulus
          ( port, usbback_driver_port_stimulus_ri );
    }
    else
    {
        usbback_driver_port_handle_stimulus
          ( port, usbback_driver_port_stimulus_rr );
    }
}

static void usbback_driver_port_kick_io( struct usbback_driver_port * port )
{
    trace();

    while
    (
        ( !list_empty( &port->request_list ) )
        &&
        ( !list_empty( &port->free_resource_list ) )
    )
    {
        struct usbback_device_request * request = list_entry
          ( port->request_list.next, struct usbback_device_request, link );

        struct usbback_driver_port_resource * resource = list_entry
        (
            port->free_resource_list.next,
            struct usbback_driver_port_resource,
            link
        );

        list_del_init( &request->link );

        list_del_init( &resource->link );

        port->current_requests++;

        usbback_driver_port_resource_start_io( resource, request );
    }
}

void usbback_driver_port_resource_completed_io
(
    struct usbback_driver_port          * port,
    struct usbback_driver_port_resource * resource
)
{
    trace();

    unsigned long flags;

    spin_lock_irqsave( &port->lock, flags );

    list_add( &resource->link, &port->free_resource_list );

    port->current_requests--;

    usbback_driver_port_handle_stimulus
      ( port, usbback_driver_port_stimulus_rq );

    spin_unlock_irqrestore( &port->lock, flags );
}

static void usbback_driver_port_perform_release
  ( struct usbback_driver_port * port )
{
    trace();

    usb_get_dev( port->usbdev );

    int scheduled = usbback_work_schedule( &port->perform_release_1_work );

    ASSERT( scheduled );
}

static void usbback_driver_port_perform_release_1( void * data )
{
    trace();

    struct usbback_driver_port * port = (struct usbback_driver_port *)data;

    usb_lock_device( port->usbdev );

    down_write( &usb_bus_type.subsys.rwsem );

    if( port->usbdev_is_connected )
    {
        usb_driver_release_interface
          ( &usbback_driver_usb_driver, usb_ifnum_to_if( port->usbdev, 0 ) );
    }

    up_write( &usb_bus_type.subsys.rwsem );

    usb_unlock_device( port->usbdev );

    usb_put_dev( port->usbdev );

    spin_lock_irq( &port->lock );

    usbback_driver_port_handle_stimulus
      ( port, usbback_driver_port_stimulus_rp );

    spin_unlock_irq( &port->lock );
}

static void usbback_driver_port_complete_release
  ( struct usbback_driver_port * port )
{
    trace();

    port->release = 1;

    usbback_work_wake_up();
}

static void usbback_driver_port_complete_disconnect
  ( struct usbback_driver_port * port )
{
    trace();

    port->disconnect = 1;

    usbback_work_wake_up();
}

static void usbback_driver_port_complete_shutdown
  ( struct usbback_driver_port * port )
{
    trace();

    port->shutdown = 1;

    usbback_work_wake_up();
}
