/*****************************************************************************/
/* Object which manages a single USB device and handles all the              */
/* usbback_device_requests for that device by assigning                      */
/* usbback_driver_port_resources to process them.                            */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

#ifndef USBBACK_DRIVER_PORT_H
#define USBBACK_DRIVER_PORT_H

#include <linux/usb.h>
#include "usbback_assert.h"
#include "usbback_driver_backend.h"
#include "usbback_driver_port_resource.h"
#include "usbback_trace.h"
#include "usbback_work.h"

int usbback_driver_port_init
(
    struct usbback_driver_port    * port,
    struct usbback_driver_backend * backend
);

void usbback_driver_port_claim
  ( struct usbback_driver_port * port, char * path );

void usbback_driver_port_release
  ( struct usbback_driver_port * port );

int usbback_driver_port_probe_usb( struct usb_interface * intf );

void usbback_driver_port_disconnect_usb( struct usb_interface * intf );

/* On success, usbback_driver_port_reset returns USBIF_RSP_FULL_SPEED,       */
/* USBIF_RSP_LOW_SPEED or USBIF_RSP_HIGH_SPEED.                              */

int usbback_driver_port_reset( struct usbback_driver_port * port );

void usbback_driver_port_set_guest_address
  ( struct usbback_driver_port * port, unsigned long guest_address );

int usbback_driver_port_match_guest_address
  ( struct usbback_driver_port * port, unsigned long guest_address );

void usbback_driver_port_handle_io
(
    struct usbback_driver_port    * port,
    struct usbback_device_request * request
);

int usbback_driver_port_try_unlink
  ( struct usbback_driver_port * port, int unlink_id );

void usbback_driver_port_flush_requests( struct usbback_driver_port * port );

void usbback_driver_port_shutdown( struct usbback_driver_port * port );

typedef enum
{
    usbback_driver_port_state_i,
    usbback_driver_port_state_i_pc,
    usbback_driver_port_state_i_pc_up,
    usbback_driver_port_state_i_pc_up_pr,
    usbback_driver_port_state_i_pc_up_ud,
    usbback_driver_port_state_i_pc_up_sd,
    usbback_driver_port_state_i_pc_up_pr_ud,
    usbback_driver_port_state_i_pc_up_pr_ri,
    usbback_driver_port_state_i_pc_up_ud_sd,
    usbback_driver_port_state_i_pc_up_sd_ri,
    usbback_driver_port_state_i_pc_up_pr_ud_ri,
    usbback_driver_port_state_i_pc_up_ud_sd_ri
}
usbback_driver_port_state;

#define USBBACK_DRIVER_PORT_RESOURCE_COUNT 4

struct usbback_driver_port
{
    struct usbback_driver_backend * backend;
    struct list_head                link;
    char                            path[ USB_PATH_LEN ];
    struct usb_device             * usbdev;
    spinlock_t                      lock;
    usbback_driver_port_state       state;
    int                             usbdev_is_connected;
    int                             enabled;
    unsigned long                   guest_address;
    struct list_head                request_list;
    struct list_head                purge_list;
    struct list_head                free_resource_list;
    unsigned long                   current_requests;
    struct usbback_driver_port_resource resources
      [ USBBACK_DRIVER_PORT_RESOURCE_COUNT ];
    struct usbback_work             purge_io_1_work;
    struct usbback_work             perform_release_1_work;
    int                             release;
    int                             disconnect;
    int                             shutdown;
};

static inline struct usb_device * usbback_driver_port_query_usbdev
  ( struct usbback_driver_port * port )
{
    trace();

    ASSERT( port->usbdev != NULL );

    return port->usbdev;
}

static inline domid_t usbback_driver_port_query_remote_domain_id
  ( struct usbback_driver_port * port )
{
    trace();

    return usbback_driver_backend_query_remote_domain_id( port->backend );
}

static inline int usbback_driver_port_usbdev_is_connected
  ( struct usbback_driver_port * port )
{
  /*    trace(); */

    return port->usbdev_is_connected;
}

void usbback_driver_port_resource_completed_io
(
    struct usbback_driver_port          * port,
    struct usbback_driver_port_resource * resource
);

#endif
