/*****************************************************************************/
/* Resource which processes a usbback_device_request by translating it into  */
/* an URB and submitting it to the USB stack.                                */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on arch/xen/drivers/usbif/backend/main.c, original copyright notice */
/* follows...                                                                */
/*****************************************************************************/

/******************************************************************************
 * arch/xen/drivers/usbif/backend/main.c
 * 
 * Backend for the Xen virtual USB driver - provides an abstraction of a
 * USB host controller to the corresponding frontend driver.
 *
 * by Mark Williamson
 * Copyright (c) 2004 Intel Research Cambridge
 * Copyright (c) 2004, 2005 Mark Williamson
 *
 * Based on arch/xen/drivers/blkif/backend/main.c
 * Copyright (c) 2003-2004, Keir Fraser & Steve Hand
 */

#include "usbback_assert.h"
#include "usbback_driver_port.h"
#include "usbback_driver_port_resource.h"
#include "usbback_trace.h"

static USBBACK_CALLBACK_SERIALISER
  ( usbback_driver_port_resource_submit_urb_serialiser );

typedef enum
{
    usbback_driver_port_resource_stimulus_st, /* Start IO          */
    usbback_driver_port_resource_stimulus_tu, /* Try unlink        */
    usbback_driver_port_resource_stimulus_ms, /* Map success       */
    usbback_driver_port_resource_stimulus_mf, /* Map failure       */
    usbback_driver_port_resource_stimulus_us, /* URB submitted     */
    usbback_driver_port_resource_stimulus_un, /* URB not submitted */
    usbback_driver_port_resource_stimulus_uc, /* URB completed     */
    usbback_driver_port_resource_stimulus_ul, /* Unlink completed  */
}
usbback_driver_port_resource_stimulus;

static void usbback_driver_port_resource_handle_stimulus
(
    struct usbback_driver_port_resource   * resource,
    usbback_driver_port_resource_stimulus   stimulus
);

static void usbback_driver_port_resource_submit_urb_1
  ( struct usbback_callback * callback );

static void usbback_driver_port_resource_unlink_urb_1( void * data );

static void usbback_driver_port_resource_complete_1( void * data );

static void usbback_driver_port_resource_map_callback
  ( struct usbback_callback * callback );

void usbback_driver_port_resource_init
(
    struct usbback_driver_port_resource * resource,
    struct usbback_driver_port          * port
)
{
    trace();

    resource->port = port;

    INIT_LIST_HEAD( &resource->link );

    spin_lock_init( &resource->lock );

    resource->state = usbback_driver_port_resource_state_i;

    resource->device_request = NULL;

    resource->buffer = 0;

    usbback_callback_init
    (
        &resource->submit_urb_1_callback,
        usbback_driver_port_resource_submit_urb_1
    );

    usbback_work_init
    (
        &resource->unlink_urb_1_work,
        usbback_driver_port_resource_unlink_urb_1,
        resource
    );

    usbback_work_init
    (
        &resource->complete_1_work,
        usbback_driver_port_resource_complete_1,
        resource
    );

    usbback_callback_init
    (
        usbback_mapper_request_to_callback( &resource->mapper_request ),
        usbback_driver_port_resource_map_callback
    );
}

void usbback_driver_port_resource_start_io
(
    struct usbback_driver_port_resource * resource,
    struct usbback_device_request       * request
)
{
    trace();

    unsigned long flags;

    spin_lock_irqsave( &resource->lock, flags );

    resource->status        = USBIF_RSP_OKAY;
    resource->actual_length = 0;

    resource->device_request = request;

    usbback_driver_port_resource_handle_stimulus
      ( resource, usbback_driver_port_resource_stimulus_st );

    spin_unlock_irqrestore( &resource->lock, flags );
}

int usbback_driver_port_resource_try_unlink
  ( struct usbback_driver_port_resource * resource, int unlink_id )
{
    trace();

    int return_value = 0;

    unsigned long flags;

    spin_lock_irqsave( &resource->lock, flags );

    if
    (
        ( resource->device_request != NULL )
        &&
        ( resource->device_request->usbif_request.unlink_id == unlink_id )
    )
    {
        usbback_driver_port_resource_handle_stimulus
          ( resource, usbback_driver_port_resource_stimulus_tu );

        return_value = 1;
    }

    spin_unlock_irqrestore( &resource->lock, flags );

    return return_value;
}

void usbback_driver_port_resource_flush_request
  ( struct usbback_driver_port_resource * resource )
{
    trace();

    unsigned long flags;

    spin_lock_irqsave( &resource->lock, flags );

    if( resource->device_request != NULL )
    {
        usbback_driver_port_resource_handle_stimulus
          ( resource, usbback_driver_port_resource_stimulus_tu );
    }

    spin_unlock_irqrestore( &resource->lock, flags );
}

static void usbback_driver_port_resource_invalid_stimulus
(
    struct usbback_driver_port_resource   * resource,
    usbback_driver_port_resource_stimulus   stimulus
);

static void usbback_driver_port_resource_map_buffer
  ( struct usbback_driver_port_resource * resource );

static void usbback_driver_port_resource_abort_map
  ( struct usbback_driver_port_resource * resource );

static void usbback_driver_port_resource_submit_urb
  ( struct usbback_driver_port_resource * resource );

static void usbback_driver_port_resource_unlink_urb
  ( struct usbback_driver_port_resource * resource );

static void usbback_driver_port_resource_set_error
  ( struct usbback_driver_port_resource * resource );

static void usbback_driver_port_resource_complete
  ( struct usbback_driver_port_resource * resource );

static void usbback_driver_port_resource_handle_stimulus
(
    struct usbback_driver_port_resource   * resource,
    usbback_driver_port_resource_stimulus   stimulus
)
{
    trace3
    (
        "port resource %p in state %d received stimulus %d",
        resource,
        resource->state,
        stimulus
    );

    switch( resource->state )
    {
      case usbback_driver_port_resource_state_i:
        /* Idle */
        switch( stimulus )
        {
          case usbback_driver_port_resource_stimulus_st:
            resource->state = usbback_driver_port_resource_state_i_st;
            usbback_driver_port_resource_map_buffer( resource );
            break;
          default:
            usbback_driver_port_resource_invalid_stimulus
              ( resource, stimulus );
            break;
        }
        break;
      case usbback_driver_port_resource_state_i_st:
        /* Handling request       */
        /* map_buffer outstanding */
        switch( stimulus )
        {
          case usbback_driver_port_resource_stimulus_tu:
            resource->state = usbback_driver_port_resource_state_i_st_tu;
            usbback_driver_port_resource_abort_map( resource );
            break;
          case usbback_driver_port_resource_stimulus_ms:
            resource->state = usbback_driver_port_resource_state_i_st_ms;
            usbback_driver_port_resource_submit_urb( resource );
            break;
          case usbback_driver_port_resource_stimulus_mf:
            resource->state = usbback_driver_port_resource_state_i;
            usbback_driver_port_resource_complete( resource );
            break;
          default:
            usbback_driver_port_resource_invalid_stimulus
              ( resource, stimulus );
            break;
        }
        break;
      case usbback_driver_port_resource_state_i_st_tu:
        /* Handling request       */
        /* map_buffer outstanding */
        /* unlinking              */
        switch( stimulus )
        {
          case usbback_driver_port_resource_stimulus_tu:
            break;
          case usbback_driver_port_resource_stimulus_ms:
            resource->state = usbback_driver_port_resource_state_i;
            usbback_driver_port_resource_set_error( resource );
            usbback_driver_port_resource_complete( resource );
            break;
          case usbback_driver_port_resource_stimulus_mf:
            resource->state = usbback_driver_port_resource_state_i;
            usbback_driver_port_resource_complete( resource );
            break;
          default:
            usbback_driver_port_resource_invalid_stimulus
              ( resource, stimulus );
            break;
        }
        break;
      case usbback_driver_port_resource_state_i_st_ms:
        /* Handling request       */
        /* submit_urb outstanding */
        switch( stimulus )
        {
          case usbback_driver_port_resource_stimulus_tu:
            resource->state = usbback_driver_port_resource_state_i_st_ms_tu;
            break;
          case usbback_driver_port_resource_stimulus_us:
            resource->state = usbback_driver_port_resource_state_i_st_ms_us;
            break;
          case usbback_driver_port_resource_stimulus_un:
            resource->state = usbback_driver_port_resource_state_i;
            usbback_driver_port_resource_complete( resource );
            break;
          case usbback_driver_port_resource_stimulus_uc:
            resource->state = usbback_driver_port_resource_state_i_st_ms_uc;
            break;
          default:
            usbback_driver_port_resource_invalid_stimulus
              ( resource, stimulus );
            break;
        }
        break;
      case usbback_driver_port_resource_state_i_st_ms_tu:
        /* Handling request       */
        /* submit_urb outstanding */
        /* unlinking              */
        switch( stimulus )
        {
          case usbback_driver_port_resource_stimulus_tu:
            break;
          case usbback_driver_port_resource_stimulus_us:
            resource->state = usbback_driver_port_resource_state_i_st_ms_tu_us;
            usbback_driver_port_resource_unlink_urb( resource );
            break;
          case usbback_driver_port_resource_stimulus_un:
            resource->state = usbback_driver_port_resource_state_i;
            usbback_driver_port_resource_complete( resource );
            break;
          case usbback_driver_port_resource_stimulus_uc:
            resource->state = usbback_driver_port_resource_state_i_st_ms_uc;
            break;
          default:
            usbback_driver_port_resource_invalid_stimulus
              ( resource, stimulus );
            break;
        }
        break;
      case usbback_driver_port_resource_state_i_st_ms_us:
        /* Handling request */
        /* URB submitted    */
        switch( stimulus )
        {
          case usbback_driver_port_resource_stimulus_tu:
            resource->state = usbback_driver_port_resource_state_i_st_ms_tu_us;
            usbback_driver_port_resource_unlink_urb( resource );
            break;
          case usbback_driver_port_resource_stimulus_uc:
            resource->state = usbback_driver_port_resource_state_i;
            usbback_driver_port_resource_complete( resource );
            break;
          default:
            usbback_driver_port_resource_invalid_stimulus
              ( resource, stimulus );
            break;
        }
        break;
      case usbback_driver_port_resource_state_i_st_ms_uc:
        /* Handling request       */
        /* submit_urb outstanding */
        /* URB completed          */
        switch( stimulus )
        {
          case usbback_driver_port_resource_stimulus_tu:
            break;
          case usbback_driver_port_resource_stimulus_us:
            resource->state = usbback_driver_port_resource_state_i;
            usbback_driver_port_resource_complete( resource );
            break;
          default:
            usbback_driver_port_resource_invalid_stimulus
              ( resource, stimulus );
            break;
        }
        break;
      case usbback_driver_port_resource_state_i_st_ms_tu_us:
        /* Handling request       */
        /* URB submitted    */
        /* unlink_urb outstanding */
        switch( stimulus )
        {
          case usbback_driver_port_resource_stimulus_tu:
            break;
          case usbback_driver_port_resource_stimulus_uc:
          case usbback_driver_port_resource_stimulus_ul:
            resource->state =
              usbback_driver_port_resource_state_i_st_ms_tu_us_uc;
            break;
          default:
            usbback_driver_port_resource_invalid_stimulus
              ( resource, stimulus );
            break;
        }
        break;
      case usbback_driver_port_resource_state_i_st_ms_tu_us_uc:
        /* Handling request                        */
        /* URB submitted or unlink_urb outstanding */
        switch( stimulus )
        {
          case usbback_driver_port_resource_stimulus_tu:
            break;
          case usbback_driver_port_resource_stimulus_uc:
          case usbback_driver_port_resource_stimulus_ul:
            resource->state = usbback_driver_port_resource_state_i;
            usbback_driver_port_resource_complete( resource );
            break;
          default:
            usbback_driver_port_resource_invalid_stimulus
              ( resource, stimulus );
            break;
        }
        break;
      default:
        usbback_driver_port_resource_invalid_stimulus( resource, stimulus );
        break;
    }
}

static void usbback_driver_port_resource_invalid_stimulus
(
    struct usbback_driver_port_resource   * resource,
    usbback_driver_port_resource_stimulus   stimulus
)
{
    trace();

    printk
    (
        KERN_ERR "usbback: port resource %p in state %d"
        "received invalid stimulus %d",
        resource,
        resource->state,
        stimulus
    );
}

static void usbback_driver_port_resource_map_buffer
  ( struct usbback_driver_port_resource * resource )
{
    trace();

    int element_count = 1;

    {
        usbif_request_t * usbif_request =
          &resource->device_request->usbif_request;

        struct usbback_mapper_element * element = resource->mapper_element;

        element->domain  =
          usbback_driver_port_query_remote_domain_id( resource->port );
        element->address = usbif_request->transfer_buffer;
        element->length  = usbif_request->length;

        if( usbif_request->num_iso > 0 )
        {
            element_count++;
            element++;

            element->domain  =
              usbback_driver_port_query_remote_domain_id( resource->port );
            element->address = usbif_request->iso_schedule;
            /* FIXME: more than a page of schedule? */
            element->length  = PAGE_SIZE;
        }
    }

    usbback_mapper_map
    (
        element_count,
        resource->mapper_element,
        &resource->mapper_request
    );
}

static void usbback_driver_port_resource_abort_map
  ( struct usbback_driver_port_resource * resource )
{
    trace();

    usbback_mapper_abort_map( &resource->mapper_request );
}

static void usbback_driver_port_resource_map_callback
  ( struct usbback_callback * callback )
{
    trace();

    struct usbback_mapper_request * mapper_request =
      usbback_mapper_request_callback_to( callback );

    struct usbback_driver_port_resource * resource = container_of
      ( mapper_request, struct usbback_driver_port_resource, mapper_request );

    unsigned long flags;

    spin_lock_irqsave( &resource->lock, flags );

    if( !usbback_callback_query_error( callback ) )
    {
        resource->buffer = 1;

        usbback_driver_port_resource_handle_stimulus
          ( resource, usbback_driver_port_resource_stimulus_ms );
    }
    else
    {
        resource->status = USBIF_RSP_ERROR;

        usbback_driver_port_resource_handle_stimulus
          ( resource, usbback_driver_port_resource_stimulus_mf );
    }

    spin_unlock_irqrestore( &resource->lock, flags );
}

static void usbback_driver_port_resource_submit_urb
  ( struct usbback_driver_port_resource * resource )
{
    trace();

    usbback_callback_serialiser_complete_callback
    (
        &usbback_driver_port_resource_submit_urb_serialiser,
        &resource->submit_urb_1_callback,
        0
    );
}

static void usbback_driver_port_resource_end_io
  ( struct urb * urb, struct pt_regs * regs );

static int check_iso_schedule( struct urb * urb )
{
    int i;

    unsigned long total_length = 0;

    for( i = 0; i < urb->number_of_packets; i++ )
    {
        struct usb_iso_packet_descriptor * desc = &urb->iso_frame_desc[ i ];

        total_length += desc->length;

        if
        (
            ( desc->offset > urb->transfer_buffer_length )
            ||
            ( desc->length > ( urb->transfer_buffer_length - desc->offset ) )
            ||
            ( total_length > urb->transfer_buffer_length )
        )
        {
            return -EINVAL;
        }
    }

    return 0;
}

static void usbback_driver_port_resource_submit_urb_1
  ( struct usbback_callback * callback )
{
    trace();

    struct usbback_driver_port_resource * resource = container_of
      ( callback, struct usbback_driver_port_resource, submit_urb_1_callback );

    struct usbback_device_request * request = resource->device_request;

    usbif_request_t * usbif_request = &request->usbif_request;

    if( usbif_request->pipe_type == PIPE_CONTROL )
    {
        struct usb_ctrlrequest * ctrl =
          (struct usb_ctrlrequest *)usbif_request->setup;

        if
        (
            (
                ctrl->bRequestType
                ==
                ( USB_DIR_OUT | USB_TYPE_STANDARD | USB_RECIP_DEVICE )
            )
        )
        {
            if( ctrl->bRequest == USB_REQ_SET_ADDRESS )
            {
                usbback_driver_port_set_guest_address
                  ( resource->port, le16_to_cpu( ctrl->wValue ) );

                goto HANDLED_SPECIAL_CASE;
            }
            else if( ctrl->bRequest == USB_REQ_SET_CONFIGURATION )
            {
                /* FIXME: what to do for set configuration? */

                goto HANDLED_SPECIAL_CASE;
            }
        }
    }

    struct urb * urb = resource->urb =
      usb_alloc_urb( usbif_request->num_iso, GFP_KERNEL );

    if( urb == NULL )
    {
        resource->status = USBIF_RSP_ERROR;

        goto NO_URB;
    }

    unsigned int pipe =
    (
        ( (unsigned int)usbif_request->direction << 7 )
        |
        (
            (unsigned int)usbback_driver_port_query_usbdev( resource->port )->
              devnum
            <<
            8
        )
        |
        ( (unsigned int)usbif_request->endpoint << 15 )
        |
        ( (unsigned int)usbif_request->pipe_type << 30 )
    );

    if( usbif_request->pipe_type == PIPE_CONTROL )
    {
        usb_fill_control_urb
        (
            urb,
            usbback_driver_port_query_usbdev( resource->port ),
            pipe,
            usbif_request->setup,
            resource->mapper_element[ 0 ].buffer,
            usbif_request->length,
            usbback_driver_port_resource_end_io,
            resource
        );
    }
    else if( usbif_request->pipe_type == PIPE_BULK )
    {
        usb_fill_bulk_urb
        (
            urb,
            usbback_driver_port_query_usbdev( resource->port ),
            pipe,
            resource->mapper_element[ 0 ].buffer,
            usbif_request->length,
            usbback_driver_port_resource_end_io,
            resource
        );
    }
    else if( usbif_request->pipe_type == PIPE_INTERRUPT )
    {
        /* FIXME: hacking the interval like this is a bit unclean. */
        /* should probably convert back to exponential form for    */
        /* high speed transfers and then pass the value into       */
        /* fill_int_urb.                                           */
        usb_fill_int_urb
        (
            urb,
            usbback_driver_port_query_usbdev( resource->port ),
            pipe,
            resource->mapper_element[ 0 ].buffer,
            usbif_request->length,
            usbback_driver_port_resource_end_io,
            resource,
            1 /* For the time being... */
        );

        /* ...now set the real value. */
        urb->interval = usbif_request->interval;
    }
    else
    {
        ASSERT( usbif_request->pipe_type == PIPE_ISOCHRONOUS );

        /* FIXME: where's usb_fill_isoc_urb ?!? */

        spin_lock_init(&urb->lock);
        urb->dev = usbback_driver_port_query_usbdev( resource->port );
        urb->pipe = pipe;
        urb->transfer_buffer = resource->mapper_element[ 0 ].buffer;
        urb->transfer_buffer_length = usbif_request->length;
        urb->complete = usbback_driver_port_resource_end_io;
        urb->context = resource;
        urb->number_of_packets = usbif_request->num_iso;
        urb->interval = usbif_request->interval;
        urb->start_frame = 0;
        urb->transfer_flags |= URB_ISO_ASAP;
        {
            usbif_iso_t * schedule = resource->mapper_element[ 1 ].buffer;

            int i;

            for( i = 0; i < usbif_request->num_iso; i++ )
            {
                urb->iso_frame_desc[ i ].offset = schedule[ i ].buffer_offset;
                urb->iso_frame_desc[ i ].length = schedule[ i ].length;

                urb->iso_frame_desc[ i ].actual_length = 0;
                urb->iso_frame_desc[ i ].status        = 0;
            }
        }

        if( !check_iso_schedule( urb ) )
        {
            resource->status = USBIF_RSP_ERROR;

            goto SCHEDULE_ERROR;
        }
    }

    /* On the backend, all unlinks are asynchronous. */

    urb->transfer_flags |= URB_ASYNC_UNLINK;

    /* FIXME: transfer flags are ugly and linux specific */

    if( usbif_request->transfer_flags & 0x0001 )
    {
        urb->transfer_flags |= URB_SHORT_NOT_OK;
    }
    if( usbif_request->transfer_flags & 0x0040 )
    {
        urb->transfer_flags |= URB_ZERO_PACKET;
    }

    if( usb_submit_urb( urb, GFP_KERNEL ) != 0 )
    {
        printk( KERN_WARNING "URB %p submission failed.\n", urb );

        resource->status = USBIF_RSP_ERROR;

        goto URB_ERROR;
    }

    {
        unsigned long flags;

        spin_lock_irqsave( &resource->lock, flags );

        usbback_driver_port_resource_handle_stimulus
          ( resource, usbback_driver_port_resource_stimulus_us );

        spin_unlock_irqrestore( &resource->lock, flags );
    }

    return;

  URB_ERROR:

  SCHEDULE_ERROR:

    usb_free_urb( urb );

  NO_URB:
  HANDLED_SPECIAL_CASE:

    {
        unsigned long flags;

        spin_lock_irqsave( &resource->lock, flags );

        usbback_driver_port_resource_handle_stimulus
          ( resource, usbback_driver_port_resource_stimulus_un );

        spin_unlock_irqrestore( &resource->lock, flags );
    }
}

static void usbback_driver_port_resource_end_io
  ( struct urb * urb, struct pt_regs * regs )
{
    trace();

    struct usbback_driver_port_resource * resource = urb->context;

    if( urb->status != 0 )
    {
        printk( KERN_WARNING "URB %p failed. Status %d\n", urb, urb->status );

        resource->status = USBIF_RSP_ERROR;
    }

    resource->actual_length = urb->actual_length;

    if( usb_pipetype( urb->pipe ) == PIPE_ISOCHRONOUS )
    {
        usbif_iso_t * schedule = resource->mapper_element[ 1 ].buffer;

        int i;

        for( i = 0; i < urb->number_of_packets; i++ )
        {
            schedule[ i ].length = urb->iso_frame_desc[ i ].actual_length;
            schedule[ i ].status = urb->iso_frame_desc[ i ].status;
        }
    }

    {
        unsigned long flags;

        spin_lock_irqsave( &resource->lock, flags );

        usbback_driver_port_resource_handle_stimulus
          ( resource, usbback_driver_port_resource_stimulus_uc );

        spin_unlock_irqrestore( &resource->lock, flags );
    }

    usb_free_urb( urb );
}

static void usbback_driver_port_resource_unlink_urb
  ( struct usbback_driver_port_resource * resource )
{
    trace();

    usb_get_urb( resource->urb );

    int scheduled = usbback_work_schedule( &resource->unlink_urb_1_work );

    ASSERT( scheduled );
}

static void usbback_driver_port_resource_unlink_urb_1( void * data )
{
    trace();

    struct usbback_driver_port_resource * resource =
      (struct usbback_driver_port_resource *)data;

    usb_unlink_urb( resource->urb );

    usb_put_urb( resource->urb );

    {
        unsigned long flags;

        spin_lock_irqsave( &resource->lock, flags );

        usbback_driver_port_resource_handle_stimulus
          ( resource, usbback_driver_port_resource_stimulus_ul );

        spin_unlock_irqrestore( &resource->lock, flags );
    }
}

static void usbback_driver_port_resource_set_error
  ( struct usbback_driver_port_resource * resource )
{
    trace();

    resource->status = USBIF_RSP_ERROR;
}

static void usbback_driver_port_resource_complete
  ( struct usbback_driver_port_resource * resource )
{
    trace();

    resource->completing_device_request = resource->device_request;
    resource->device_request = NULL;

    int scheduled = usbback_work_schedule( &resource->complete_1_work );

    ASSERT( scheduled );
}

static void usbback_driver_port_resource_complete_1( void * data )
{
    trace();

    struct usbback_driver_port_resource * resource =
      (struct usbback_driver_port_resource *)data;

    if( resource->buffer )
    {
        resource->buffer = 0;

        usbback_mapper_unmap( &resource->mapper_request );
    }

    usbback_device_complete_request
    (
        resource->completing_device_request,
        resource->status,
        resource->actual_length
    );

    usbback_driver_port_resource_completed_io( resource->port, resource );
}
