/*****************************************************************************/
/* Service for mapping memory from remote domains                            */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on arch/xen/drivers/usbif/backend/main.c, original copyright notice */
/* follows...                                                                */
/*****************************************************************************/

/******************************************************************************
 * arch/xen/drivers/usbif/backend/main.c
 * 
 * Backend for the Xen virtual USB driver - provides an abstraction of a
 * USB host controller to the corresponding frontend driver.
 *
 * by Mark Williamson
 * Copyright (c) 2004 Intel Research Cambridge
 * Copyright (c) 2004, 2005 Mark Williamson
 *
 * Based on arch/xen/drivers/blkif/backend/main.c
 * Copyright (c) 2003-2004, Keir Fraser & Steve Hand
 */

#include <asm/pgtable.h>
#include <asm-xen/hypervisor.h>
#include <linux/errno.h>
#include "usbback_assert.h"
#include "usbback_callback.h"
#include "usbback_mapper.h"
#include "usbback_trace.h"

static inline unsigned long usbback_mapper_element_query_page_count
  ( struct usbback_mapper_element * element )
{
    return
    (
        ( element->address & ~PAGE_MASK ) /* Offset into 1st page.   */
        +
        element->length                   /* Length of request.      */
        +
        PAGE_SIZE - 1                     /* Round up to whole page. */
    )
    / PAGE_SIZE;
}

#define USBBACK_MAPPER_PAGES 256

static int usbback_mapper_in_use = 0;

static u32 usbback_mapper_page_in_use_bits
  [ ( USBBACK_MAPPER_PAGES + 31 ) / 32 ] = { 0, };

static inline void usbback_mapper_set_page_in_use( unsigned long page_index )
{
    unsigned long word_index = page_index / 32;
    int           bit_index  = page_index % 32;

    usbback_mapper_page_in_use_bits[ word_index ] |= ( (u32)1 ) << bit_index;
}

static inline void usbback_mapper_reset_page_in_use( unsigned long page_index )
{
    unsigned long word_index = page_index / 32;
    int           bit_index  = page_index % 32;

    usbback_mapper_page_in_use_bits[ word_index ] &=
      ~( ((u32)1) << bit_index );
}

static inline int usbback_mapper_test_page_in_use( unsigned long page_index )
{
    unsigned long word_index = page_index / 32;
    int           bit_index  = page_index % 32;

    return
    (
        (
            usbback_mapper_page_in_use_bits[ word_index ]
            &
            ( ( (u32)1 ) << bit_index )
        )
        !=
        0
    );
}

static multicall_entry_t usbback_mapper_multicall[ USBBACK_MAPPER_PAGES ];

static unsigned long usbback_mapper_vstart;

static DEFINE_SPINLOCK( usbback_mapper_lock );

static LIST_HEAD( usbback_mapper_list );

static USBBACK_CALLBACK_SERIALISER( usbback_mapper_callback_serialiser );

static int usbback_mapper_init_or_exit( int exit );

int usbback_mapper_init( void )
{
    trace();

    return usbback_mapper_init_or_exit( 0 );
}

void usbback_mapper_exit( void )
{
    trace();

    (void)usbback_mapper_init_or_exit( 1 );
}

static int usbback_mapper_init_or_exit( int exit )
{
    trace();

    int return_value = 0;

    if( exit )
    {
        goto EXIT;
    }

    usbback_mapper_vstart =
      allocate_empty_lowmem_region( USBBACK_MAPPER_PAGES );

    if( usbback_mapper_vstart == 0 )
    {
        return_value = -ENOMEM;

        goto EXIT_NO_LOWMEM;
    }

    return 0;

  EXIT:

    /* FIXME: free lowmem region */

  EXIT_NO_LOWMEM:

    return return_value;
}

static void usbback_mapper_kick_requests( void );

void usbback_mapper_map
(
    int                             element_count,
    struct usbback_mapper_element * element_array,
    struct usbback_mapper_request * request
)
{
    trace();

    request->element_count = element_count;
    request->element_array = element_array;

    unsigned long flags;

    spin_lock_irqsave( &usbback_mapper_lock, flags );

    list_add_tail
      ( usbback_mapper_request_to_link( request ), &usbback_mapper_list );

    spin_unlock_irqrestore( &usbback_mapper_lock, flags );

    usbback_mapper_kick_requests();
}

static void usbback_mapper_fast_flush_area
  ( unsigned long page_index, unsigned long page_count );

static void usbback_mapper_kick_requests( void )
{
    trace();

    unsigned long flags;

    spin_lock_irqsave( &usbback_mapper_lock, flags );

    while
    (
        ( !usbback_mapper_in_use )
        &&
        ( !list_empty( &usbback_mapper_list ) )
    )
    {
        int callback_value = 0;

        struct usbback_mapper_request * request =
          usbback_mapper_request_link_to( usbback_mapper_list.next );

        list_del_init( usbback_mapper_request_to_link( request ) );

        unsigned long page_count = 0;

        {
            int e;

            for( e = 0; e < request->element_count; e++ )
            {
                page_count += usbback_mapper_element_query_page_count
                  ( &request->element_array[ e ] );
            }
        }

        if( page_count == 0 )
        {
            goto ZERO_LENGTH;
        }

        if( page_count > USBBACK_MAPPER_PAGES )
        {
            printk
            (
                KERN_WARNING "usbback: request requires mapping %lu pages "
                "which exceeds supported maximum of %lu.\n",
                page_count,
                (unsigned long)USBBACK_MAPPER_PAGES
            );

            callback_value = -ENOMEM;

            goto UNABLE_TO_MAP;
        }

        int page_index = 0;

        {
            int success = 0;

            do
            {
                int i;

                for( i = 0; i < page_count; i++ )
                {
                    if( usbback_mapper_test_page_in_use( page_index + i ) )
                    {
                        page_index = page_index + i + 1;

                        break;
                    }
                }

                if( i == page_count )
                {
                    for( i = 0; i < page_count; i++ )
                    {
                        usbback_mapper_set_page_in_use( page_index + i );
                    }

                    success = 1;

                    break;
                }
            }
            while( page_index <= ( USBBACK_MAPPER_PAGES - page_count ) );

            if( !success )
            {
                list_add
                (
                    usbback_mapper_request_to_link( request ),
                    &usbback_mapper_list
                );

                break;
            }
        }

        usbback_mapper_in_use = 1;

        spin_unlock_irqrestore( &usbback_mapper_lock, flags );

        {
            unsigned long remap_prot =
              _PAGE_PRESENT | _PAGE_DIRTY | _PAGE_ACCESSED | _PAGE_RW;

            unsigned long element_page_index = page_index;

            int e;

            for( e = 0; e < request->element_count; e++ )
            {
                struct usbback_mapper_element * element =
                  &request->element_array[ e ];

                unsigned long element_page_count =
                  usbback_mapper_element_query_page_count( element );

                unsigned long i;
                unsigned long offset;

                for
                (
                    i = element_page_index, offset = 0;
                    i < ( element_page_index + element_page_count );
                    i++, offset += PAGE_SIZE
                )
                {
                    unsigned long vaddress =
                      offset + ( element_page_index * PAGE_SIZE ) +
                        usbback_mapper_vstart;
                    unsigned long maddress =
                      offset + ( element->address & PAGE_MASK );

                    usbback_mapper_multicall[ i ].op =
                      __HYPERVISOR_update_va_mapping_otherdomain;
                    usbback_mapper_multicall[ i ].args[ 0 ] = vaddress;
                    usbback_mapper_multicall[ i ].args[ 1 ] =
                      maddress | remap_prot;
                    usbback_mapper_multicall[ i ].args[ 2 ] = 0;
                    usbback_mapper_multicall[ i ].args[ 3 ] = element->domain;

                    phys_to_machine_mapping[ __pa( vaddress ) >> PAGE_SHIFT ] =
                      FOREIGN_FRAME( maddress >> PAGE_SHIFT );
                }

                element_page_index += element_page_count;
            }
        }

        if
        (
            unlikely
            (
                HYPERVISOR_multicall
                  ( &usbback_mapper_multicall[ page_index ], page_count )
                !=
                0
            )
        )
        {
            BUG();
        }

        {
            int i;

            for( i = page_index; i < ( page_index + page_count ); i++ )
            {
                if( unlikely( usbback_mapper_multicall[ i ].args[ 5 ] != 0 ) )
                {
                    printk( KERN_WARNING "usbback: unable to map buffer.\n" );

                    usbback_mapper_fast_flush_area( page_index, page_count );

                    callback_value = -1; /* FIXME: what value? */

                    spin_lock_irqsave( &usbback_mapper_lock, flags );

                    usbback_mapper_in_use = 0;

                    goto UNABLE_TO_MAP;
                }
            }
        }

        request->page_index = page_index;
        request->page_count = page_count;

        {
            int e;

            for( e = 0; e < request->element_count; e++ )
            {
                struct usbback_mapper_element * element =
                  &request->element_array[ e ];

                element->buffer =
                (
                    (char *)usbback_mapper_vstart
                    +
                    ( element->address & ~PAGE_MASK )
                    +
                    ( page_index * PAGE_SIZE )
                );

                page_index +=
                  usbback_mapper_element_query_page_count( element );
            }
        }

        usbback_callback_serialiser_complete_callback
        (
            &usbback_mapper_callback_serialiser,
            &request->callback,
            0
        );

        spin_lock_irqsave( &usbback_mapper_lock, flags );

        usbback_mapper_in_use = 0;

        continue;

      ZERO_LENGTH:
      UNABLE_TO_MAP:

        request->page_index = 0;
        request->page_count = 0;

        {
            int e;

            for( e = 0; e < request->element_count; e++ )
            {
                request->element_array[ e ].buffer = NULL;
            }
        }

        usbback_callback_serialiser_complete_callback
        (
            &usbback_mapper_callback_serialiser,
            &request->callback,
            callback_value
        );
    }

    spin_unlock_irqrestore( &usbback_mapper_lock, flags );
}

void usbback_mapper_abort_map( struct usbback_mapper_request * request )
{
    unsigned long flags;

    spin_lock_irqsave( &usbback_mapper_lock, flags );

    if( !list_empty( usbback_mapper_request_to_link( request ) ) )
    {
        list_del_init( usbback_mapper_request_to_link( request ) );

        request->page_index = 0;
        request->page_count = 0;

        {
            int e;

            for( e = 0; e < request->element_count; e++ )
            {
                request->element_array[ e ].buffer = NULL;
            }
        }

        usbback_callback_serialiser_complete_callback
        (
            &usbback_mapper_callback_serialiser,
            &request->callback,
            -ECANCELED
        );
    }

    spin_unlock_irqrestore( &usbback_mapper_lock, flags );
}

static void usbback_mapper_fast_flush_area
  ( unsigned long page_index, unsigned long page_count )
{
    ASSERT( page_count > 0 );

    int i;

    for( i = page_index; i < ( page_index + page_count ); i++ )
    {
        usbback_mapper_multicall[ i ].op = __HYPERVISOR_update_va_mapping;
        usbback_mapper_multicall[ i ].args[ 0 ] =
          usbback_mapper_vstart + ( i * PAGE_SIZE );
        usbback_mapper_multicall[ i ].args[ 1 ] = 0;
        usbback_mapper_multicall[ i ].args[ 2 ] = 0;
    }

    usbback_mapper_multicall[ i - 1 ].args[ 2 ] = UVMF_FLUSH_TLB;

    if
    (
        unlikely
        (
            HYPERVISOR_multicall
              ( &usbback_mapper_multicall[ page_index ], page_count )
            !=
            0
        )
    )
    {
        BUG();
    }
}

void usbback_mapper_unmap( struct usbback_mapper_request * request )
{
    trace();

    int page_index = request->page_index;
    int page_count = request->page_count;

    if( page_count != 0 )
    {
        usbback_mapper_fast_flush_area( page_index, page_count );

        unsigned long flags;

        spin_lock_irqsave( &usbback_mapper_lock, flags );

        {
            int i;

            for( i = page_index; i < ( page_index + page_count ); i++ )
            {
                usbback_mapper_reset_page_in_use( i );
            }
        }

        spin_unlock_irqrestore( &usbback_mapper_lock, flags );

        usbback_mapper_kick_requests();
    }
}
