/*****************************************************************************/
/* Service for mapping memory from remote domains                            */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

#ifndef USBBACK_MAPPER_H
#define USBBACK_MAPPER_H

#include <linux/list.h>
#include <linux/list.h>
#include <asm-xen/xen-public/io/usbif.h>

#include "usbback_callback.h"

int usbback_mapper_init( void );

void usbback_mapper_exit( void );

/* A single request to the mapper may consist of multiple extents. Each      */
/* extent is represented by an element.  The mapper will wait until there is */
/* enough memory to map all of the extents for a request before performing   */
/* the request.                                                              */

struct usbback_mapper_element
{
    domid_t         domain;
    unsigned long   address;
    unsigned long   length;
    void          * buffer;
};

/* The client needs one of these resources to make a request. */

struct usbback_mapper_request
{
    struct usbback_callback         callback;
    int                             element_count;
    struct usbback_mapper_element * element_array;
    unsigned long                   page_index;
    unsigned long                   page_count;
};

/* The client needs to access the callback in the request to initialise it. */

static inline struct usbback_callback * usbback_mapper_request_to_callback
  ( struct usbback_mapper_request * request )
{
    return &request->callback;
}

/* The client needs to be able to cast the completed callback back to a      */
/* request (and then back to the client's context).                          */

static inline struct usbback_mapper_request *
  usbback_mapper_request_callback_to( struct usbback_callback * callback )
{
    return container_of( callback, struct usbback_mapper_request, callback );
}

/* The client may use the link in the request until the request is submitted */
/* and again after the request is completed.                                 */

static inline struct list_head * usbback_mapper_request_to_link
  ( struct usbback_mapper_request * request )
{
    return usbback_callback_to_link( &request->callback );
}

static inline struct usbback_mapper_request * usbback_mapper_request_link_to
  ( struct list_head * link )
{
    return usbback_mapper_request_callback_to
      ( usbback_callback_link_to( link ) );
}

/* Client must initialise callback in request and domain, address and length */
/* fields of all elements in the array before calling usbback_mapper_map.    */

void usbback_mapper_map
(
    int                             element_count,
    struct usbback_mapper_element * element_array,
    struct usbback_mapper_request * request
);

/* Callback will still happen after abort and may still be successful but    */
/* abort guarantees that callback will happen independent of any other       */
/* clients by allowing the request to fail.                                  */
/* Abort must only be called after submission returns and before request is  */
/* resubmitted after callback.  Abort may be called concurrently with        */
/* callback.                                                                 */

void usbback_mapper_abort_map( struct usbback_mapper_request * request );

/* When a request is successful, the client can use the mapping by accessing */
/* the buffer fields of the mapping elements.  Once the mapping is finished  */
/* with, the client must call unmap to free resources.                       */

void usbback_mapper_unmap( struct usbback_mapper_request * request );

#endif
