/*****************************************************************************/
/* Enhanced work queue service                                               */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/* This work queue service is a lot like the linux work queue service except */
/* that it provides the interface usbback_work_until which allows the client */
/* to wait for a condition to be met even when the client is executing on a  */
/* work thread and the condition will only be satisfied by execution of      */
/* another work item.                                                        */

#ifndef USBBACK_WORK_H
#define USBBACK_WORK_H

#include <linux/list.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/wait.h>
#include <linux/workqueue.h>

struct usbback_work
{
    struct list_head link;
    void ( * function )( void * context );
    void   * context;
};

#define USBBACK_WORK_INIT( name, fn, ctx ) \
{ LIST_HEAD_INIT( name.link ), fn, ctx }

#define USBBACK_WORK( name, fn, ctx ) \
struct usbback_work name = USBBACK_WORK_INIT( name, fn, ctx )

static inline void usbback_work_init
(
    struct usbback_work   * work,
    void                ( * function )( void * context ),
    void                  * context
)
{
    INIT_LIST_HEAD( &work->link );

    work->function = function;
    work->context  = context;
}

static inline struct list_head * usbback_work_to_link
  ( struct usbback_work * work )
{
    return &work->link;
}

static inline struct usbback_work * usbback_work_link_to
  ( struct list_head * link )
{
    return container_of( link, struct usbback_work, link );
}

/* Returns 1 if scheduled this time or 0 if already scheduled. */

int usbback_work_schedule( struct usbback_work * work );

static inline void usbback_work_perform_synchronously
  ( struct usbback_work * work )
{
    work->function( work->context );
}

/* Don't use any of these, they're just exposed for the macro below. */

extern spinlock_t usbback_work_list_lock;

extern struct list_head usbback_work_list;

extern wait_queue_head_t usbback_work_waitqueue;

extern struct list_head usbback_work_condition;

/* Wait for a condition to be met. This works whether or not you call it     */
/* from a work item and works even when the condition will only be satisfied */
/* by another work item.                                                     */

#define usbback_work_until( condition )                                 \
do                                                                      \
{                                                                       \
    unsigned long flags;                                                \
                                                                        \
    spin_lock_irqsave( &usbback_work_list_lock, flags );                \
                                                                        \
    for( ; ; )                                                          \
    {                                                                   \
        while                                                           \
        (                                                               \
            ( !list_empty( &usbback_work_list ) )                       \
            &&                                                          \
            ( !( condition ) )                                          \
        )                                                               \
        {                                                               \
            struct usbback_work * work = list_entry                     \
            (                                                           \
                usbback_work_list.next,                                 \
                struct usbback_work,                                    \
                link                                                    \
            );                                                          \
                                                                        \
            list_del_init( &work->link );                               \
                                                                        \
            spin_unlock_irqrestore( &usbback_work_list_lock, flags );   \
                                                                        \
            usbback_work_perform_synchronously( work );                 \
                                                                        \
            spin_lock_irqsave( &usbback_work_list_lock, flags );        \
        }                                                               \
                                                                        \
        if( condition )                                                 \
        {                                                               \
            break;                                                      \
        }                                                               \
                                                                        \
        struct list_head link;                                          \
                                                                        \
        INIT_LIST_HEAD( &link );                                        \
                                                                        \
        list_add_tail( &link, &usbback_work_condition );                \
                                                                        \
        spin_unlock_irqrestore( &usbback_work_list_lock, flags );       \
                                                                        \
        wait_event( usbback_work_waitqueue, list_empty( &link ) );      \
                                                                        \
        spin_lock_irqsave( &usbback_work_list_lock, flags );            \
    }                                                                   \
                                                                        \
    spin_unlock_irqrestore( &usbback_work_list_lock, flags );           \
}                                                                       \
while( 0 )

/* When you satisfy a condition, you should call this to kick any threads    */
/* waiting on the condition.                                                 */

void usbback_work_wake_up( void );

#endif
