/*****************************************************************************/
/* usbfront_device is a device which represents a connection to a back-end.  */
/* The intent was for it to have an interface like a hardware USB host       */
/* controller device.                                                        */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on arch/xen/drivers/usbif/frontend/main.c, original copyright       */
/* notice follows...                                                         */
/*****************************************************************************/

/*
 * Xen Virtual USB Frontend Driver 
 *
 * This file contains the first version of the Xen virtual USB hub
 * that I've managed not to delete by mistake (3rd time lucky!).
 *
 * Based on Linux's uhci.c, original copyright notices are displayed
 * below.  Portions also (c) 2004 Intel Research Cambridge
 * and (c) 2004, 2005 Mark Williamson
 *
 * Contact <mark.williamson@cl.cam.ac.uk> or
 * <xen-devel@lists.sourceforge.net> regarding this code.
 *
 * Still to be (maybe) implemented:
 * - migration / backend restart support?
 * - support for building / using as a module
 */

/*
 * Universal Host Controller Interface driver for USB.
 *
 * Maintainer: Johannes Erdfelt <johannes@erdfelt.com>
 *
 * (C) Copyright 1999 Linus Torvalds
 * (C) Copyright 1999-2002 Johannes Erdfelt, johannes@erdfelt.com
 * (C) Copyright 1999 Randy Dunlap
 * (C) Copyright 1999 Georg Acher, acher@in.tum.de
 * (C) Copyright 1999 Deti Fliegl, deti@fliegl.de
 * (C) Copyright 1999 Thomas Sailer, sailer@ife.ee.ethz.ch
 * (C) Copyright 1999 Roman Weissgaerber, weissg@vienna.at
 * (C) Copyright 2000 Yggdrasil Computing, Inc. (port of new PCI interface
 *               support from usb-ohci.c by Adam Richter, adam@yggdrasil.com).
 * (C) Copyright 1999 Gregory P. Smith (from usb-ohci.c)
 *
 * Intel documents this fairly well, and as far as I know there
 * are no royalties or anything like that, but even so there are
 * people who decided that they want to do the same thing in a
 * completely different way.
 *
 * WARNING! The USB documentation is downright evil. Most of it
 * is just crap, written by a committee. You're better off ignoring
 * most of it, the important stuff is:
 *  - the low-level protocol (fairly simple but lots of small details)
 *  - working around the horridness of the rest
 */

#include <linux/device.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <asm-xen/ctrl_if.h>
#include <asm-xen/evtchn.h>
#include "usbfront_assert.h"
#include "usbfront_bus.h"
#include "usbfront_device.h"
#include "usbfront_trace.h"

#define USBFRONT_DEVICE_PORT_COUNT 7

static DECLARE_RWSEM( usbfront_device_list_sem );

typedef enum
{
    usbfront_device_state_i,
    usbfront_device_state_i_dn,
    usbfront_device_state_i_sd,
    usbfront_device_state_i_dn_dt,
    usbfront_device_state_i_dn_cn,
    usbfront_device_state_i_dn_sd,
    usbfront_device_state_i_dn_c1,
    usbfront_device_state_i_sd_sd,
    usbfront_device_state_i_dn_dt_c1,
    usbfront_device_state_i_dn_cn_dn,
    usbfront_device_state_i_dn_cn_c1,
    usbfront_device_state_i_sd_sd_ud,
    usbfront_device_state_i_dn_dt_c1_sd,
    usbfront_device_state_i_dn_dt_c1_ud,
    usbfront_device_state_i_dn_cn_c1_dn,
    usbfront_device_state_i_dn_cn_c1_sd,
    usbfront_device_state_i_dn_cn_c1_c2,
    usbfront_device_state_i_dn_cn_c1_dn_dt,
    usbfront_device_state_i_dn_cn_c1_c2_dn,
    usbfront_device_state_i_dn_cn_c1_c2_sd,
    usbfront_device_state_i_dn_cn_c1_c2_pc,
    usbfront_device_state_i_dn_cn_c1_c2_dn_dt,
    usbfront_device_state_i_dn_cn_c1_c2_dn_sd,
    usbfront_device_state_i_dn_cn_c1_c2_sd_pc,
    usbfront_device_state_i_dn_cn_c1_c2_sd_sd
}
usbfront_device_state;

typedef enum
{
    usbfront_device_stimulus_dt, /* interface status changed destroyed    */
    usbfront_device_stimulus_dn, /* interface status changed disconnected */
    usbfront_device_stimulus_cn, /* interface status changed connected    */
    usbfront_device_stimulus_sd, /* shutdown                              */
    usbfront_device_stimulus_pt, /* polling tick                          */
    usbfront_device_stimulus_rq, /* request queued                        */
    usbfront_device_stimulus_c1, /* connect part one success              */
    usbfront_device_stimulus_c2, /* connect part two success              */
    usbfront_device_stimulus_ud, /* unregister done                       */
    usbfront_device_stimulus_rr, /* requests running                      */
    usbfront_device_stimulus_ri, /* requests idle                         */
    usbfront_device_stimulus_pc, /* port probes completed                 */
    usbfront_device_stimulus_invalid
}
usbfront_device_stimulus;

struct usbfront_device
{
    struct device                       dev;
    struct list_head                    link;
    u32                                 handle;
    spinlock_t                          lock;
    usbfront_device_state               state;
    struct work_struct                  connect_part_one_1_work;
    struct work_struct                  connect_part_two_1_work;
    struct work_struct                  unregister_1_work;
    usbif_sring_t                     * shared_ring;
    usbif_front_ring_t                  front_ring;
    int                                 evtchn;
    int                                 irq;
    int                                 got_irq;
    int                                 tick_count;
    struct usb_port_status              port_status
      [ USBFRONT_DEVICE_PORT_COUNT ];
    unsigned long                       next_id;
    struct list_head                    request_list;
    struct list_head                    running_list;
    struct list_head                    complete_list;
    usbif_fe_interface_status_changed_t status_changed;
    int                                 port_probe_count;
    struct usbfront_device_request      port_probe_request
      [ USBFRONT_DEVICE_PORT_COUNT ];
    wait_queue_head_t                   shutdown_wait_queue_head;
    int                                 shutdown;
    irqreturn_t                     ( * driver_irq_handler )
      ( int irq, void * context, struct pt_regs * regs );
    void                              * driver_irq_context;
};

struct usbfront_device * usbfront_device_dev_to( struct device * dev )
{
    return container_of( dev, struct usbfront_device, dev );
}

static void usbfront_device_handle_stimulus
(
    struct usbfront_device   * device,
    usbfront_device_stimulus   stimulus,
    unsigned long            * flags
);

static void usbfront_device_release( struct device * dev );

static void usbfront_device_connect_part_one_1( void * data );

static void usbfront_device_connect_part_two_1( void * data );

static void usbfront_device_unregister_1( void * data );

static int usbfront_device_init_or_exit
  ( struct usbfront_device * device, u32 handle, int exit )
{
    trace();

    int return_value = 0;

    if( exit )
    {
        goto EXIT;
    }

    memset( device, 0, sizeof( *device ) );

    device->handle = handle;

    INIT_LIST_HEAD( &device->link );

    device->dev.bus = &usbfront_bus_type;
    device->dev.parent = &usbfront_bus;

    snprintf
      ( device->dev.bus_id, BUS_ID_SIZE, "%08lx", (unsigned long)handle );

    device->dev.release = usbfront_device_release;

    spin_lock_init( &device->lock );

    device->state = usbfront_device_state_i;

    INIT_WORK
    (
        &device->connect_part_one_1_work,
        usbfront_device_connect_part_one_1,
        device
    );

    INIT_WORK
    (
        &device->connect_part_two_1_work,
        usbfront_device_connect_part_two_1,
        device
    );

    INIT_WORK
    (
        &device->unregister_1_work,
        usbfront_device_unregister_1,
        device
    );

    device->shared_ring = (usbif_sring_t *)__get_free_page( GFP_KERNEL );

    if( device->shared_ring == NULL )
    {
        return_value = -ENOMEM;

        goto EXIT_NO_RING;
    }

    INIT_LIST_HEAD( &device->request_list );
    INIT_LIST_HEAD( &device->running_list );
    INIT_LIST_HEAD( &device->complete_list );

    init_waitqueue_head( &device->shutdown_wait_queue_head );

    if( ( return_value = device_register( &device->dev ) ) != 0 )
    {
        goto EXIT_NO_DEVICE_REGISTER;
    }

    return 0;

  EXIT:

  EXIT_NO_DEVICE_REGISTER:

    free_page( (unsigned long)device->shared_ring );

  EXIT_NO_RING:

    return return_value;
}

static int usbfront_device_init( struct usbfront_device * device, u32 handle  )
{
    trace();

    return usbfront_device_init_or_exit( device, handle, 0 );
}

static void usbfront_device_exit( struct usbfront_device * device )
{
    trace();

    (void)usbfront_device_init_or_exit( device, 0, 1 );
}

void usbfront_device_get( struct usbfront_device * device )
{
    trace();

    get_device( &device->dev );
}

void usbfront_device_put( struct usbfront_device * device )
{
    trace();

    put_device( &device->dev );
}

static void usbfront_device_release( struct device * dev )
{
    trace();

    struct usbfront_device * device = usbfront_device_dev_to( dev );

    (void)usbfront_device_exit( device );

    kfree( device );
}

static int usbfront_device_shutdown( struct device * dev, void * ignored );

struct usbfront_device * usbfront_device_find_or_allocate
  ( u32 handle, int shutting_down )
{
    trace();

    static LIST_HEAD( usbfront_device_list );

    struct usbfront_device * device = NULL;

    /* Try to reference an already created device. */

    down_write( &usbfront_device_list_sem );

    {
        struct usbfront_device * cur = NULL;

        list_for_each_entry( cur, &usbfront_device_list, link )
        {
            if( handle == cur->handle )
            {
                device = cur;

                usbfront_device_get( device );

                break;
            }
        }
    }

    if( device == NULL )
    {
        /* No existing device so try to create one. */

        device = (struct usbfront_device *)kmalloc
          ( sizeof( struct usbfront_device ), GFP_KERNEL );

        if( device != NULL )
        {
            /* Managed to allocate memory. */

            if( usbfront_device_init( device, handle ) == 0 )
            {
                usbfront_device_get( device );

                list_add_tail( &device->link, &usbfront_device_list );

                if( shutting_down )
                {
                    usbfront_device_shutdown( &device->dev, NULL );
                }
            }
            else
            {
                kfree( device );

                device = NULL;
            }
        }
    }

    up_write( &usbfront_device_list_sem );

    return device;
}

void usbfront_device_handle_interface_status_changed
(
    struct usbfront_device * device,
    ctrl_msg_t             * msg,
    unsigned long            id
)
{
    trace();

    usbif_fe_interface_status_changed_t * status_changed =
      (usbif_fe_interface_status_changed_t *)msg->msg;

    usbfront_device_stimulus stimulus = usbfront_device_stimulus_invalid;

    switch( status_changed->status )
    {
      case USBIF_INTERFACE_STATUS_DESTROYED:
        stimulus = usbfront_device_stimulus_dt;
        break;
      case USBIF_INTERFACE_STATUS_DISCONNECTED:
        stimulus = usbfront_device_stimulus_dn;
        break;
      case USBIF_INTERFACE_STATUS_CONNECTED:
        stimulus = usbfront_device_stimulus_cn;
        break;
      default:
        break;
    }

    if( stimulus != usbfront_device_stimulus_invalid )
    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        device->status_changed = *status_changed;

        usbfront_device_handle_stimulus( device, stimulus, &flags );

        spin_unlock_irqrestore( &device->lock, flags );
    }

    ctrl_if_send_response( msg );
}

struct message_info
{
    size_t   length;
    void ( * handler )
    (
        struct usbfront_device * device,
        ctrl_msg_t             * msg,
        unsigned long            id
    );
};

static struct message_info message_subtype_info[] =
{
    [ CMSG_USBIF_FE_INTERFACE_STATUS_CHANGED ] =
    {
        sizeof( usbif_fe_interface_status_changed_t ),
        usbfront_device_handle_interface_status_changed
    }
};

void usbfront_device_handle_control_message
(
    struct usbfront_device * device,
    ctrl_msg_t             * msg,
    unsigned long            id
)
{
    trace();

    /* Message wouldn't have made it to us unless it contained a header with */
    /* our handle in it.                                                     */

    ASSERT( msg->length >= sizeof( usbif_fe_header_t ) );

    if
    (
        (
            msg->subtype
            <
            ( sizeof( message_subtype_info ) / sizeof( struct message_info ) )
        )
        &&
        ( msg->length == message_subtype_info[ msg->subtype ].length )
    )
    {
        message_subtype_info[ msg->subtype ].handler( device, msg, id );
    }
    else
    {
        msg->length = 0;

        ctrl_if_send_response( msg );
    }
}

static int usbfront_device_shutdown( struct device * dev, void * ignored )
{
    trace();

    struct usbfront_device * device = usbfront_device_dev_to( dev );

    device->shutdown = 0;

    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        usbfront_device_handle_stimulus
          ( device, usbfront_device_stimulus_sd, &flags );

        spin_unlock_irqrestore( &device->lock, flags );
    }

    wait_event( device->shutdown_wait_queue_head, device->shutdown );

    return 0;
}

static void usbfront_device_shutdown_all( void )
{
    trace();

    (void)bus_for_each_dev
      ( &usbfront_bus_type, NULL, NULL, usbfront_device_shutdown );
}

void usbfront_device_shutdown_all_phase_1( void )
{
    trace();

    usbfront_device_shutdown_all();
}

void usbfront_device_shutdown_all_phase_2( void )
{
    trace();

    usbfront_device_shutdown_all();
}

int usbfront_device_query_port_count( struct usbfront_device * device )
{
    /* trace(); */

    return USBFRONT_DEVICE_PORT_COUNT;
}

int usbfront_device_query_port_status_changed
  ( struct usbfront_device * device, int port_index )
{
    /* trace(); */

    ASSERT( port_index < USBFRONT_DEVICE_PORT_COUNT );

    int changed;

    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        if( ++device->tick_count == USBFRONT_DEVICE_PORT_COUNT )
        {
            device->tick_count = 0;

            usbfront_device_handle_stimulus
              ( device, usbfront_device_stimulus_pt, &flags );
        }

        changed = ( device->port_status[ port_index ].wPortChange != 0 );

        spin_unlock_irqrestore( &device->lock, flags );
    }

    return changed;
}

struct usb_port_status usbfront_device_query_port_status
  ( struct usbfront_device * device, int port_index )
{
    trace();

    ASSERT( port_index < USBFRONT_DEVICE_PORT_COUNT );

    struct usb_port_status port_status;

    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        port_status = device->port_status[ port_index ];

        spin_unlock_irqrestore( &device->lock, flags );
    }

    return port_status;
}

void usbfront_device_set_port_power
  ( struct usbfront_device * device, int port_index )
{
    trace();

    ASSERT( port_index < USBFRONT_DEVICE_PORT_COUNT );

    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        device->port_status[ port_index ].wPortStatus |= USB_PORT_STAT_POWER;

        spin_unlock_irqrestore( &device->lock, flags );
    }
}

void usbfront_device_set_port_reset
  ( struct usbfront_device * device, int port_index )
{
    trace();

    ASSERT( port_index < USBFRONT_DEVICE_PORT_COUNT );

    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        device->port_status[ port_index ].wPortStatus |= USB_PORT_STAT_RESET;
        device->port_status[ port_index ].wPortStatus &= ~USB_PORT_STAT_ENABLE;

        usbfront_device_handle_stimulus
          ( device, usbfront_device_stimulus_pt, &flags );

        spin_unlock_irqrestore( &device->lock, flags );
    }
}

void usbfront_device_clear_port_enable
  ( struct usbfront_device * device, int port_index )
{
    trace();

    ASSERT( port_index < USBFRONT_DEVICE_PORT_COUNT );

    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        device->port_status[ port_index ].wPortStatus &= ~USB_PORT_STAT_ENABLE;

        spin_unlock_irqrestore( &device->lock, flags );
    }
}

void usbfront_device_clear_port_connection_change
  ( struct usbfront_device * device, int port_index )
{
    trace();

    ASSERT( port_index < USBFRONT_DEVICE_PORT_COUNT );

    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        device->port_status[ port_index ].wPortChange &=
          ~USB_PORT_STAT_C_CONNECTION;

        spin_unlock_irqrestore( &device->lock, flags );
    }
}

void usbfront_device_clear_port_reset_change
  ( struct usbfront_device * device, int port_index )
{
    trace();

    ASSERT( port_index < USBFRONT_DEVICE_PORT_COUNT );

    {
        unsigned long flags;

        spin_lock_irqsave( &device->lock, flags );

        device->port_status[ port_index ].wPortChange &=
          ~USB_PORT_STAT_C_RESET;

        spin_unlock_irqrestore( &device->lock, flags );
    }
}

void usbfront_device_install_irq_handler
(
    struct usbfront_device * device,
    irqreturn_t ( * handler )
      ( int irq, void * context, struct pt_regs * regs ),
    void * context
)
{
    trace();

    device->driver_irq_handler = handler;
    device->driver_irq_context = context;
}

static void usbfront_device_queue_internal_request
(
    struct usbfront_device         * device,
    struct usbfront_device_request * request,
    void                         ( * internal_completion )
    (
        struct usbfront_device         * device,
        struct usbfront_device_request * request
    )
)
{
    /* trace(); */

    INIT_LIST_HEAD( &request->link );

    request->internal_completion = internal_completion;

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    list_add_tail( &request->link, &device->request_list );

    usbfront_device_handle_stimulus
      ( device, usbfront_device_stimulus_rq, &flags );

    spin_unlock_irqrestore( &device->lock, flags );
}

void usbfront_device_queue_request
  ( struct usbfront_device * device, struct usbfront_device_request * request )
{
    trace();

    usbfront_device_queue_internal_request( device, request, NULL );
}

struct usbfront_device_request * usbfront_device_collect_completed_request
  ( struct usbfront_device * device )
{
    trace();

    struct usbfront_device_request * request = NULL;

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    if( !list_empty( &device->complete_list ) )
    {
        request = list_entry
          ( device->complete_list.next, struct usbfront_device_request, link );

        list_del_init( &request->link );
    }

    spin_unlock_irqrestore( &device->lock, flags );

    return request;
}

void usbfront_device_remove_irq_handler( struct usbfront_device * device )
{
    trace();

    device->driver_irq_handler = NULL;
    device->driver_irq_context = NULL;
}

static void usbfront_device_invalid_stimulus
  ( struct usbfront_device * device, usbfront_device_stimulus stimulus );

static void usbfront_device_reject_message
  ( struct usbfront_device * device, usbfront_device_stimulus stimulus );

static void usbfront_device_connect_part_one
  ( struct usbfront_device * device );

static void usbfront_device_connect_part_two
  ( struct usbfront_device * device );

static void usbfront_device_disconnect( struct usbfront_device * device );

static void usbfront_device_unregister( struct usbfront_device * device );

static void usbfront_device_probe_all_ports
  ( struct usbfront_device * device, unsigned long * flags );

static void usbfront_device_disconnect_all_ports
  ( struct usbfront_device * device );

static void usbfront_device_kick_requests
  ( struct usbfront_device * device, unsigned long * flags );

static void usbfront_device_purge_requests
  ( struct usbfront_device * device, unsigned long * flags );

static void usbfront_device_complete_shutdown
  ( struct usbfront_device * device );

static void usbfront_device_handle_stimulus
(
    struct usbfront_device   * device,
    usbfront_device_stimulus   stimulus,
    unsigned long            * flags
)
{
#if 0
    trace3
    (
        "device %p in state %d received stimulus %d",
        device,
        device->state,
        stimulus
    );
#endif

    switch( device->state )
    {
      case usbfront_device_state_i:
        /* Registered            */
        /* Maybe requests queued */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_dn:
            device->state = usbfront_device_state_i_dn;
            usbfront_device_connect_part_one( device );
            break;
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_sd;
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn:
        /* Registered               */
        /* Connectable              */
        /* Maybe requests queued    */
        /* Connect part one running */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            device->state = usbfront_device_state_i_dn_dt;
            break;
          case usbfront_device_stimulus_dn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_cn:
            device->state = usbfront_device_state_i_dn_cn;
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_sd;
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          case usbfront_device_stimulus_c1:
            device->state = usbfront_device_state_i_dn_c1;
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_sd:
        /* Shutdown 1 completed   */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt: /* FALLS THROUGH */
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn: /* FALLS THROUGH */
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_sd_sd;
            usbfront_device_unregister( device );
            usbfront_device_purge_requests( device, flags );
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_dt:
        /* Registered               */
        /* Destroyed                */
        /* Connect part one running */
        /* Maybe requests queued    */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt: /* FALLS THROUGH */
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_sd;
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          case usbfront_device_stimulus_c1:
            device->state = usbfront_device_state_i_dn_dt_c1;
            usbfront_device_unregister( device );
            usbfront_device_purge_requests( device, flags );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn:
        /* Registered               */
        /* Connected                */
        /* Maybe requests queued    */
        /* Connect part one running */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_dn:
            device->state = usbfront_device_state_i_dn_cn_dn;
            break;
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_sd;
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          case usbfront_device_stimulus_c1:
            device->state = usbfront_device_state_i_dn_cn_c1;
            usbfront_device_connect_part_two( device );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_sd:
        /* Registered               */
        /* Maybe requests queued    */
        /* Connect part one running */
        /* Shutdown 1 outstanding   */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt: /* FALLS THROUGH */
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn: /* FALLS THROUGH */
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          case usbfront_device_stimulus_c1:
            device->state = usbfront_device_state_i_sd;
            usbfront_device_complete_shutdown( device );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_c1:
        /* Registered               */
        /* Connectable              */
        /* Maybe requests queued    */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            device->state = usbfront_device_state_i_dn_dt_c1;
            usbfront_device_unregister( device );
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_dn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_cn:
            device->state = usbfront_device_state_i_dn_cn_c1;
            usbfront_device_connect_part_two( device );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_sd;
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_sd_sd:
        /* Shutdown 2 completed   */
        /* Unregister outstanding */
        switch( stimulus )
        {
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_ud:
            device->state = usbfront_device_state_i_sd_sd_ud;
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_dt_c1:
        /* Unregister outstanding   */
        /* Destroyed                */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt: /* FALLS THROUGH */
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_dt_c1_sd;
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_ud:
            device->state = usbfront_device_state_i_dn_dt_c1_ud;
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_dn:
        /* Registered                      */
        /* Disconnected                    */
        /* Maybe requests queued           */
        /* Connect part one running        */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            device->state = usbfront_device_state_i_dn_dt;
            break;
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_sd;
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          case usbfront_device_stimulus_c1:
            device->state = usbfront_device_state_i_dn;
            usbfront_device_connect_part_one( device );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1:
        /* Registered                   */
        /* Connected                    */
        /* Maybe requests queued        */
        /* Connect part two outstanding */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_dn:
            device->state = usbfront_device_state_i_dn_cn_c1_dn;
            break;
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_cn_c1_sd;
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          case usbfront_device_stimulus_c2:
            device->state = usbfront_device_state_i_dn_cn_c1_c2;
            usbfront_device_probe_all_ports( device, flags );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_sd_sd_ud:
        /* Shutdown 2 completed */
        switch( stimulus )
        {
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_purge_requests( device, flags );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_dt_c1_sd:
        /* Unregister outstanding   */
        /* Destroyed                */
        /* Shutdown 1 completed     */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt: /* FALLS THROUGH */
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_sd_sd;
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_ud:
            device->state = usbfront_device_state_i_dn_dt_c1_ud;
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_dt_c1_ud:
        /* Unregisted */
        /* Destroyed  */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt: /* FALLS THROUGH */
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_purge_requests( device, flags );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_dn:
        /* Registered                   */
        /* Disconnected                 */
        /* Maybe requests queued        */
        /* Connect part two outstanding */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            device->state = usbfront_device_state_i_dn_cn_c1_dn_dt;
            break;
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_cn_c1_sd;
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          case usbfront_device_stimulus_c2:
            device->state = usbfront_device_state_i_dn;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_connect_part_one( device );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_sd:
        /* Registered                   */
        /* Connected                    */
        /* Maybe requests queued        */
        /* Connect part two outstanding */
        /* Shutdown 1 outstanding       */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt: /* FALLS THROUGH */
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn: /* FALLS THROUGH */
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          case usbfront_device_stimulus_c2:
            device->state = usbfront_device_state_i_sd;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_complete_shutdown( device );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_c2:
        /* Registered                   */
        /* Connected                    */
        /* Maybe requests queued        */
        /* Probe running                */
        /* Maybe requests running       */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_dn:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_dn;
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_sd;
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_kick_requests( device, flags );
            break;
          case usbfront_device_stimulus_rr:
          case usbfront_device_stimulus_ri:
            break;
          case usbfront_device_stimulus_pc:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_pc;
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_dn_dt:
        /* Registered                   */
        /* Destroyed                    */
        /* Maybe requests queued        */
        /* Connect part two outstanding */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt: /* FALLS THROUGH */
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_cn_c1_sd;
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq:
            break;
          case usbfront_device_stimulus_c2:
            device->state = usbfront_device_state_i_dn_dt_c1;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_unregister( device );
            usbfront_device_purge_requests( device, flags );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_c2_dn:
        /* Registered                   */
        /* Disconnected                 */
        /* Probe running                */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_dn_dt;
            break;
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_dn_sd;
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_pc:
            device->state = usbfront_device_state_i_dn;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_connect_part_one( device );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_c2_sd:
        /* Registered                   */
        /* Connected                    */
        /* Maybe requests queued        */
        /* Probe running                */
        /* Maybe requests running       */
        /* Shutdown 1 completed         */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_dn:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_dn_sd;
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_sd_sd;
            usbfront_device_purge_requests( device, flags );
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_kick_requests( device, flags );
            break;
          case usbfront_device_stimulus_rr:
          case usbfront_device_stimulus_ri:
            break;
          case usbfront_device_stimulus_pc:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_sd_pc;
            usbfront_device_kick_requests( device, flags );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_c2_pc:
        /* Registered                   */
        /* Connected                    */
        /* Maybe requests queued        */
        /* Maybe requests running       */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_dn:
            device->state = usbfront_device_state_i_dn;
            usbfront_device_purge_requests( device, flags );
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_connect_part_one( device );
            break;
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_sd_pc;
            usbfront_device_complete_shutdown( device );
            usbfront_device_kick_requests( device, flags );
            break;
          case usbfront_device_stimulus_pt:
            device->state = usbfront_device_state_i_dn_cn_c1_c2;
            usbfront_device_probe_all_ports( device, flags );
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_kick_requests( device, flags );
            break;
          case usbfront_device_stimulus_rr:
          case usbfront_device_stimulus_ri:
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_c2_dn_dt:
        /* Registered                   */
        /* Destroyed                    */
        /* Probe running                */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt: /* FALLS THROUGH */
          case usbfront_device_stimulus_dn: /* FALLS THROUGH */
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_dn_sd;
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_pc:
            device->state = usbfront_device_state_i_dn_dt_c1;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_unregister( device );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_c2_dn_sd:
        /* Registered                   */
        /* Disconnected                 */
        /* Probe running                */
        /* Shutdown 1 completed         */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
          case usbfront_device_stimulus_dn:
          case usbfront_device_stimulus_cn:
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_dn_cn_c1_c2_sd_sd;
            usbfront_device_complete_shutdown( device );
            break;
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_pc:
            device->state = usbfront_device_state_i_sd;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_c2_sd_pc:
        /* Registered                   */
        /* Connected                    */
        /* Maybe requests queued        */
        /* Maybe requests running       */
        /* Shutdown 1 completed         */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_dn:
            device->state = usbfront_device_state_i_sd;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_cn:
            usbfront_device_reject_message( device, stimulus );
            break;
          case usbfront_device_stimulus_sd:
            device->state = usbfront_device_state_i_sd_sd;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_unregister( device );
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_pt: /* FALLS THROUGH */
          case usbfront_device_stimulus_rq: /* FALLS THROUGH */
          case usbfront_device_stimulus_rr: /* FALLS THROUGH */
            break;
          case usbfront_device_stimulus_ri:
            device->state = usbfront_device_state_i_sd;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_purge_requests( device, flags );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      case usbfront_device_state_i_dn_cn_c1_c2_sd_sd:
        /* Registered                   */
        /* Probe running                */
        /* Shutdown 2 completed         */
        switch( stimulus )
        {
          case usbfront_device_stimulus_dt:
          case usbfront_device_stimulus_dn:
          case usbfront_device_stimulus_cn:
            break;
          case usbfront_device_stimulus_pt:
            break;
          case usbfront_device_stimulus_rq:
            usbfront_device_purge_requests( device, flags );
            break;
          case usbfront_device_stimulus_rr:
          case usbfront_device_stimulus_ri:
            break;
          case usbfront_device_stimulus_pc:
            device->state = usbfront_device_state_i_sd_sd;
            usbfront_device_disconnect( device );
            usbfront_device_disconnect_all_ports( device );
            usbfront_device_unregister( device );
            break;
          default:
            usbfront_device_invalid_stimulus( device, stimulus );
            break;
        }
        break;
      default:
        usbfront_device_invalid_stimulus( device, stimulus );
        break;
    }
}

static void usbfront_device_invalid_stimulus
  ( struct usbfront_device * device, usbfront_device_stimulus stimulus )
{
    trace();

    printk
    (
        KERN_ERR "usbfront: device %p in state %d"
        "received invalid stimulus %d",
        device,
        device->state,
        stimulus
    );
}

static void usbfront_device_reject_message
  ( struct usbfront_device * device, usbfront_device_stimulus stimulus )
{
    trace();

    printk
    (
        KERN_ERR "usbfront: received an invalid control interface message..."
    );

    usbfront_device_invalid_stimulus( device, stimulus );
}

static void usbfront_device_init_ring( struct usbfront_device * device )
{
    trace();

    SHARED_RING_INIT( device->shared_ring );

    FRONT_RING_INIT( &device->front_ring, device->shared_ring );
}

static void usbfront_device_send_interface_connect
  ( struct usbfront_device * device )
{
    trace();

    control_msg_t msg =
    {
        .type    = CMSG_USBIF_FE,
        .subtype = CMSG_USBIF_FE_INTERFACE_CONNECT,
        .length  = sizeof( usbif_fe_interface_connect_t ),
    };

    {
        usbif_fe_interface_connect_t cn =
        {
	    .handle      = device->handle,
            .shmem_frame =
              virt_to_machine( device->shared_ring ) >> PAGE_SHIFT,
        };

        memcpy( msg.msg, &cn, sizeof( cn ) );
    }

    if
    (
        ctrl_if_send_message_block( &msg, NULL, 0, TASK_UNINTERRUPTIBLE )
        !=
        0
    )
    {
        /* TASK_UNINTERRUPTIBLE ensures this can't happen. */

        ASSERT( 0 );
    }
}

static void usbfront_device_connect_part_one( struct usbfront_device * device )
{
    trace();

    usbfront_device_get( device );

    int scheduled = schedule_work( &device->connect_part_one_1_work );

    ASSERT( scheduled );
}

static void usbfront_device_connect_part_one_1( void * data )
{
    trace();

    struct usbfront_device * device = (struct usbfront_device *)data;

    usbfront_device_init_ring( device );

    usbfront_device_send_interface_connect( device );

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    usbfront_device_handle_stimulus
      ( device, usbfront_device_stimulus_c1, &flags );

    spin_unlock_irqrestore( &device->lock, flags );

    usbfront_device_put( device );
}

static void usbfront_device_connect_part_two( struct usbfront_device * device )
{
    trace();

    device->evtchn = device->status_changed.evtchn;

    usbfront_device_get( device );

    int scheduled = schedule_work( &device->connect_part_two_1_work );

    ASSERT( scheduled );
}

static irqreturn_t usbfront_device_interrupt
  ( int irq, void * context, struct pt_regs * regs );

static void usbfront_device_connect_part_two_1( void * data )
{
    trace();

    struct usbfront_device * device = (struct usbfront_device *)data;

    device->irq = bind_evtchn_to_irq( device->evtchn );

    device->got_irq =
    (
        request_irq
        (
            device->irq,
            usbfront_device_interrupt,
            SA_SAMPLE_RANDOM,
            "usbfront",
            device
        )
        ==
        0
    );

    ASSERT( device->got_irq );

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    usbfront_device_handle_stimulus
      ( device, usbfront_device_stimulus_c2, &flags );

    spin_unlock_irqrestore( &device->lock, flags );

    usbfront_device_put( device );
}

static void usbfront_device_disconnect( struct usbfront_device * device )
{
    trace();

    free_irq( device->irq, device );

    unbind_evtchn_from_irq( device->evtchn );
}

static void usbfront_device_unregister( struct usbfront_device * device )
{
    trace();

    usbfront_device_get( device );

    int scheduled = schedule_work( &device->unregister_1_work );

    ASSERT( scheduled );
}

static void usbfront_device_unregister_1( void * data )
{
    trace();

    struct usbfront_device * device = (struct usbfront_device *)data;

    down_write( &usbfront_device_list_sem );

    list_del_init( &device->link );

    up_write( &usbfront_device_list_sem );

    device_unregister( &device->dev );

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    usbfront_device_handle_stimulus
      ( device, usbfront_device_stimulus_ud, &flags );

    spin_unlock_irqrestore( &device->lock, flags );

    usbfront_device_put( device );
}

static void usbfront_device_probe_all_ports_1
(
    struct usbfront_device         * device,
    struct usbfront_device_request * request
);

static void usbfront_device_probe_all_ports
  ( struct usbfront_device * device, unsigned long * flags )
{
    /* trace(); */

    device->port_probe_count = USBFRONT_DEVICE_PORT_COUNT;

    spin_unlock_irqrestore( &device->lock, *flags );

    {
        int i;

        for( i = 0; i < USBFRONT_DEVICE_PORT_COUNT; i++ )
        {
            struct usbfront_device_request * request =
              &device->port_probe_request[ i ];

            usbif_request_t * parameters = &request->usbif_request;

            memset( parameters, 0, sizeof( *parameters ) );

            if
            (
                ( device->port_status[ i ].wPortStatus & USB_PORT_STAT_RESET )
                !=
                USB_PORT_STAT_RESET
            )
            {
                parameters->operation = USBIF_OP_PROBE;
                parameters->port      = i;
            }
            else
            {
                parameters->operation = USBIF_OP_RESET;
                parameters->port      = i;
            }

            usbfront_device_queue_internal_request
              ( device, request, usbfront_device_probe_all_ports_1 );
        }
    }

    spin_lock_irqsave( &device->lock, *flags );
}

static void usbfront_device_probe_all_ports_1
(
    struct usbfront_device         * device,
    struct usbfront_device_request * request
)
{
    /* trace(); */

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    int port_index = request->usbif_request.port;

    struct usb_port_status * port_status = &device->port_status[ port_index ];

    if( request->usbif_request.operation == USBIF_OP_PROBE )
    {
        if( request->usbif_response.status == 1 )
        {
            /* There is a device attached to the port. */

            if
            (
                ( port_status->wPortStatus & USB_PORT_STAT_CONNECTION )
                !=
                USB_PORT_STAT_CONNECTION
            )
            {
                /* There wasn't a device attached to the port before. */

                port_status->wPortStatus |= USB_PORT_STAT_CONNECTION;
                port_status->wPortChange |= USB_PORT_STAT_C_CONNECTION;
            }
        }
        else if( request->usbif_response.status == 0 )
        {
            /* There isn't a device attached to the port. */

            if
            (
                ( port_status->wPortStatus & USB_PORT_STAT_CONNECTION )
                ==
                USB_PORT_STAT_CONNECTION
            )
            {
                /* There was a device attached to the port before. */

                port_status->wPortStatus &= ~USB_PORT_STAT_CONNECTION;
                port_status->wPortChange |= USB_PORT_STAT_C_CONNECTION;
            }
        }
        else
        {
            trace3
            (
                "device %p: error %d probing port %d",
                device,
                (int)request->usbif_response.status,
                port_index
            );
        }
    }
    else
    {
        ASSERT( request->usbif_request.operation == USBIF_OP_RESET );

        if( request->usbif_response.status >= 0 )
        {
            /* Reset successful. */

            if( request->usbif_response.status == USBIF_RSP_LOW_SPEED )
            {
                port_status->wPortStatus |=  USB_PORT_STAT_LOW_SPEED;
                port_status->wPortStatus &= ~USB_PORT_STAT_HIGH_SPEED;
            }
            else if( request->usbif_response.status == USBIF_RSP_HIGH_SPEED )
            {
                port_status->wPortStatus &= ~USB_PORT_STAT_LOW_SPEED;
                port_status->wPortStatus |=  USB_PORT_STAT_HIGH_SPEED;
            }
            else
            {
                port_status->wPortStatus &= ~USB_PORT_STAT_LOW_SPEED;
                port_status->wPortStatus &= ~USB_PORT_STAT_HIGH_SPEED;
            }

            port_status->wPortStatus &= ~USB_PORT_STAT_RESET;
            port_status->wPortStatus |= USB_PORT_STAT_ENABLE;
            port_status->wPortChange |= USB_PORT_STAT_C_RESET;
        }
        else
        {
            /* Reset failed. */

            printk
            (
                KERN_ERR "usbfront: device %p: error %d resetting port %d",
                device,
                (int)request->usbif_response.status,
                port_index
            );

            port_status->wPortStatus &= ~USB_PORT_STAT_LOW_SPEED;
            port_status->wPortStatus &= ~USB_PORT_STAT_HIGH_SPEED;

            port_status->wPortStatus &= ~USB_PORT_STAT_RESET;
            port_status->wPortChange |= USB_PORT_STAT_C_RESET;
        }
    }

    if( --device->port_probe_count == 0 )
    {
        usbfront_device_handle_stimulus
          ( device, usbfront_device_stimulus_pc, &flags );
    }

    spin_unlock_irqrestore( &device->lock, flags );
}

static void usbfront_device_disconnect_all_ports
  ( struct usbfront_device * device )
{
    trace();

    int i;

    for( i = 0; i < USBFRONT_DEVICE_PORT_COUNT; i++ )
    {
        struct usb_port_status * port_status = &device->port_status[ i ];

        if
        (
            ( port_status->wPortStatus & USB_PORT_STAT_CONNECTION )
            ==
            USB_PORT_STAT_CONNECTION
        )
        {
            /* There was a device attached to the port before. */

            port_status->wPortStatus &= ~USB_PORT_STAT_CONNECTION;
            port_status->wPortChange |= USB_PORT_STAT_C_CONNECTION;
        }
    }
}

static void usbfront_device_kick_requests
  ( struct usbfront_device * device, unsigned long * flags )
{
    /* trace(); */

    int need_notify = 0;

    while
    (
        !list_empty( &device->request_list )
        &&
        !RING_FULL( &device->front_ring )
    )
    {
        usbif_request_t * ring_request = RING_GET_REQUEST
          ( &device->front_ring, device->front_ring.req_prod_pvt );

        struct usbfront_device_request * request = list_entry
          ( device->request_list.next, struct usbfront_device_request, link );

        list_del_init( &request->link );

        /* FIXME --- this is O(N) when it could be O(1) */

      ID_IN_USE_TRY_AGAIN:

        request->usbif_request.id = device->next_id;

        device->next_id++;

        {
            struct usbfront_device_request * running_request;

            list_for_each_entry( running_request, &device->running_list, link )
            {
                if
                (
                    running_request->usbif_request.id
                    ==
                    request->usbif_request.id
                )
                {
                    goto ID_IN_USE_TRY_AGAIN;
                }
            }
        }

        list_add_tail( &request->link, &device->running_list );

        memcpy
        (
            ring_request,
            &request->usbif_request,
            sizeof( *ring_request )
        );

        device->front_ring.req_prod_pvt++;

        RING_PUSH_REQUESTS( &device->front_ring );

        need_notify = 1;
    }

    if( !list_empty( &device->running_list ) )
    {
        usbfront_device_handle_stimulus
          ( device, usbfront_device_stimulus_rr, flags );
    }
    else
    {
        usbfront_device_handle_stimulus
          ( device, usbfront_device_stimulus_ri, flags );
    }

    spin_unlock_irqrestore( &device->lock, *flags );

    if( need_notify )
    {
        notify_via_evtchn( device->evtchn );
    }

    spin_lock_irqsave( &device->lock, *flags );
}

static irqreturn_t usbfront_device_interrupt
  ( int irq, void * context, struct pt_regs * regs )
{
    /* trace(); */

    struct list_head internal_list;

    INIT_LIST_HEAD( &internal_list );

    struct usbfront_device * device = (struct usbfront_device *)context;

    unsigned long flags;

    spin_lock_irqsave( &device->lock, flags );

    RING_IDX response_producer = device->front_ring.sring->rsp_prod;

    rmb();

    {
        RING_IDX i;

        for( i = device->front_ring.rsp_cons; i != response_producer; i++ )
        {
            usbif_response_t * response =
              RING_GET_RESPONSE( &device->front_ring, i );

            struct usbfront_device_request * request;

            list_for_each_entry( request, &device->running_list, link )
            {
                if( response->id == request->usbif_request.id )
                {
                    list_del_init( &request->link );

                    if( request->internal_completion == NULL )
                    {
                        list_add_tail
                          ( &request->link, &device->complete_list );
                    }
                    else
                    {
                        list_add_tail( &request->link, &internal_list );
                    }

                    memcpy
                    (
                        &request->usbif_response,
                        response,
                        sizeof( *response )
                    );

                    break;
                }
            }
        }

        device->front_ring.rsp_cons = i;
    }

    usbfront_device_kick_requests( device, &flags );

    int ping_driver_irq_handler = !list_empty( &device->complete_list );

    spin_unlock_irqrestore( &device->lock, flags );

    while( !list_empty( &internal_list ) )
    {
        struct usbfront_device_request * request = list_entry
          ( internal_list.next, struct usbfront_device_request, link );

        list_del_init( &request->link );

        request->internal_completion( device, request );
    }

    if( ping_driver_irq_handler )
    {
        (void)device->driver_irq_handler
          ( 0, device->driver_irq_context, regs );
    }

    return IRQ_HANDLED;
}

static void usbfront_device_purge_list
(
    struct list_head       * list,
    struct list_head       * internal_list,
    struct list_head       * complete_list,
    char                   * message
)
{
    trace();

    while( !list_empty( list ) )
    {
        trace1( "%s", message );

        struct usbfront_device_request * request = list_entry
          ( list->next, struct usbfront_device_request, link );

        list_del_init( &request->link );

        if( request->internal_completion == NULL )
        {
            list_add_tail( &request->link, complete_list );
        }
        else
        {
            list_add_tail( &request->link, internal_list );
        }

        request->usbif_response.status = USBIF_RSP_ERROR;
    }
}

static void usbfront_device_purge_requests
  ( struct usbfront_device * device, unsigned long * flags )
{
    trace();

    struct list_head internal_list;

    INIT_LIST_HEAD( &internal_list );

    usbfront_device_purge_list
    (
        &device->running_list,
        &internal_list,
        &device->complete_list,
        "Purging running request"
    );

    usbfront_device_purge_list
    (
        &device->request_list,
        &internal_list,
        &device->complete_list,
        "Purging queued request"
    );

    int ping_driver_irq_handler = !list_empty( &device->complete_list );

    /* FIXME: this chunk of code isn't allowed to release the lock in        */
    /* a state machine response.                                             */

    spin_unlock_irqrestore( &device->lock, *flags );

    while( !list_empty( &internal_list ) )
    {
        struct usbfront_device_request * request = list_entry
          ( internal_list.next, struct usbfront_device_request, link );

        list_del_init( &request->link );

        request->internal_completion( device, request );
    }

    if( ping_driver_irq_handler )
    {
        /* Intention here is to disable interrupts to provide driver with    */
        /* the execution context it is expecting in an interrupt handler.    */

        unsigned long flags;

        local_irq_save( flags );

        (void)device->driver_irq_handler
          ( 0, device->driver_irq_context, NULL );

        local_irq_restore( flags );
    }

    spin_lock_irqsave( &device->lock, *flags );
}

static void usbfront_device_complete_shutdown
  ( struct usbfront_device * device )
{
    trace();

    device->shutdown = 1;

    wake_up( &device->shutdown_wait_queue_head );
}
