/*****************************************************************************/
/* Discovery code discovers connections to back-ends and allocates           */
/* usbfront_device devices on the usbfront_bus.                              */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on arch/xen/drivers/usbif/frontend/main.c, original copyright       */
/* notice follows...                                                         */
/*****************************************************************************/

/*
 * Xen Virtual USB Frontend Driver 
 *
 * This file contains the first version of the Xen virtual USB hub
 * that I've managed not to delete by mistake (3rd time lucky!).
 *
 * Based on Linux's uhci.c, original copyright notices are displayed
 * below.  Portions also (c) 2004 Intel Research Cambridge
 * and (c) 2004, 2005 Mark Williamson
 *
 * Contact <mark.williamson@cl.cam.ac.uk> or
 * <xen-devel@lists.sourceforge.net> regarding this code.
 *
 * Still to be (maybe) implemented:
 * - migration / backend restart support?
 * - support for building / using as a module
 */

/*
 * Universal Host Controller Interface driver for USB.
 *
 * Maintainer: Johannes Erdfelt <johannes@erdfelt.com>
 *
 * (C) Copyright 1999 Linus Torvalds
 * (C) Copyright 1999-2002 Johannes Erdfelt, johannes@erdfelt.com
 * (C) Copyright 1999 Randy Dunlap
 * (C) Copyright 1999 Georg Acher, acher@in.tum.de
 * (C) Copyright 1999 Deti Fliegl, deti@fliegl.de
 * (C) Copyright 1999 Thomas Sailer, sailer@ife.ee.ethz.ch
 * (C) Copyright 1999 Roman Weissgaerber, weissg@vienna.at
 * (C) Copyright 2000 Yggdrasil Computing, Inc. (port of new PCI interface
 *               support from usb-ohci.c by Adam Richter, adam@yggdrasil.com).
 * (C) Copyright 1999 Gregory P. Smith (from usb-ohci.c)
 *
 * Intel documents this fairly well, and as far as I know there
 * are no royalties or anything like that, but even so there are
 * people who decided that they want to do the same thing in a
 * completely different way.
 *
 * WARNING! The USB documentation is downright evil. Most of it
 * is just crap, written by a committee. You're better off ignoring
 * most of it, the important stuff is:
 *  - the low-level protocol (fairly simple but lots of small details)
 *  - working around the horridness of the rest
 */

#include <linux/module.h>
#include <asm-xen/ctrl_if.h>
#include "usbfront_assert.h"
#include "usbfront_discovery.h"
#include "usbfront_device.h"
#include "usbfront_trace.h"

static int usbfront_discovery_done;

static DECLARE_RWSEM( usbfront_discovery_shutting_down_sem );

static int usbfront_discovery_shutting_down = 0;

static void usbfront_discovery_ctrl_if_receiver
  ( ctrl_msg_t * msg, unsigned long id )
{
    trace();

    if( msg->length >= sizeof( usbif_be_header_t ) )
    {
        usbif_be_header_t * header = (usbif_be_header_t *)&msg->msg[ 0 ];

        down_read( &usbfront_discovery_shutting_down_sem );

        struct usbfront_device * device = usbfront_device_find_or_allocate
          ( header->domid, usbfront_discovery_shutting_down );

        up_read( &usbfront_discovery_shutting_down_sem );

        if( device != NULL )
        {
            usbfront_device_handle_control_message( device, msg, id );

            usbfront_device_put( device );

            return;
        }
    }

    msg->length = 0;

    ctrl_if_send_response( msg );
}

static void usbfront_discovery_timed_wait_for_discovery_done( void )
{
    trace();

    int i;

    for( i = 0; ( !usbfront_discovery_done ) && ( i < ( 10 * HZ ) ); i++ )
    {
        set_current_state( TASK_INTERRUPTIBLE );

        schedule_timeout( 1 );
    }

    if( !usbfront_discovery_done )
    {
        printk( KERN_WARNING "Timeout connecting USB frontend driver!\n" );
    }
}

static void usbfront_discovery_send_driver_status( int status )
{
    trace();

    control_msg_t msg =
    {
        .type    = CMSG_USBIF_FE,
        .subtype = CMSG_USBIF_FE_DRIVER_STATUS_CHANGED,
        .length  = sizeof( usbif_fe_driver_status_changed_t ),
    };

    {
        usbif_fe_driver_status_changed_t st = { .status = status, };

        memcpy( msg.msg, &st, sizeof( st ) );
    }

    if
    (
        ctrl_if_send_message_and_get_response
          ( &msg, &msg, TASK_INTERRUPTIBLE )
        !=
        0
    )
    {
        ASSERT( 0 );
    }
}

static int usbfront_discovery_init_or_exit( int exit )
{
    trace();

    int return_value = 0;

    if( exit )
    {
        goto EXIT;
    }

    if
    (
        !ctrl_if_register_receiver
        (
            CMSG_USBIF_FE,
            usbfront_discovery_ctrl_if_receiver,
            CALLBACK_IN_BLOCKING_CONTEXT
        )
    )
    {
        return_value = -EBUSY;

        goto EXIT_NO_CTRL_IF;
    }

    usbfront_discovery_send_driver_status( USBIF_DRIVER_STATUS_UP );

    usbfront_discovery_timed_wait_for_discovery_done();

    return 0;

  EXIT:

    down_write( &usbfront_discovery_shutting_down_sem );

    usbfront_discovery_shutting_down = 1;

    usbfront_device_shutdown_all_phase_1();

    up_write( &usbfront_discovery_shutting_down_sem );

    usbfront_discovery_send_driver_status( USBIF_DRIVER_STATUS_DOWN );

    ctrl_if_unregister_receiver
      ( CMSG_USBIF_FE, usbfront_discovery_ctrl_if_receiver );

    usbfront_device_shutdown_all_phase_2();

  EXIT_NO_CTRL_IF:

    return return_value;
}

int usbfront_discovery_init( void )
{
    trace();

    return usbfront_discovery_init_or_exit( 0 );
}

void usbfront_discovery_exit( void )
{
    trace();

    (void)usbfront_discovery_init_or_exit( 1 );
}
