/*****************************************************************************/
/* A resource used by the usbfront host controller device driver to process  */
/* an URB by translating the URB and optional unlink request into            */
/* usbfront_device_requests for submission to the back end.                  */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on arch/xen/drivers/usbif/frontend/main.c, original copyright       */
/* notice follows...                                                         */
/*****************************************************************************/

/*
 * Xen Virtual USB Frontend Driver 
 *
 * This file contains the first version of the Xen virtual USB hub
 * that I've managed not to delete by mistake (3rd time lucky!).
 *
 * Based on Linux's uhci.c, original copyright notices are displayed
 * below.  Portions also (c) 2004 Intel Research Cambridge
 * and (c) 2004, 2005 Mark Williamson
 *
 * Contact <mark.williamson@cl.cam.ac.uk> or
 * <xen-devel@lists.sourceforge.net> regarding this code.
 *
 * Still to be (maybe) implemented:
 * - migration / backend restart support?
 * - support for building / using as a module
 */

/*
 * Universal Host Controller Interface driver for USB.
 *
 * Maintainer: Johannes Erdfelt <johannes@erdfelt.com>
 *
 * (C) Copyright 1999 Linus Torvalds
 * (C) Copyright 1999-2002 Johannes Erdfelt, johannes@erdfelt.com
 * (C) Copyright 1999 Randy Dunlap
 * (C) Copyright 1999 Georg Acher, acher@in.tum.de
 * (C) Copyright 1999 Deti Fliegl, deti@fliegl.de
 * (C) Copyright 1999 Thomas Sailer, sailer@ife.ee.ethz.ch
 * (C) Copyright 1999 Roman Weissgaerber, weissg@vienna.at
 * (C) Copyright 2000 Yggdrasil Computing, Inc. (port of new PCI interface
 *               support from usb-ohci.c by Adam Richter, adam@yggdrasil.com).
 * (C) Copyright 1999 Gregory P. Smith (from usb-ohci.c)
 *
 * Intel documents this fairly well, and as far as I know there
 * are no royalties or anything like that, but even so there are
 * people who decided that they want to do the same thing in a
 * completely different way.
 *
 * WARNING! The USB documentation is downright evil. Most of it
 * is just crap, written by a committee. You're better off ignoring
 * most of it, the important stuff is:
 *  - the low-level protocol (fairly simple but lots of small details)
 *  - working around the horridness of the rest
 */

#include "usbfront_assert.h"
#include "usbfront_driver.h"
#include "usbfront_hcd_resource.h"
#include "usbfront_trace.h"

static DEFINE_RWLOCK( dequeue_lock );

static int usbfront_hcd_resource_init_or_exit
  ( struct usbfront_hcd_resource * resource, int index, int exit )
{
    trace();

    int return_value = 0;

    if( exit )
    {
        goto EXIT;
    }

    INIT_LIST_HEAD( &resource->link );

    resource->index = index;

    resource->enqueue_request.resource = resource;
    resource->dequeue_request.resource = resource;

    resource->schedule = (usbif_iso_t *)__get_free_page( GFP_KERNEL );

    if( resource->schedule == NULL )
    {
        return_value = -ENOMEM;

        goto EXIT_NO_SCHEDULE;
    }

    return 0;

  EXIT:

    free_page( (unsigned long)resource->schedule );

  EXIT_NO_SCHEDULE:

    return return_value;
}

int usbfront_hcd_resource_init
  ( struct usbfront_hcd_resource * resource, int index )
{
    trace();

    return usbfront_hcd_resource_init_or_exit( resource, index, 0 );
}

void usbfront_hcd_resource_exit( struct usbfront_hcd_resource * resource )
{
    trace();

    usbfront_hcd_resource_init_or_exit( resource, 0, 1 );
}

/* usbfront_hcd_resource_start_urb must return before */
/* usbfront_hcd_resource_dequeue_urb may be called.   */

void usbfront_hcd_resource_start_urb
(
    struct usbfront_hcd_resource * resource,
    struct urb                   * urb,
    struct usbfront_device       * device
)
{
    trace();

    urb->hcpriv = resource;

    resource->urb = urb;

    atomic_set( &resource->requests_out, 1 );

    struct usbfront_device_request * request =
      &resource->enqueue_request.request;

    usbif_request_t * usbif_request = &request->usbif_request;

    memset( usbif_request, 0, sizeof( *usbif_request ) );

    usbif_request->operation       = USBIF_OP_IO;
    usbif_request->unlink_id       = resource->index;
    usbif_request->devnum          = usb_pipedevice( urb->pipe );
    usbif_request->endpoint        = usb_pipeendpoint( urb->pipe );
    usbif_request->direction       = !!usb_pipein( urb->pipe );
    usbif_request->pipe_type       = usb_pipetype( urb->pipe );
    usbif_request->transfer_buffer = virt_to_machine( urb->transfer_buffer );
    usbif_request->length          = urb->transfer_buffer_length;
    /* FIXME: transfer flags are ugly and linux specific */
    if( urb->transfer_flags & URB_SHORT_NOT_OK )
    {
        usbif_request->transfer_flags |= 0x0001;
    }
    if( urb->transfer_flags & URB_ZERO_PACKET )
    {
        usbif_request->transfer_flags |= 0x0040;
    }
    if( urb->setup_packet != NULL )
    {
        memcpy( usbif_request->setup, urb->setup_packet, 8 );
    }

    if( usb_pipetype( urb->pipe ) == PIPE_INTERRUPT )
    {
        usbif_request->interval = urb->interval;
    }
    else if( usb_pipetype( urb->pipe ) == PIPE_ISOCHRONOUS )
    {
        usbif_iso_t * schedule = resource->schedule;

        usbif_request->interval = urb->interval;

        usbif_request->iso_schedule = virt_to_machine( schedule );

        usbif_request->num_iso = urb->number_of_packets;

        int i;

        for( i = 0; i < urb->number_of_packets; i++ )
        {
            schedule[ i ].buffer_offset = urb->iso_frame_desc[ i ].offset;
            schedule[ i ].length        = urb->iso_frame_desc[ i ].length;
        }
    }

    usbfront_device_queue_request( device, request );
}

void usbfront_hcd_resource_request_completed
(
    struct usbfront_device_request * device_request,
    struct usbfront_hcd            * uhcd,
    struct usb_hcd                 * hcd,
    struct pt_regs                 * regs
)
{
    trace();

    unsigned long flags;

    /* Once this read lock is taken by the completing enqueue_request, a     */
    /* dequeue request will not be issued because...                         */

    read_lock_irqsave( &dequeue_lock, flags );

    /* ...a dequeue request will not be issued whilst the read lock is held  */
    /* since usbfront_hcd_resource_dequeue_urb takes a write lock...         */

    struct usbfront_hcd_resource_request * resource_request =
      usbfront_hcd_resource_request_device_request_to_resource_request
        ( device_request );

    struct usbfront_hcd_resource * resource = resource_request->resource;

    struct urb * urb = NULL;

    if( resource_request == &resource->enqueue_request )
    {
        urb = resource->urb;

        urb->actual_length = device_request->usbif_response.length;
        /* FIXME: map this status to/from linux error code */
        urb->status        = device_request->usbif_response.status;

        trace2
        (
            "status: %d, length %d",
            (int)device_request->usbif_response.status,
            (int)device_request->usbif_response.length
        );

        if( usb_pipetype( urb->pipe ) == PIPE_ISOCHRONOUS )
        {
            int i;

            for( i = 0; i < urb->number_of_packets; i++ )
            {
                usbif_iso_t * schedule = &resource->schedule[ i ];

                urb->iso_frame_desc[ i ].status        = schedule->status;
                urb->iso_frame_desc[ i ].actual_length = schedule->length;
            }
        }

        /* ...and the completing enqueue_request clears hcpriv here. */
        urb->hcpriv = 0;
    }

    read_unlock_irqrestore( &dequeue_lock, flags );

    /* So we are now guaranteed that a dequeue_request will not be issued    */
    /* unless it has already been issues and is already accounted for in     */
    /* requests_out.                                                         */

    /* We complete the urb when the enqueue_request completes. */

    if( urb != NULL )
    {
        usb_hcd_giveback_urb( hcd, urb, regs );
    }

    /* The resource is free when all requests are complete. The requests     */
    /* might complete concurrently so we need to use atomic dec here.        */

    if( atomic_dec_and_test( &resource->requests_out ) )
    {
        usbfront_hcd_resource_completed( uhcd, resource );
    }
}

void usbfront_hcd_resource_dequeue_urb
  ( struct urb * urb, struct usbfront_device * device )
{
    trace();

    struct usbfront_hcd_resource * resource = NULL;

    unsigned long flags;

    write_lock_irqsave( &dequeue_lock, flags );

    if( urb->hcpriv != NULL )
    {
        resource = urb->hcpriv;

        atomic_inc( &resource->requests_out );
    }

    write_unlock_irqrestore( &dequeue_lock, flags );

    if( resource )
    {
        struct usbfront_device_request * request =
          &resource->dequeue_request.request;

        usbif_request_t * usbif_request = &request->usbif_request;

        memset( usbif_request, 0, sizeof( *usbif_request ) );

        usbif_request->operation = USBIF_OP_UNLINK;
        usbif_request->unlink_id = resource->index;

        usbfront_device_queue_request( device, request );
    }
}
