/*****************************************************************************/
/* A program which prints out a python implementation of the enumeration of  */
/* the state machine for the UsbifBackendInterface object in usbif.py.       */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <assert.h>

/* Here are all of the stimuli. */

typedef enum
{
    STIM_fdu,
    STIM_fdd,
    STIM_bdu,
    STIM_bdd,
    STIM_fic,
    STIM_dst,
    STIM_add,
    STIM_nfy,
    STIM_fer,
    STIM_ber,
    STIM_Count
}
STIM;

/* This is a list of components of the state vector. */

typedef enum
{
    STATE_FeDown               = 0x00000000,
    STATE_FeUp                 = 0x00000001,
    STATE_FeDowning            = 0x00000002,
    STATE_FeMask               = 0x00000003,

    STATE_BeDown               = 0x00000000,
    STATE_BeUp                 = 0x00000004,
    STATE_BeDowning            = 0x00000008,
    STATE_BeMask               = 0x0000000C,

    STATE_SendBeIdle           = 0x00000000,
    STATE_SendBeCreate         = 0x00000010,
    STATE_SendBeDestroy        = 0x00000020,
    STATE_SendBeConnect        = 0x00000030,
    STATE_SendBeDisconnect     = 0x00000040,
    STATE_SendBeDontCare       = 0x00000050,
    STATE_SendBeMask           = 0x00000070,

/* GAP from futzing around */

    STATE_BeIfDestroyed        = 0x00000000,
    STATE_BeIfCreating         = 0x00000200,
    STATE_BeIfDestroying       = 0x00000400,
    STATE_BeIfDisconnected     = 0x00000600,
    STATE_BeIfConnecting       = 0x00000800,
    STATE_BeIfDisconnecting    = 0x00000A00,
    STATE_BeIfConnected        = 0x00000E00,
    STATE_BeIfMask             = 0x00000E00,

    STATE_FeIfDestroyed        = 0x00000000,
    STATE_FeIfDisconnected     = 0x00001000,
    STATE_FeIfConnectable      = 0x00002000,
    STATE_FeIfConnected        = 0x00003000,
    STATE_FeIfConnectedStale   = 0x00004000,
    STATE_FeIfMask             = 0x00007000,

    STATE_DevicesDown          = 0x00000000,
    STATE_DevicesUping         = 0x00008000,
    STATE_DevicesUp            = 0x00010000,
    STATE_DevicesDowning       = 0x00018000,
    STATE_DevicesMask          = 0x00018000,

    STATE_AddClean             = 0x00000000,
    STATE_AddDirty             = 0x00020000,
    STATE_AddMask              = 0x00020000,

    STATE_EvtChanClosed        = 0x00000000,
    STATE_EvtChanOpen          = 0x00040000,
    STATE_EvtChanMask          = 0x00040000,

    STATE_DestroyIdle          = 0x00000000,
    STATE_DestroyOutstanding   = 0x00080000,
    STATE_DestroyMask          = 0x00080000,

    STATE_SendFeIdle           = 0x00000000,
    STATE_SendFeISDisconnected = 0x00100000,
    STATE_SendFeISConnected    = 0x00200000,
    STATE_SendFeISDestroyed    = 0x00300000,
    STATE_SendFeDontCare       = 0x00400000,
    STATE_SendFeMask           = 0x00700000,

    STATE_Count                = 0x00800000
}
STATE;

/* Here are all of the responses. */

typedef enum
{
    /* Allow claim/release port messages to BE, completes after all BE ports */
    /* are claimed.                                                          */
    RESP_NotifyDevicesUp,

    /* Devices assume BE state reset, all claim/release port messages        */
    /* quiesced before completion.                                           */
    RESP_NotifyDevicesDown,

    /* Must guarantee no more/in-flight control messages to FE and no FE     */
    /* resources in use by BE before completion.                             */
    RESP_CompleteFeDown,

    /* Must guarantee no more/in-flight control messages to BE before        */
    /* completion including claim/release port messages so must notify       */
    /* devices down.                                                         */
    RESP_CompleteBeDown,

    /* Must guarantee FE and BE interfaces torn down before completion.      */
    RESP_CompleteDestroy,

    /* Must Open before be_connect.                                          */
    RESP_OpenEvtChan,

    /* Must close to guarantee no more interrupts before be_disconnect/fe is */
    /* disconnected and before fd down/be down acknowledged??*/
    RESP_CloseEvtChan,

    RESP_SendBeCreate,
    RESP_SendBeDestroy,
    RESP_SendBeConnect,
    RESP_SendBeDisconnect,
    RESP_SendFeISDisconnected,
    RESP_SendFeISConnected,
    RESP_SendFeISDestroyed,
    RESP_Count
}
RESP;

/* This ENR structure is for an enumeration state. */

typedef struct enr_struct ENR;

struct enr_struct
{
    /* This is the ASCII name of the state. */
    char  ENR_Name[1024];

    /* This is the state vector for the state. */
    STATE ENR_ThisState;

    /* These are state vectors for the transitions. */
    STATE ENR_NextState[ STIM_Count ];

    /* These are responses. */
    RESP  ENR_Resp[ STIM_Count ][ RESP_Count ];
};

/* Given a state vector State, and state Name, this function fills in the    */
/* ENR *pEnr.                                                                */

void new_state( ENR * pEnr, STATE State, char * Name )
{
    /* Copy the ASCII state name into the ENR. */
    strncpy( pEnr->ENR_Name, Name, 1023 );
    pEnr->ENR_Name[ 1023 ] = 0;

    /* Store the state vector. */
    pEnr->ENR_ThisState = State;

    /* Initialise all of the stimuli to be invalid. */
    {
        int StimIndex;

        for( StimIndex = 0; StimIndex < STIM_Count; StimIndex++ )
        {
            pEnr->ENR_NextState[ StimIndex ] = STATE_Count;

            {
                int RespIndex;

                for( RespIndex = 0; RespIndex < RESP_Count; RespIndex++ )
                {
                    pEnr->ENR_Resp[ StimIndex ][ RespIndex ] = RESP_Count;
                }
            }
        }
    }

    if
    (
        ( ( State & STATE_FeMask ) == STATE_FeDown )
    )
    {
        int rspidx = 0;

        assert( ( State & STATE_SendFeMask ) == STATE_SendFeIdle );
        assert( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected );
        assert( ( State & STATE_FeIfMask ) == STATE_FeIfDestroyed );
        assert
          ( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

        pEnr->ENR_NextState[ STIM_fdu ] = State;

        pEnr->ENR_NextState[ STIM_fdu ] &= ~STATE_FeMask;
        pEnr->ENR_NextState[ STIM_fdu ] |=  STATE_FeUp;

        if
        (
            ( ( State & STATE_BeMask ) == STATE_BeUp )
            &&
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDestroyed )
        )
        {
            assert( ( State & STATE_DevicesMask ) == STATE_DevicesDown );

            pEnr->ENR_Resp[ STIM_fdu ][ rspidx++ ] = RESP_SendBeCreate;

            pEnr->ENR_NextState[ STIM_fdu ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_fdu ] |=  STATE_SendBeCreate;

            pEnr->ENR_NextState[ STIM_fdu ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_fdu ] |=  STATE_BeIfCreating;
        }

        {
            pEnr->ENR_Resp[ STIM_fdu ][ rspidx++ ] = RESP_SendFeISDisconnected;

            pEnr->ENR_NextState[ STIM_fdu ] &= ~STATE_SendFeMask;
            pEnr->ENR_NextState[ STIM_fdu ] |=  STATE_SendFeISDisconnected;

            pEnr->ENR_NextState[ STIM_fdu ] &= ~STATE_FeIfMask;
            pEnr->ENR_NextState[ STIM_fdu ] |=  STATE_FeIfDisconnected;
        }
    }

    if
    (
        ( ( State & STATE_FeMask ) == STATE_FeUp )
    )
    {
        int rspidx = 0;

        pEnr->ENR_NextState[ STIM_fdd ] = State;

        pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_FeMask;
        pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_FeDowning;

        pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_FeIfMask;
        pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_FeIfDestroyed;

        pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_AddMask;
        pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_AddClean;

        if( ( State & STATE_SendFeMask ) != STATE_SendFeIdle )
        {
            pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_SendFeMask;
            pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_SendFeDontCare;
        }

        if( ( State & STATE_EvtChanMask ) == STATE_EvtChanOpen )
        {
            pEnr->ENR_Resp[ STIM_fdd ][ rspidx++ ] = RESP_CloseEvtChan;

            pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_EvtChanMask;
            pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_EvtChanClosed;
        }

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_BeIfMask ) > STATE_BeIfDisconnected )
        )
        {
            assert( ( State & STATE_BeMask ) == STATE_BeUp );
            assert( ( State & STATE_BeIfMask ) == STATE_BeIfConnected );

            pEnr->ENR_Resp[ STIM_fdd ][ rspidx++ ] = RESP_SendBeDisconnect;

            pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_SendBeDisconnect;

            pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_BeIfDisconnecting;
        }

        if
        (
            ( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
            &&
            ( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected )
        )
        {
            pEnr->ENR_Resp[ STIM_fdd ][ rspidx++ ] = RESP_CompleteFeDown;

            pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_FeMask;
            pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_FeDown;
        }

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_BeIfMask ) > STATE_BeIfDestroyed )
            &&
            ( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesUp )
        )
        {
            assert( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnected );
            assert( ( State & STATE_BeMask ) == STATE_BeUp );

            pEnr->ENR_Resp[ STIM_fdd ][ rspidx++ ] = RESP_NotifyDevicesDown;

            pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_DevicesMask;
            pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_DevicesDowning;
        }

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_BeIfMask ) > STATE_BeIfDestroyed )
            &&
            ( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDown )
        )
        {
            assert( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnected );
            assert( ( State & STATE_BeMask ) == STATE_BeUp );

            pEnr->ENR_Resp[ STIM_fdd ][ rspidx++ ] = RESP_SendBeDestroy;

            pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_SendBeDestroy;

            pEnr->ENR_NextState[ STIM_fdd ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_fdd ] |=  STATE_BeIfDestroying;
        }
    }

    if
    (
        ( ( State & STATE_BeMask ) == STATE_BeDown )
    )
    {
        int rspidx = 0;

        assert( ( State & STATE_SendBeMask ) == STATE_SendBeIdle );
        assert( ( State & STATE_BeIfMask ) == STATE_BeIfDestroyed );
        assert( ( State & STATE_DevicesMask ) == STATE_DevicesDown );
        assert( ( State & STATE_AddMask ) == STATE_AddClean );
        assert
          ( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

        pEnr->ENR_NextState[ STIM_bdu ] = State;

        pEnr->ENR_NextState[ STIM_bdu ] &= ~STATE_BeMask;
        pEnr->ENR_NextState[ STIM_bdu ] |=  STATE_BeUp;

        if
        (
            ( ( State & STATE_FeMask ) == STATE_FeUp )
            &&
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyIdle )
        )
        {
            pEnr->ENR_Resp[ STIM_bdu ][ rspidx++ ] = RESP_SendBeCreate;

            pEnr->ENR_NextState[ STIM_bdu ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_bdu ] |=  STATE_SendBeCreate;

            pEnr->ENR_NextState[ STIM_bdu ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_bdu ] |=  STATE_BeIfCreating;
        }
    }

    if
    (
        ( ( State & STATE_BeMask ) == STATE_BeUp )
    )
    {
        int rspidx = 0;

        pEnr->ENR_NextState[ STIM_bdd ] = State;

        pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_BeMask;
        pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_BeDowning;

        pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_BeIfMask;
        pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_BeIfDestroyed;

        if( ( State & STATE_SendBeMask ) != STATE_SendBeIdle )
        {
            pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_SendBeDontCare;
        }

        pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_AddMask;
        pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_AddClean;

        if( ( State & STATE_EvtChanMask ) == STATE_EvtChanOpen )
        {
            pEnr->ENR_Resp[ STIM_bdd ][ rspidx++ ] = RESP_CloseEvtChan;

            pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_EvtChanMask;
            pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_EvtChanClosed;
        }

        if
        (
            ( ( State & STATE_DevicesMask ) == STATE_DevicesUp )
        )
        {
            pEnr->ENR_Resp[ STIM_bdd ][ rspidx++ ] = RESP_NotifyDevicesDown;

            pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_DevicesMask;
            pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_DevicesDowning;
        }

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDown )
        )
        {
            pEnr->ENR_Resp[ STIM_bdd ][ rspidx++ ] = RESP_CompleteBeDown;

            pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_BeMask;
            pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_BeDown;
        }

        if
        (
            ( ( State & STATE_FeIfMask ) == STATE_FeIfConnected )
        )
        {
            if( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
            {
                pEnr->ENR_Resp[ STIM_bdd ][ rspidx++ ] =
                  RESP_SendFeISDisconnected;

                pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_SendFeMask;
                pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_SendFeISDisconnected;

                pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_FeIfMask;
                pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_FeIfDisconnected;
            }
            else
            {
                pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_FeIfMask;
                pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_FeIfConnectedStale;
            }
        }

        if
        (
            ( ( State & STATE_FeMask ) == STATE_FeDowning )
            &&
            ( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
        )
        {
            pEnr->ENR_Resp[ STIM_bdd ][ rspidx++ ] = RESP_CompleteFeDown;

            pEnr->ENR_NextState[ STIM_bdd ] &= ~STATE_FeMask;
            pEnr->ENR_NextState[ STIM_bdd ] |=  STATE_FeDown;
        }
    }

    if
    (
        ( ( State & STATE_FeIfMask ) == STATE_FeIfDisconnected )
    )
    {
        int rspidx = 0;

        assert( ( State & STATE_FeMask ) == STATE_FeUp );
        assert
          ( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

        pEnr->ENR_NextState[ STIM_fic ] = State;

        pEnr->ENR_NextState[ STIM_fic ] &= ~STATE_FeIfMask;
        pEnr->ENR_NextState[ STIM_fic ] |=  STATE_FeIfConnectable;

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnected )
        )
        {
            assert( ( State & STATE_BeMask ) == STATE_BeUp );

            pEnr->ENR_Resp[ STIM_fic ][ rspidx++ ] = RESP_OpenEvtChan;

            pEnr->ENR_NextState[ STIM_fic ] &= ~STATE_EvtChanMask;
            pEnr->ENR_NextState[ STIM_fic ] |=  STATE_EvtChanOpen;

            pEnr->ENR_Resp[ STIM_fic ][ rspidx++ ] = RESP_SendBeConnect;

            pEnr->ENR_NextState[ STIM_fic ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_fic ] |=  STATE_SendBeConnect;

            pEnr->ENR_NextState[ STIM_fic ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_fic ] |=  STATE_BeIfConnecting;
        }
    }

    if
    (
        ( ( State & STATE_DestroyMask ) == STATE_DestroyIdle )
    )
    {
        int rspidx = 0;

        pEnr->ENR_NextState[ STIM_dst ] = State;

        pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_DestroyMask;
        pEnr->ENR_NextState[ STIM_dst ] |=  STATE_DestroyOutstanding;

        pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_AddMask;
        pEnr->ENR_NextState[ STIM_dst ] |=  STATE_AddClean;

        if( ( State & STATE_EvtChanMask ) == STATE_EvtChanOpen )
        {
            pEnr->ENR_Resp[ STIM_dst ][ rspidx++ ] = RESP_CloseEvtChan;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_EvtChanMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_EvtChanClosed;
        }

        if
        (
            ( ( State & STATE_DevicesMask ) == STATE_DevicesUp )
        )
        {
            pEnr->ENR_Resp[ STIM_dst ][ rspidx++ ] = RESP_NotifyDevicesDown;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_DevicesMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_DevicesDowning;
        }

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfConnected )
        )
        {
            pEnr->ENR_Resp[ STIM_dst ][ rspidx++ ] = RESP_SendBeDisconnect;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_SendBeDisconnect;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_BeIfDisconnecting;
        }

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnected )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDown )
        )
        {
            pEnr->ENR_Resp[ STIM_dst ][ rspidx++ ] = RESP_SendBeDestroy;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_SendBeDestroy;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_BeIfDestroying;
        }

        if
        (
            ( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
            &&
            ( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected )
            &&
            ( ( State & STATE_FeIfMask ) == STATE_FeIfConnected )
        )
        {
            pEnr->ENR_Resp[ STIM_dst ][ rspidx++ ] = RESP_SendFeISDisconnected;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_SendFeMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_SendFeISDisconnected;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_FeIfMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_FeIfDisconnected;
        }

        if
        (
            ( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
            &&
            ( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected )
            &&
            (
                ( ( State & STATE_FeIfMask ) == STATE_FeIfConnectable )
                ||
                ( ( State & STATE_FeIfMask ) == STATE_FeIfDisconnected )
            )
        )
        {
            pEnr->ENR_Resp[ STIM_dst ][ rspidx++ ] = RESP_SendFeISDestroyed;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_SendFeMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_SendFeISDestroyed;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_FeIfMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_FeIfDestroyed;
        }

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDestroyed )
            &&
            ( ( State & STATE_FeIfMask ) == STATE_FeIfDestroyed )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDown )
        )
        {
            assert( ( State & STATE_AddMask ) == STATE_AddClean );
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

            pEnr->ENR_Resp[ STIM_dst ][ rspidx++ ] = RESP_CompleteDestroy;

            pEnr->ENR_NextState[ STIM_dst ] &= ~STATE_DestroyMask;
            pEnr->ENR_NextState[ STIM_dst ] |=  STATE_DestroyIdle;
        }
    }

    if
    (
        ( ( State & STATE_DestroyMask ) == STATE_DestroyIdle )
    )
    {
        int rspidx = 0;

        pEnr->ENR_NextState[ STIM_add ] = State;

        if
        (
            ( ( State & STATE_FeMask ) == STATE_FeUp )
            &&
            ( ( State & STATE_BeMask ) == STATE_BeUp )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyIdle )
        )
        {
            if
            (
                ( ( State & STATE_DevicesMask ) == STATE_DevicesUping )
            )
            {
                pEnr->ENR_NextState[ STIM_add ] &= ~STATE_AddMask;
                pEnr->ENR_NextState[ STIM_add ] |=  STATE_AddDirty;
            }

            if
            (
                ( ( State & STATE_DevicesMask ) == STATE_DevicesUp )
            )
            {
                pEnr->ENR_Resp[ STIM_add ][ rspidx++ ] = RESP_NotifyDevicesUp;

                pEnr->ENR_NextState[ STIM_add ] &= ~STATE_DevicesMask;
                pEnr->ENR_NextState[ STIM_add ] |=  STATE_DevicesUping;
            }
        }
    }

    if
    (
        ( ( State & STATE_DevicesMask ) == STATE_DevicesUping )
        ||
        ( ( State & STATE_DevicesMask ) == STATE_DevicesDowning )
    )
    {
        int rspidx = 0;

        pEnr->ENR_NextState[ STIM_nfy ] = State;

        if
        (
            ( ( State & STATE_DevicesMask ) == STATE_DevicesUping )
        )
        {
            pEnr->ENR_NextState[ STIM_nfy ] &= ~STATE_DevicesMask;
            pEnr->ENR_NextState[ STIM_nfy ] |=  STATE_DevicesUp;
        }
        else
        {
            pEnr->ENR_NextState[ STIM_nfy ] &= ~STATE_DevicesMask;
            pEnr->ENR_NextState[ STIM_nfy ] |=  STATE_DevicesDown;
        }

        if( ( State & STATE_AddMask ) == STATE_AddDirty )
        {
            pEnr->ENR_Resp[ STIM_nfy ][ rspidx++ ] = RESP_NotifyDevicesUp;

            pEnr->ENR_NextState[ STIM_nfy ] &= ~STATE_DevicesMask;
            pEnr->ENR_NextState[ STIM_nfy ] |=  STATE_DevicesUping;
        }

        if
        (
            ( ( State & STATE_BeMask ) == STATE_BeDowning )
            &&
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDowning )
        )
        {
            assert( ( State & STATE_AddMask ) == STATE_AddClean );

            pEnr->ENR_Resp[ STIM_nfy ][ rspidx++ ] = RESP_CompleteBeDown;

            pEnr->ENR_NextState[ STIM_nfy ] &= ~STATE_BeMask;
            pEnr->ENR_NextState[ STIM_nfy ] |=  STATE_BeDown;
        }

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnected )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDowning )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyOutstanding )
        )
        {
            pEnr->ENR_Resp[ STIM_nfy ][ rspidx++ ] = RESP_SendBeDestroy;

            pEnr->ENR_NextState[ STIM_nfy ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_nfy ] |=  STATE_SendBeDestroy;

            pEnr->ENR_NextState[ STIM_nfy ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_nfy ] |=  STATE_BeIfDestroying;
        }
    }

    if
    (
        ( ( State & STATE_SendFeMask ) != STATE_SendFeIdle )
    )
    {
        int rspidx = 0;

        pEnr->ENR_NextState[ STIM_fer ] = State;

        pEnr->ENR_NextState[ STIM_fer ] &= ~STATE_SendFeMask;
        pEnr->ENR_NextState[ STIM_fer ] |=  STATE_SendFeIdle;

        if
        (
            ( ( State & STATE_FeMask ) == STATE_FeDowning )
            &&
            ( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected )
        )
        {
            pEnr->ENR_Resp[ STIM_fer ][ rspidx++ ] = RESP_CompleteFeDown;

            pEnr->ENR_NextState[ STIM_fer ] &= ~STATE_FeMask;
            pEnr->ENR_NextState[ STIM_fer ] |=  STATE_FeDown;
        }

        if
        (
            ( ( State & STATE_SendBeMask ) == STATE_SendBeIdle )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDestroyed )
            &&
            ( ( State & STATE_FeIfMask ) == STATE_FeIfDestroyed )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDown )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyOutstanding )
        )
        {
            assert( ( State & STATE_AddMask ) == STATE_AddClean );
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

            pEnr->ENR_Resp[ STIM_fer ][ rspidx++ ] = RESP_CompleteDestroy;

            pEnr->ENR_NextState[ STIM_fer ] &= ~STATE_DestroyMask;
            pEnr->ENR_NextState[ STIM_fer ] |=  STATE_DestroyIdle;
        }

        if
        (
            ( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected )
            &&
            (
                ( ( State & STATE_FeIfMask ) == STATE_FeIfConnectedStale )
                ||
                (
                    ( ( State & STATE_FeIfMask ) == STATE_FeIfConnected )
                    &&
                    (
                        ( State & STATE_DestroyMask )
                        ==
                        STATE_DestroyOutstanding
                    )
                )
            )
        )
        {
            assert( ( State & STATE_FeMask ) == STATE_FeUp );

            pEnr->ENR_Resp[ STIM_fer ][ rspidx++ ] = RESP_SendFeISDisconnected;

            pEnr->ENR_NextState[ STIM_fer ] &= ~STATE_SendFeMask;
            pEnr->ENR_NextState[ STIM_fer ] |=  STATE_SendFeISDisconnected;

            pEnr->ENR_NextState[ STIM_fer ] &= ~STATE_FeIfMask;
            pEnr->ENR_NextState[ STIM_fer ] |=  STATE_FeIfDisconnected;
        }

        if
        (
            ( ( State & STATE_FeMask ) == STATE_FeUp )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfConnected )
            &&
            ( ( State & STATE_FeIfMask ) == STATE_FeIfConnectable )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyIdle )
        )
        {
            assert( ( State & STATE_BeMask ) == STATE_BeUp );
            assert( ( State & STATE_SendBeMask ) == STATE_SendBeIdle );
/* FIXME           assert
            (
                ( ( State & STATE_DevicesMask ) == STATE_DevicesUp )
                ||
                ( ( State & STATE_DevicesMask ) == STATE_DevicesUping )
            );*/
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanOpen );

            pEnr->ENR_Resp[ STIM_fer ][ rspidx++ ] = RESP_SendFeISConnected;

            pEnr->ENR_NextState[ STIM_fer ] &= ~STATE_SendFeMask;
            pEnr->ENR_NextState[ STIM_fer ] |=  STATE_SendFeISConnected;

            pEnr->ENR_NextState[ STIM_fer ] &= ~STATE_FeIfMask;
            pEnr->ENR_NextState[ STIM_fer ] |=  STATE_FeIfConnected;
        }

        if
        (
            ( ( State & STATE_FeMask ) == STATE_FeUp )
            &&
            (
                ( ( State & STATE_FeIfMask ) == STATE_FeIfDisconnected )
                ||
                (
                    ( ( State & STATE_FeIfMask ) == STATE_FeIfConnectable )
                    &&
                    ( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected )
                )
            )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyOutstanding )
        )
        {
            assert( ( State & STATE_BeIfMask ) <= STATE_BeIfDisconnected );
            assert( ( State & STATE_AddMask ) == STATE_AddClean );
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

            pEnr->ENR_Resp[ STIM_fer ][ rspidx++ ] = RESP_SendFeISDestroyed;

            pEnr->ENR_NextState[ STIM_fer ] &= ~STATE_SendFeMask;
            pEnr->ENR_NextState[ STIM_fer ] |=  STATE_SendFeISDestroyed;

            pEnr->ENR_NextState[ STIM_fer ] &= ~STATE_FeIfMask;
            pEnr->ENR_NextState[ STIM_fer ] |=  STATE_FeIfDestroyed;
        }
    }

    if
    (
        ( ( State & STATE_SendBeMask ) != STATE_SendBeIdle )
    )
    {
        int rspidx = 0;

        pEnr->ENR_NextState[ STIM_ber ] = State;

        pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_SendBeMask;
        pEnr->ENR_NextState[ STIM_ber ] |=  STATE_SendBeIdle;

        if( ( State & STATE_BeIfMask ) == STATE_BeIfCreating )
        {
            assert( ( State & STATE_SendBeMask ) == STATE_SendBeCreate );

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_BeIfDisconnected;
        }
        else if( ( State & STATE_BeIfMask ) == STATE_BeIfDestroying )
        {
            assert( ( State & STATE_SendBeMask ) == STATE_SendBeDestroy );

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_BeIfDestroyed;
        }
        else if( ( State & STATE_BeIfMask ) == STATE_BeIfConnecting )
        {
            assert( ( State & STATE_SendBeMask ) == STATE_SendBeConnect );

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_BeIfConnected;
        }
        else if( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnecting )
        {
            assert( ( State & STATE_SendBeMask ) == STATE_SendBeDisconnect );

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_BeIfDisconnected;
        }

        if
        (
            ( ( State & STATE_FeMask ) == STATE_FeUp )
            &&
            ( ( State & STATE_BeMask ) == STATE_BeUp )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfCreating )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyIdle )
        )
        {
            assert( ( State & STATE_DevicesMask ) == STATE_DevicesDown );
            assert( ( State & STATE_AddMask ) == STATE_AddClean );
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_NotifyDevicesUp;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_DevicesMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_DevicesUping;
        }

        if
        (
            ( ( State & STATE_BeMask ) == STATE_BeDowning )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDown )
        )
        {
            assert( ( State & STATE_AddMask ) == STATE_AddClean );
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_CompleteBeDown;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_BeMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_BeDown;
        }

        if
        (
            ( ( State & STATE_FeMask ) == STATE_FeDowning )
            &&
            ( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnecting )
        )
        {
            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_CompleteFeDown;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_FeMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_FeDown;
        }

        if
        (
            ( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDestroying )
            &&
            ( ( State & STATE_FeIfMask ) == STATE_FeIfDestroyed )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDown )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyOutstanding )
        )
        {
            assert( ( State & STATE_AddMask ) == STATE_AddClean );
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_CompleteDestroy;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_DestroyMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_DestroyIdle;
        }

        if
        (
            ( ( State & STATE_FeMask ) == STATE_FeUp )
            &&
            ( ( State & STATE_BeMask ) == STATE_BeUp )
            &&
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDestroying )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyIdle )
        )
        {
            assert( ( State & STATE_DevicesMask ) == STATE_DevicesDown );
            assert( ( State & STATE_AddMask ) == STATE_AddClean );
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_SendBeCreate;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_SendBeCreate;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_BeIfCreating;
        }

        if
        (
            (
                ( ( State & STATE_BeIfMask ) == STATE_BeIfCreating )
                ||
                ( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnecting )
            )
            &&
            ( ( State & STATE_DevicesMask ) == STATE_DevicesDown )
            &&
            (
                ( ( State & STATE_DestroyMask ) == STATE_DestroyOutstanding )
                ||
                ( ( State & STATE_FeMask ) != STATE_FeUp )
            )
        )
        {
            assert( ( State & STATE_BeMask ) == STATE_BeUp );

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_SendBeDestroy;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_SendBeDestroy;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_BeIfDestroying;
        }

        if
        (
            (
                ( ( State & STATE_BeIfMask ) == STATE_BeIfCreating )
                ||
                ( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnecting )
            )
            &&
            ( ( State & STATE_FeIfMask ) == STATE_FeIfConnectable )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyIdle )
        )
        {
            assert( ( State & STATE_FeMask ) == STATE_FeUp );
            assert( ( State & STATE_BeMask ) == STATE_BeUp );
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_OpenEvtChan;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_EvtChanMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_EvtChanOpen;

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_SendBeConnect;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_SendBeConnect;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_BeIfConnecting;
        }

        if
        (
            ( ( State & STATE_BeIfMask ) == STATE_BeIfConnecting )
            &&
            (
                ( ( State & STATE_FeMask ) == STATE_FeDowning )
                ||
                ( ( State & STATE_DestroyMask ) == STATE_DestroyOutstanding )
            )
        )
        {
            assert( ( State & STATE_BeMask ) == STATE_BeUp );
            assert( ( State & STATE_AddMask ) == STATE_AddClean );

            if( ( State & STATE_EvtChanMask ) == STATE_EvtChanOpen )
            {
                pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_CloseEvtChan;

                pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_EvtChanMask;
                pEnr->ENR_NextState[ STIM_ber ] |=  STATE_EvtChanClosed;
            }

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_SendBeDisconnect;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_SendBeMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_SendBeDisconnect;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_BeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_BeIfDisconnecting;
        }

        if
        (
            ( ( State & STATE_BeIfMask ) == STATE_BeIfDisconnecting )
            &&
            (
                ( ( State & STATE_FeIfMask ) == STATE_FeIfConnectedStale )
                ||
                ( ( State & STATE_FeIfMask ) == STATE_FeIfConnected )
            )
            &&
            ( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
        )
        {
            assert( ( State & STATE_FeMask ) == STATE_FeUp );
            assert( ( State & STATE_BeMask ) == STATE_BeUp );
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanClosed );

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_SendFeISDisconnected;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_SendFeMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_SendFeISDisconnected;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_FeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_FeIfDisconnected;
        }

        if
        (
            ( ( State & STATE_BeIfMask ) == STATE_BeIfConnecting )
            &&
            ( ( State & STATE_FeIfMask ) == STATE_FeIfConnectable )
            &&
            ( ( State & STATE_SendFeMask ) == STATE_SendFeIdle )
            &&
            ( ( State & STATE_DestroyMask ) == STATE_DestroyIdle )
        )
        {
            assert( ( State & STATE_FeMask ) == STATE_FeUp );
            assert( ( State & STATE_BeMask ) == STATE_BeUp );
/* FIXME           assert
            (
                ( ( State & STATE_DevicesMask ) == STATE_DevicesUp )
                ||
                ( ( State & STATE_DevicesMask ) == STATE_DevicesUping )
            ); */
            assert( ( State & STATE_EvtChanMask ) == STATE_EvtChanOpen );

            pEnr->ENR_Resp[ STIM_ber ][ rspidx++ ] = RESP_SendFeISConnected;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_SendFeMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_SendFeISConnected;

            pEnr->ENR_NextState[ STIM_ber ] &= ~STATE_FeIfMask;
            pEnr->ENR_NextState[ STIM_ber ] |=  STATE_FeIfConnected;
        }
    }
}

/* This is a simple function which prints out the enumeration in Funnelweb   */
/* form so it can be pasted into this file.                                  */

static void print_states( ENR * pEnrIn, int EnrCountIn )
{
    ENR * pEnr = pEnrIn;
    int EnrCount = EnrCountIn;
    int First = 1;

    while( EnrCount-- > 0 )
    {
        if( First )
        {
            First = 0;

            printf
            (
               "        if self.__state == '%s':\n",
                pEnr->ENR_Name
            );
        }
        else
        {
            printf
            (
               "        elif self.__state == '%s':\n",
                pEnr->ENR_Name
            );
        }

        switch( pEnr->ENR_ThisState & STATE_FeMask )
        {
          case STATE_FeDown:
            printf( "            # FeDown\n" );
            break;
          case STATE_FeUp:
            printf( "            # FeUp\n" );
            break;
          case STATE_FeDowning:
            printf( "            # FeDowning\n" );
            break;
        }

        switch( pEnr->ENR_ThisState & STATE_BeMask )
        {
          case STATE_BeDown:
            printf( "            # BeDown\n" );
            break;
          case STATE_BeUp:
            printf( "            # BeUp\n" );
            break;
          case STATE_BeDowning:
            printf( "            # BeDowning\n" );
            break;
        }

        switch( pEnr->ENR_ThisState & STATE_SendBeMask )
        {
          case STATE_SendBeIdle:
            printf( "            # SendBeIdle\n" );
            break;
          case STATE_SendBeCreate:
            printf( "            # SendBeCreate\n" );
            break;
          case STATE_SendBeDestroy:
            printf( "            # SendBeDestroy\n" );
            break;
          case STATE_SendBeConnect:
            printf( "            # SendBeConnect\n" );
            break;
          case STATE_SendBeDisconnect:
            printf( "            # SendBeDisconnect\n" );
            break;
          case STATE_SendBeDontCare:
            printf( "            # SendBeDontCare\n" );
            break;
        }

        switch( pEnr->ENR_ThisState & STATE_SendFeMask )
        {
          case STATE_SendFeIdle:
            printf( "            # SendFeIdle\n" );
            break;
          case STATE_SendFeISDisconnected:
            printf( "            # SendFeISDisconnected\n" );
            break;
          case STATE_SendFeISConnected:
            printf( "            # SendFeISConnected\n" );
            break;
          case STATE_SendFeISDestroyed:
            printf( "            # SendFeISDestroyed\n" );
            break;
          case STATE_SendFeDontCare:
            printf( "            # SendFeDontCare\n" );
            break;
        }

        switch( pEnr->ENR_ThisState & STATE_BeIfMask )
        {
          case STATE_BeIfDestroyed:
            printf( "            # BeIfDestroyed\n" );
            break;
          case STATE_BeIfCreating:
            printf( "            # BeIfCreating\n" );
            break;
          case STATE_BeIfDisconnected:
            printf( "            # BeIfDisconnected\n" );
            break;
          case STATE_BeIfConnecting:
            printf( "            # BeIfConnecting\n" );
            break;
          case STATE_BeIfConnected:
            printf( "            # BeIfConnected\n" );
            break;
          case STATE_BeIfDisconnecting:
            printf( "            # BeIfDisconnecting\n" );
            break;
          case STATE_BeIfDestroying:
            printf( "            # BeIfDestroying\n" );
            break;
        }

        switch( pEnr->ENR_ThisState & STATE_FeIfMask )
        {
          case STATE_FeIfDestroyed:
            printf( "            # FeIfDestroyed\n" );
            break;
          case STATE_FeIfDisconnected:
            printf( "            # FeIfDisconnected\n" );
            break;
          case STATE_FeIfConnectable:
            printf( "            # FeIfConnectable\n" );
            break;
          case STATE_FeIfConnected:
            printf( "            # FeIfConnected\n" );
            break;
        }

        switch( pEnr->ENR_ThisState & STATE_DevicesMask )
        {
          case STATE_DevicesDown:
            printf( "            # DevicesDown\n" );
            break;
          case STATE_DevicesUping:
            printf( "            # DevicesUping\n" );
            break;
          case STATE_DevicesUp:
            printf( "            # DevicesUp\n" );
            break;
          case STATE_DevicesDowning:
            printf( "            # DevicesDowning\n" );
            break;
        }

        switch( pEnr->ENR_ThisState & STATE_AddMask )
        {
          case STATE_AddClean:
            printf( "            # AddClean\n" );
            break;
          case STATE_AddDirty:
            printf( "            # AddDirty\n" );
            break;
        }

        switch( pEnr->ENR_ThisState & STATE_EvtChanMask )
        {
          case STATE_EvtChanClosed:
            printf( "            # EvtChanClosed\n" );
            break;
          case STATE_EvtChanOpen:
            printf( "            # EvtChanOpen\n" );
            break;
        }

        switch( pEnr->ENR_ThisState & STATE_DestroyMask )
        {
          case STATE_DestroyIdle:
            printf( "            # DestroyIdle\n" );
            break;
          case STATE_DestroyOutstanding:
            printf( "            # DestroyOutstanding\n" );
            break;
        }

        {
            int StimIndex;
            int First = 1;

            for( StimIndex = 0; StimIndex < STIM_Count; StimIndex++ )
            {
                if( pEnr->ENR_NextState[ StimIndex ] != STATE_Count )
                {
                    int RespIndex = 0, RespCount = 0;

                    while
                    (
                        pEnr->ENR_Resp[ StimIndex ][ RespCount ]
                        !=
                        RESP_Count
                    )
                    {
                        RespCount++;
                    }

                    if( First )
                    {
                        First = 0;

                        printf( "            if stimulus == " );
                    }
                    else
                    {
                        printf( "            elif stimulus == " );
                    }

                    switch( StimIndex )
                    {
                      case STIM_fdu:
                        printf( "'fdu'" );
                        break;
                      case STIM_fdd:
                        printf( "'fdd'" );
                        break;
                      case STIM_bdu:
                        printf( "'bdu'" );
                        break;
                      case STIM_bdd:
                        printf( "'bdd'" );
                        break;
                      case STIM_fic:
                        printf( "'fic'" );
                        break;
                      case STIM_dst:
                        printf( "'dst'" );
                        break;
                      case STIM_add:
                        printf( "'add'" );
                        break;
                      case STIM_nfy:
                        printf( "'nfy'" );
                        break;
                      case STIM_fer:
                        printf( "'fer'" );
                        break;
                      case STIM_ber:
                        printf( "'ber'" );
                        break;
                    }

                    printf( ":\n" );

                    if
                    (
                        pEnr->ENR_NextState[ StimIndex ]
                        !=
                        pEnr->ENR_ThisState
                    )
                    {
                        int EnrIndex;

                        for( EnrIndex = 0; EnrIndex < EnrCountIn; EnrIndex++ )
                        {
                            if
                            (
                                pEnr->ENR_NextState[ StimIndex ]
                                ==
                                pEnrIn[ EnrIndex ].ENR_ThisState
                            )
                            {
                                printf
                                (
                                    "                self.__state = '%s'\n",
                                    pEnrIn[ EnrIndex ].ENR_Name
                                );

                                break;
                            }
                        }
                    }
                    else if( RespCount == 0 )
                    {
                        printf( "                pass\n" );
                    }

                    for( RespIndex = 0; RespIndex < RespCount; RespIndex++ )
                    {
                        switch( pEnr->ENR_Resp[ StimIndex ][ RespIndex ] )
                        {
                          case RESP_NotifyDevicesUp:
                            printf( "                self.__notify_devices('up')\n" );
                            break;
                          case RESP_NotifyDevicesDown:
                            printf( "                self.__notify_devices('down')\n" );
                            break;
                          case RESP_CompleteFeDown:
                            printf( "                self.__complete_fe_down()\n" );
                            break;
                          case RESP_CompleteBeDown:
                            printf( "                self.__complete_be_down()\n" );
                            break;
                          case RESP_CompleteDestroy:
                            printf( "                self.__complete_destroy()\n" );
                            break;
                          case RESP_OpenEvtChan:
                            printf( "                self.__open_event_channel()\n" );
                            break;
                          case RESP_CloseEvtChan:
                            printf( "                self.__close_event_channel()\n" );
                            break;
                          case RESP_SendBeCreate:
                            printf( "                self.__send_be_create()\n" );
                            break;
                          case RESP_SendBeDestroy:
                            printf( "                self.__send_be_destroy()\n" );
                            break;
                          case RESP_SendBeConnect:
                            printf( "                self.__send_be_connect()\n" );
                            break;
                          case RESP_SendBeDisconnect:
                            printf( "                self.__send_be_disconnect()\n" );
                            break;
                          case RESP_SendFeISDisconnected:
                            printf( "                self.__send_fe_is_disconnected()\n" );
                            break;
                          case RESP_SendFeISConnected:
                            printf( "                self.__send_fe_is_connected()\n" );
                            break;
                          case RESP_SendFeISDestroyed:
                            printf( "                self.__send_fe_is_destroyed()\n" );
                            break;
                        }
                    }
                }
            }

            if( First )
            {
                printf( "            self.__stimulus_error(stimulus)\n" );
            }
            else
            {
                printf( "            else:\n                self.__stimulus_error(stimulus)\n" );
            }
        }

        pEnr++;
    }

    printf( "        else:\n            self.__stimulus_error(stimulus)\n" );
}

int main( int argc, char ** argv, char ** envp )
{
    ENR Enr[ 1024 ];

    int EnrIndex;

    /* Start with one state: i. */

    int EnrCount = 1;

    new_state( &Enr[ 0 ], 0, "i" );

    /* Fill in states in the table one by one. */

    for( EnrIndex = 0; EnrIndex < EnrCount; EnrIndex++ )
    {
        ENR * pEnr = &Enr[ EnrIndex ];

        int StimIndex;

        /* If the enumeration gets too big and we run out of states, there   */
        /* is probably a bug in the new_state function.                      */

        if( EnrCount == 1024 )
        {
            printf( "FAILED\n" );

            print_states( Enr, EnrCount );

            printf( "FAILED\n" );

            return 1;
        }

        /* Iterate over all stimuli. */

        for( StimIndex = 0; StimIndex < STIM_Count; StimIndex++ )
        {
            /* If the current state makes a transition for that stimulus... */

            if( pEnr->ENR_NextState[ StimIndex ] != STATE_Count )
            {
                int EnrSearch;

                /* ...see if the next state is equivalent to any of the      */
                /* states we have already enumerated.                        */

                for
                (
                    EnrSearch = 0;
                    EnrSearch < EnrCount;
                    EnrSearch++
                )
                {
                    if
                    (
                        Enr[ EnrSearch ].ENR_ThisState
                        ==
                        pEnr->ENR_NextState[ StimIndex ]
                    )
                    {
                        break;
                    }
                }

                /* If there was no equivalent... */

                if( EnrSearch == EnrCount )
                {
                    char Name[ 1024 ];

                    strncpy( Name, pEnr->ENR_Name, 1023 );

                    Name[ 1023 ] = 0;

                    switch( StimIndex )
                    {
                      case STIM_fdu:
                        strcat( Name, "_fdu" );
                        break;
                      case STIM_fdd:
                        strcat( Name, "_fdd" );
                        break;
                      case STIM_bdu:
                        strcat( Name, "_bdu" );
                        break;
                      case STIM_bdd:
                        strcat( Name, "_bdd" );
                        break;
                      case STIM_fic:
                        strcat( Name, "_fic" );
                        break;
                      case STIM_dst:
                        strcat( Name, "_dst" );
                        break;
                      case STIM_add:
                        strcat( Name, "_add" );
                        break;
                      case STIM_nfy:
                        strcat( Name, "_nfy" );
                        break;
                      case STIM_fer:
                        strcat( Name, "_fer" );
                        break;
                      case STIM_ber:
                        strcat( Name, "_ber" );
                        break;
                    }

                    printf( "%s\n", Name );

                    /* ...create the new state. */

                    new_state
                    (
                        &Enr[ EnrCount++ ],
                        pEnr->ENR_NextState[ StimIndex ],
                        Name
                    );
                }
            }
        }
    }

    /* We'll only get here once there are no transitions which map to states */
    /* that haven't been enumerated yet.                                     */

    printf( "SUCCESS\n" );

    print_states( Enr, EnrCount );

    printf( "SUCCESS\n" );

    return 0;
}
