#  Support for the virtualising split USB driver
#
#  Copyright (c) 2005 Harry Butterworth IBM Corporation
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or (at your
#  option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# Based on usbif.py, original copyright notice follows:

# Copyright (C) 2004 Mike Wray <mike.wray@hp.com>
# Copyright (C) 2004 Intel Research Cambridge
# Copyright (C) 2004 Mark Williamson <mark.williamson@cl.cam.ac.uk>

"""Support for virtual USB hubs.
"""

from twisted.internet import defer

from xen.xend import sxp
from xen.xend.XendLogging import log
from xen.xend.XendError import XendError, VmError
from xen.xend.XendRoot import get_component

import channel
import controller
from messages import *

class UsbifControllerFactory(controller.SplitControllerFactory):

    def __init__(self):
        log.debug("UsbifControllerFactory")
        controller.SplitControllerFactory.__init__(self)

    def createController(self, dom, recreate=0):
        log.debug("UsbifControllerFactory.createController")
        if recreate:
            raise NotImplementedError()
        return UsbifController(self, dom)

    def createBackendController(self, dom):
        log.debug("UsbifControllerFactory.createBackendController")
        return UsbifBackendController(self, dom)

    def createBackendInterface(self, feController, beDomain, handle):
        log.debug("UsbifControllerFactory.createBackendInterface")
        return UsbifBackendInterface(feController, beDomain, handle)

class UsbifController(controller.SplitController):

    def __init__(self, factory, dom):
        log.debug("UsbifController")
        controller.SplitController.__init__(self, factory, dom)
        self.__state = 'i'
        self.addMethod(CMSG_USBIF_FE,
                       CMSG_USBIF_FE_DRIVER_STATUS_CHANGED,
                       self.__recv_fe_driver_status_changed)
        self.addMethod(CMSG_USBIF_FE,
                       CMSG_USBIF_FE_INTERFACE_CONNECT,
                       self.__recv_fe_interface_connect)
        self.addMethod(CMSG_USBIF_FE,
                       CMSG_USBIF_FE_INTERFACE_DISCONNECT,
                       self.__recv_fe_interface_disconnect)
        self.registerChannel()

    def attachDevice(self, idx, config, recreate=0):
        log.debug("UsbifController.attachDevice")
        if recreate:
            raise NotImplementedError()
        if idx in self.devices:
            raise XendError('device exists: ' + str(idx))
        dev = UsbDev(idx, self, config)
        self.devices[idx] = dev
        return defer.succeed(dev)

    def destroy(self):
        log.debug("UsbifController.destroy")
        self.__destroy_devices_outstanding = 1
        for dev in self.getDevices():
            self.__destroy_devices_outstanding = \
              self.__destroy_devices_outstanding + 1
            deferred = defer.Deferred()
            deferred.addCallback(self.__destroy_1)
            dev.destroy( deferred )

    def __destroy_1(self, ignored=0):
        log.debug("UsbifController.__destroy_1")
        self.__destroy_devices_outstanding = \
          self.__destroy_devices_outstanding - 1
        if self.__destroy_devices_outstanding == 0:
            self.__destroy_interfaces_outstanding = 1
            for iface in self.getBackendInterfaces():
                self.__destroy_interfaces_outstanding = \
                  self.__destroy_interfaces_outstanding + 1
                deferred = defer.Deferred()
                deferred.addCallback(self.__destroy_2)
                iface.destroy( deferred )

    def __destroy_2(self, ignored=0):
        log.debug("UsbifController.__destroy_2")
        self.__destroy_interfaces_outstanding = \
          self.__destroy_interfaces_outstanding - 1
        if self.__destroy_interfaces_outstanding == 0:
            self.close()

    def __recv_fe_driver_status_changed(self, msg, req):
        log.debug("UsbifController.__recv_fe_driver_status_changed")
        status = unpackMsg('usbif_fe_driver_status_changed_t', msg)['status']
        if status:
            self.__stimulus('fdu')
            return 0
        else:
            self.__msg = msg
            self.__stimulus('fdd')
            return 1

    def __recv_fe_interface_connect(self, msg, req):
        log.debug("UsbifController.__recv_fe_interface_connect")
        val = unpackMsg('usbif_fe_interface_connect_t', msg)
        handle = val['handle']
        backendInterface = self.getBackendInterfaceByHandle(handle)
        if backendInterface:
            backendInterface.fe_interface_connect(val)
        else:
            log.error('connect for unknown interface: handle=%d', handle)

    def __recv_fe_interface_disconnect(self, msg, req):
        log.debug("UsbifController.__recv_fe_interface_disconnect")
        handle = unpackMsg('usbif_fe_interface_disconnect_t', msg)['handle']
        backendInterface = self.getBackendInterfaceByHandle(handle)
        if backendInterface:
            backendInterface.fe_interface_disconnect()
        else:
            log.error('disconnect for unknown interface: handle=%d', handle)

    def __stimulus(self, stimulus):
        log.debug("UsbifController.__stimulus")
        log.debug("state:%s stimulus:%s", self.__state, stimulus)
        if self.__state == 'i':
            if stimulus == 'fdu':
                self.__state = 'i_fdu'
                self.__notify_up()
            else:
                log.error("state:%s stimulus:%s", self.__state, stimulus)
        elif self.__state == 'i_fdu':
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd'
                self.__notify_down()
            else:
                log.error("state:%s stimulus:%s", self.__state, stimulus)
        elif self.__state == 'i_fdu_fdd':
            if stimulus == 'ndc':
                self.__state = 'i'
                self.__acknowledge_fe_driver_down()
            else:
                log.error("state:%s stimulus:%s", self.__state, stimulus)
        else:
            log.error("state:%s stimulus:%s", self.__state, stimulus)

    def __notify_up(self):
        log.debug("UsbifController.__notify_up")
        for backendInterface in self.getBackendInterfaces():
            backendInterface.fe_up()

    def __notify_down(self):
        log.debug("UsbifController.__notify_down")
        self.__fe_down_outstanding = 1
        for backendInterface in self.getBackendInterfaces():
            self.__fe_down_outstanding = self.__fe_down_outstanding + 1
            deferred = defer.Deferred()
            deferred.addCallback(self.__notify_down_1)
            backendInterface.fe_down(deferred)
        self.__notify_down_1()

    def __notify_down_1(self, ignored=0):
        log.debug("UsbifController.__notify_down_1")
        self.__fe_down_outstanding = self.__fe_down_outstanding - 1
        if self.__fe_down_outstanding == 0:
            self.__stimulus('ndc')

    def __acknowledge_fe_driver_down(self):
        log.debug("UsbifController.__acknowledge_fe_driver_down")
        self.writeResponse(self.__msg)
        self.__msg = None

    def getBackendInterface(self, dom):
        log.debug("UsbifController.getBackendInterface")
        b = self.getBackendInterfaceByDomain(dom)
        if b is None:
            b = controller.SplitController.getBackendInterface(self, dom)
            if self.__state == 'i_fdu':
                backendInterface.fe_up()
        return b

class UsbifBackendController(controller.BackendController):

    def __init__(self, factory, dom):
        log.debug("UsbifBackendController")
        controller.BackendController.__init__(self, factory, dom)
        self.__state = 'i'
        self.addMethod(CMSG_USBIF_BE,
                       CMSG_USBIF_BE_DRIVER_STATUS_CHANGED,
                       self.__recv_be_driver_status_changed)
        self.registerChannel()

    def __recv_be_driver_status_changed(self, msg, req):
        log.debug("UsbifBackendController.__recv_be_driver_status_changed")
        status = unpackMsg('usbif_be_driver_status_changed_t', msg)['status']
        if status:
            self.__stimulus('bdu')
            return 0
        else:
            self.__msg = msg
            self.__stimulus('bdd')
            return 1

    def __stimulus(self, stimulus):
        log.debug("UsbifBackendController.__stimulus")
        log.debug("state:%s stimulus:%s", self.__state, stimulus)
        if self.__state == 'i':
            if stimulus == 'bdu':
                self.__state = 'i_bdu'
                self.__notify_up()
            else:
                log.error("state:%s stimulus:%s", self.__state, stimulus)
        elif self.__state == 'i_bdu':
            if stimulus == 'bdd':
                self.__state = 'i_bdu_bdd'
                self.__notify_down()
            else:
                log.error("state:%s stimulus:%s", self.__state, stimulus)
        elif self.__state == 'i_bdu_bdd':
            if stimulus == 'ndc':
                self.__state = 'i'
                self.__acknowledge_be_driver_down()
            else:
                log.error("state:%s stimulus:%s", self.__state, stimulus)
        else:
            log.error("state:%s stimulus:%s", self.__state, stimulus)

    def __notify_up(self):
        log.debug("UsbifBackendController.__notify_up")
        for backendInterface in self.backendInterfaces.values():
            backendInterface.be_up()

    def __notify_down(self):
        log.debug("UsbifBackendController.__notify_down")
        self.__be_down_outstanding = 1
        for backendInterface in self.backendInterfaces.values():
            self.__be_down_outstanding = self.__be_down_outstanding + 1
            deferred = defer.Deferred()
            deferred.addCallback(self.__notify_down_1)
            backendInterface.be_down(deferred)
        self.__notify_down_1()

    def __notify_down_1(self, ignored=0):
        log.debug("UsbifBackendController.__notify_down_1")
        self.__be_down_outstanding = self.__be_down_outstanding - 1
        if self.__be_down_outstanding == 0:
            self.__stimulus('ndc')

    def __acknowledge_be_driver_down(self):
        log.debug("UsbifController.__acknowledge_be_driver_down")
        self.writeResponse(self.__msg)
        self.__msg = None

    def registerInterface(self, backendInterface):
        log.debug("UsbifBackendController.registerInterface")
        controller.BackendController.registerInterface(self, backendInterface)
        if self.__state == 'i_bdu':
            backendInterface.be_up()

    def deregisterInterface(self, backendInterface):
        log.debug("UsbifBackendController.deregisterInterface")
        controller.BackendController.deregisterInterface\
          (self, backendInterface)
        if not self.backendInterfaces:
            log.debug("UsbifBackendController closing")
            self.close()

class UsbifBackendInterface(controller.BackendInterface):

    def __init__(self, feController, beDomain, handle):
        log.debug("UsbifBackendInterface")
        controller.BackendInterface.__init__\
          (self, feController, beDomain, handle)
        self.__state = 'i'
        self.__devices = {}

    def fe_up(self):
        log.debug("UsbifBackendInterface.fe_up")
        self.__stimulus('fdu')

    def fe_interface_connect(self, val):
        log.debug("UsbifBackendInterface.fe_interface_connect")
        self.__shmem_frame = val['shmem_frame']
        self.__stimulus('fic')

    def fe_interface_disconnect(self):
        log.debug("UsbifBackendInterface.fe_interface_disconnect")
        self.__stimulus('fid')

    def fe_down(self, deferred):
        log.debug("UsbifBackendInterface.fe_down")
        self.__fe_deferred = deferred
        self.__stimulus('fdd')

    def be_up(self):
        log.debug("UsbifBackendInterface.be_up")
        self.__stimulus('bdu')

    def be_down(self, deferred):
        log.debug("UsbifBackendInterface.be_down")
        self.__be_deferred = deferred
        self.__stimulus('bdd')

    def destroy(self, deferred):
        log.debug("UsbifBackendInterface.destroy")
        self.__destroy_deferred = deferred
        if self.__devices:
            raise NotImplementedError()
        self.__stimulus('dst')

    # add and remove provided for UsbDev

    def add_device(self, device, idx):
        log.debug("UsbifBackendInterface.add_device")
        self.__devices[idx]=device
        self.__stimulus('add')

    def remove_device(self, device, idx):
        log.debug("UsbifBackendInterface.remove_device")
        del self.__devices[idx]

    # State machine stimuli:
    #
    # fdu: frontend driver up
    # fic: frontend interface connect
    # fid: frontend interface disconnect
    # fdd: frontend driver down
    # bdu: backend driver up
    # bdd: backend driver down
    # dst: destroy (no devices)
    # add: device add
    # nfy: notify done
    # rsp: response received
    #
    # State machine responses:
    #
    # __notify_devices up/down  : nfy
    # __complete_fe_down
    # __complete_be_down
    # __complete_destroy
    # __open_event_channel
    # __close_event_channel
    # __send_be_create          : rsp
    # __send_be_destroy         : rsp
    # __send_be_connect         : rsp
    # __send_be_disconnect      : rsp
    # __send_fe_is_disconnected : rsp
    # __send_fe_is_connected    : rsp
    # __send_fe_is_destroyed    : rsp

    # When a device is attached:
    #
    # add
    # __notify_devices up/down

    # Establishing the connection:
    #
    # bdu (backend driver up)
    # __send_be_create
    # __notify_devices up
    # fdu (frontend driver up)
    # __send_fe_is_disconnected
    # fic (frontend interface connect)
    # __send_be_connect
    # rsp
    # __send_fe_is_connected

    # Back-end driver module unload whilst front-end is up
    #
    # bdd (backend driver down)
    # __send_fe_is_disconnected

    # Front-end driver module unload whilst back-end is up
    #
    # fid (frontend interface disconnect)
    # __send_be_disconnect
    # rsp
    # __send_fe_is_disconnected
    # fdd (frontend driver down)

    # Destroy whilst back end and front end are up
    #
    # dst
    # __send_be_disconnect
    # rsp
    # __send_be_destroy
    # __send_fe_is_destroyed
    # __complete_destroy

    def __stimulus_error(self, stimulus):
        log.debug( "USB BE interface in state %s received unexpected stimulus %s", self.__state, stimulus )

    def __stimulus(self, stimulus):
        log.debug("UsbifBackendInterface.__stimulus")
        log.debug("state:%s stimulus:%s", self.__state, stimulus)

        if self.__state == 'i':
            # FeDown
            # BeDown
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdu':
                self.__state = 'i_bdu'
            elif stimulus == 'dst':
                self.__complete_destroy()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd'
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_bdu'
                self.__send_be_create()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_bdu':
            # FeDown
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu'
                self.__send_be_create()
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i'
                self.__complete_be_down()
            elif stimulus == 'dst':
                self.__complete_destroy()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_fdd':
            # FeDowning
            # BeDown
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdu':
                self.__state = 'i_fdu_fdd_bdu'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_fdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeISDisconnected
            # BeIfCreating
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber'
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_fic':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd'
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_bdu_fic'
                self.__send_be_create()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_fic_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_fic_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd_dst'
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_dst_bdu'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_fic_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_dst_fer'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_fer':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i'
                self.__complete_fe_down()
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_bdu_fer'
                self.__send_be_create()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_fic_fer'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_fdd_bdu':
            # FeDowning
            # BeUp
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_fdd'
                self.__complete_be_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_fdd_bdu_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_bdu'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_fdd_dst':
            # FeDowning
            # BeDown
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdu':
                self.__state = 'i_fdu_fdd_bdu_dst'
            elif stimulus == 'fer':
                self.__state = 'i'
                self.__complete_fe_down()
                self.__complete_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd':
            # FeDowning
            # BeUp
            # SendBeCreate
            # SendFeDontCare
            # BeIfCreating
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fdd_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_bdd_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_bdd_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeISDisconnected
            # BeIfCreating
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber'
                self.__notify_devices('up')
                self.__open_event_channel()
                self.__send_be_connect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_dst':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeISDisconnected
            # BeIfCreating
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_dst'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_dst_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeIdle
            # BeIfCreating
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fer'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_fer'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber'
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_ber'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_dst'
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_ber_add'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_fic_dst':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd_dst'
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_fic_dst_bdu'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_dst_fer'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_fic_fer':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i'
                self.__complete_fe_down()
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_bdu_fic_fer'
                self.__send_be_create()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_bdu':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd_bdu_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_dst'
                self.__complete_be_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_fic_dst_bdu'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_dst_bdu_fer'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeISDestroyed
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd_dst'
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_dst_bdu_fer'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_dst_fer_fer'
                self.__complete_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_fdd_bdu_dst':
            # FeDowning
            # BeUp
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_fdd_dst'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_bdu'
                self.__complete_fe_down()
                self.__complete_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_bdd':
            # FeDowning
            # BeDowning
            # SendBeDontCare
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fdd_bdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_fdd'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_dst':
            # FeDowning
            # BeUp
            # SendBeCreate
            # SendFeDontCare
            # BeIfCreating
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fdd_dst_fer'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_fer':
            # FeDown
            # BeUp
            # SendBeCreate
            # SendFeIdle
            # BeIfCreating
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_fer'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fdd_dst_fer'
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fdd_fer_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_ber':
            # FeDowning
            # BeUp
            # SendBeDestroy
            # SendFeDontCare
            # BeIfDestroying
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fdd_fer_ber'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_fdd_bdu'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd_fic':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_bdd_fic_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_bdd_fic_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_fic'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd_dst':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_bdd_fic_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_bdd_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_dst'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_bdd_fic_fer'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_bdd_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_fer'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_dst':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeISDisconnected
            # BeIfCreating
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_dst_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeIdle
            # BeIfCreating
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic_fer'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber'
                self.__notify_devices('up')
                self.__open_event_channel()
                self.__send_be_connect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_dst'
                self.__close_event_channel()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_ber_add'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_dst_fer':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeISDestroyed
            # BeIfCreating
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_dst_fer'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_dst_fer_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_dst_fer_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_dst_ber':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeISDisconnected
            # BeIfDestroying
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_dst'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_dst_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_dst_fer_ber'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_dst_bdu'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_fer_ber'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fer_ber_add'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd':
            # FeDowning
            # BeUp
            # SendBeIdle
            # SendFeDontCare
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_fdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_ber_bdd_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_dst':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd_dst'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_ber_dst_fic'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_dst'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_add':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesUping
            # AddDirty
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_ber_add'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_add'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_nfy_fdd'
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd'
                self.__notify_devices('down')
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_ber_nfy'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_ber'
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_fic_dst_bdu':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd_bdu_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_fic_dst'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_dst_bdu_fer'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_bdu_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDestroyed
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd_bdu_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_dst_fer'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_dst_bdu_fer_fer'
                self.__complete_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer_fer':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i'
                self.__complete_fe_down()
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_dst_fer_fer_bdu'
                self.__send_be_create()
            elif stimulus == 'dst':
                self.__complete_destroy()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_bdd_dst':
            # FeDowning
            # BeDowning
            # SendBeDontCare
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_fdd_dst'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_bdd_fer':
            # FeDown
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_bdd'
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer'
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_dst_fer':
            # FeDown
            # BeUp
            # SendBeCreate
            # SendFeIdle
            # BeIfCreating
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fdd_dst_fer_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_dst_ber':
            # FeDowning
            # BeUp
            # SendBeDestroy
            # SendFeDontCare
            # BeIfDestroying
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fdd_dst_fer_ber'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_fdd_bdu_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_fer_ber':
            # FeDown
            # BeUp
            # SendBeDestroy
            # SendFeIdle
            # BeIfDestroying
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_fdd_fer_ber_fdu'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_fer'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fdd_dst_fer_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_bdu'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd_fic_dst':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_bdd_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_fic_dst'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd_fic_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_bdd_dst_fer'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_fic_fer'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd_dst_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDestroyed
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_bdd_dst_fer_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_dst_fer'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_dst_ber':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeISDisconnected
            # BeIfDestroying
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_dst_fer_ber'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_fic_dst_bdu'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst'
                self.__close_event_channel()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_fer_ber_add'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeDontCare
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd_fic'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_dst':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_dst_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_add':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUping
            # AddDirty
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_dst'
                self.__close_event_channel()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_add'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_add_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_nfy':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_bdd'
                self.__close_event_channel()
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst'
                self.__close_event_channel()
                self.__notify_devices('down')
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_ber'
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfConnected
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd_fic'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_dst_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_ber_add_ber'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_dst_fer_fer':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeIdle
            # BeIfCreating
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst_fer'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_dst_fer_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_dst_fer_fer_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_dst_fer_ber':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeISDestroyed
            # BeIfDestroying
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_dst_fer'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_dst_fer_fer_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_dst_bdu_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd':
            # FeDown
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd'
                self.__complete_fe_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_dst':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDestroyed
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_dst_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_add':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesUping
            # AddDirty
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_fer_ber_add'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd'
                self.__complete_fe_down()
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_bdd'
                self.__notify_devices('down')
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_dst'
                self.__notify_devices('down')
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fer_ber'
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_bdd':
            # FeDowning
            # BeDowning
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_dst':
            # FeDowning
            # BeUp
            # SendBeIdle
            # SendFeDontCare
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_fdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_dst'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_nfy':
            # FeDowning
            # BeUp
            # SendBeIdle
            # SendFeDontCare
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_bdd'
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd_fic':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_fic'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd_dst':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_nfy'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_dst_fic':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanOpen
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_dst'
                self.__close_event_channel()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_dst_fic_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_dst_fic_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_dst_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_dst_nfy':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_ber_dst_fic_nfy'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_dst_nfy'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy_fdd':
            # FeDowning
            # BeUp
            # SendBeIdle
            # SendFeDontCare
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_nfy_fdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy_bdd':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy_dst':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy_dst'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_ber_nfy_dst_fic'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_dst_ber'
                self.__send_be_destroy()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_dst'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_bdu_fer_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_bdu'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_dst_fer_fer'
                self.__complete_be_down()
            elif stimulus == 'dst':
                self.__complete_destroy()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer_fer_bdu':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeIdle
            # BeIfCreating
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_dst_fer_fer_bdu_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_dst_fer_fer'
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_dst_fer_fer_bdu_ber'
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_bdd_dst_fer':
            # FeDown
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_bdd_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer_ber'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_dst_fer_ber':
            # FeDown
            # BeUp
            # SendBeDestroy
            # SendFeIdle
            # BeIfDestroying
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_dst_ber'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer'
            elif stimulus == 'ber':
                self.__state = 'i_bdu'
                self.__complete_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_fer_ber_fdu':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeISDisconnected
            # BeIfDestroying
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fdd_fer_ber_fdu_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_dst_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fdd_fer_ber_fdu_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu'
                self.__send_be_create()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd_dst_fer_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_bdd_dst_fer_fer_ber'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_fic'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_bdd'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_add':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUping
            # AddDirty
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst'
                self.__close_event_channel()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__notify_devices('up')
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_add_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_bdd'
                self.__close_event_channel()
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst'
                self.__close_event_channel()
                self.__notify_devices('down')
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_fer_ber'
                self.__notify_devices('up')
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfConnected
            # FeIfConnected
            # DevicesUping
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_fer_ber_add_ber'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_bdd':
            # FeDowning
            # BeDowning
            # SendBeDontCare
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_dst':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeDontCare
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_nfy':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeDontCare
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_bdd'
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeDontCare
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_bdd_dst':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_dst_nfy':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_nfy'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_dst_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_dst_ber':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeISDisconnected
            # BeIfDisconnecting
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_dst_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_dst_ber_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_add_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfConnected
            # FeIfConnectable
            # DevicesUping
            # AddDirty
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd_fic'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_dst_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_add_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_nfy_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_bdd_fic'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_bdd'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd_fic'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_nfy_dst':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_nfy_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfConnected
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd_fic'
                self.__close_event_channel()
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_ber'
                self.__close_event_channel()
                self.__notify_devices('down')
                self.__send_be_disconnect()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_ber_ber'
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_dst_fer_fer_ber':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeIdle
            # BeIfDestroying
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst_fer_ber'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_dst_fer_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_dst_bdu_fer_fer'
                self.__complete_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_bdd':
            # FeDown
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_bdd'
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_dst':
            # FeDown
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_dst_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_nfy':
            # FeDown
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_nfy'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_bdd'
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd_fic':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_fic_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd_dst':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDestroyed
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy'
                self.__complete_fe_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_fic_nfy'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst'
                self.__notify_devices('down')
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_dst_nfy':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDestroyed
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_dst_nfy_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_dst_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_dst'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_fer'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_dst_nfy_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd':
            # FeDown
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_bdd':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_bdd'
                self.__complete_fe_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_bdd_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_fer'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_dst':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDestroyed
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_dst_fer_ber'
                self.__send_be_destroy()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_dst_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_bdd_dst':
            # FeDowning
            # BeDowning
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_bdd_nfy':
            # FeDowning
            # BeDowning
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_dst_nfy':
            # FeDowning
            # BeUp
            # SendBeIdle
            # SendFeDontCare
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_dst_nfy'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_nfy_bdd':
            # FeDowning
            # BeDowning
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_fdd'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_bdd'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_nfy_dst':
            # FeDowning
            # BeUp
            # SendBeIdle
            # SendFeDontCare
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
                self.__send_be_destroy()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_dst'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd_fic_dst':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd_fic_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_fic_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd_dst_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_nfy'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd_nfy_dst':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_dst'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_dst_fic_nfy':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanOpen
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_nfy'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy_dst'
                self.__close_event_channel()
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_dst_fic_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_dst_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_dst_fic_fer':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanOpen
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_bdd'
                self.__close_event_channel()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_dst_fic_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy_fdd_nfy':
            # FeDowning
            # BeUp
            # SendBeIdle
            # SendFeDontCare
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_fdd'
                self.__complete_be_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
                self.__send_be_destroy()
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_nfy'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy_bdd_fic':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_fic'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_bdd_fic'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy_dst_fic':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanOpen
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy_dst'
                self.__close_event_channel()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_nfy_dst_fic_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_nfy_dst_fic_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer_fer_bdu_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_bdd_dst_fer_fer'
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_dst_fer_fer'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer_fer_bdu_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_dst_fer_fer_bdu_ber_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_dst_fer'
            elif stimulus == 'add':
                self.__state = 'i_fdu_dst_fer_fer_bdu_ber_add'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_dst_fer_fer_bdu_ber_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_bdd_dst_fer_ber':
            # FeDown
            # BeDown
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer_ber_bdu'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_fer_ber_fdu_fic':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeISDisconnected
            # BeIfDestroying
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_dst_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fdd_fer_ber_fdu_fic_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic'
                self.__send_be_create()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_fer_ber_fdu_fer':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeIdle
            # BeIfDestroying
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_fer_ber'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fer'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fdd_fer_ber_fdu_fic_fer'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_dst_fer_ber'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer'
                self.__send_be_create()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd_dst_fer_fer_ber':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer_ber'
                self.__complete_fe_down()
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_bdu_bdd_dst_fer_fer_ber_bdu'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd':
            # FeDown
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu'
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_dst':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_nfy':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd'
                self.__notify_devices('down')
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_bdd_dst':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDestroyed
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy_dst'
                self.__notify_devices('down')
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_fic_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_bdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_bdd_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_nfy':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd'
                self.__notify_devices('down')
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_ber':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_bdd'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_ber_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_add_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfConnected
            # FeIfConnected
            # DevicesUping
            # AddDirty
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_add_ber_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_bdd_fic_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_bdd_fic'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_dst':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfConnected
            # FeIfConnected
            # DevicesUp
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd'
                self.__close_event_channel()
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst'
                self.__close_event_channel()
                self.__notify_devices('down')
                self.__send_be_disconnect()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber'
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_bdd':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_bdd'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_dst':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeISConnected
            # BeIfDisconnecting
            # FeIfConnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_bdd'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfConnected
            # FeIfConnected
            # DevicesUping
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd'
                self.__close_event_channel()
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_fer'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_fer_ber_add_ber_fer'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_bdd_dst':
            # FeDowning
            # BeDowning
            # SendBeDontCare
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_bdd_nfy':
            # FeDowning
            # BeDowning
            # SendBeDontCare
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_dst_nfy':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeDontCare
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_dst_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeDontCare
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_nfy_bdd':
            # FeDowning
            # BeDowning
            # SendBeDontCare
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fdd_bdd'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_nfy_dst':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeDontCare
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_nfy_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeDontCare
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_bdd'
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_ber'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_bdd_dst_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst_nfy'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_dst_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_bdd_nfy_dst':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_bdd_fic_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_nfy_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_dst_nfy_ber':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeISDisconnected
            # BeIfDisconnecting
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_nfy_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_dst_nfy_ber_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_dst_ber_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_dst_nfy_ber_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_dst'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_nfy_dst_nfy':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_nfy_dst_ber':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeISDisconnected
            # BeIfDisconnecting
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_ber_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_bdd_dst':
            # FeDown
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_bdd_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_bdd_nfy':
            # FeDown
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy'
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_dst_nfy':
            # FeDown
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_dst_nfy'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_nfy_bdd':
            # FeDown
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd'
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_nfy_dst':
            # FeDown
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_nfy_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fdd_dst_fer_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd_fic_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst'
                self.__notify_devices('down')
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd_dst_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDestroyed
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_nfy_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd_dst_fer':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_nfy_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd_nfy_dst':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDestroyed
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_dst_fer'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_dst_nfy_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_dst'
                self.__complete_fe_down()
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst_fer'
                self.__notify_devices('down')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_nfy_fdd'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_ber_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_nfy':
            # FeDown
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_nfy'
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i'
                self.__complete_be_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fdd_dst_fer_ber'
                self.__send_be_destroy()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_bdd_fic':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_fic_fer'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_dst_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_dst'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst_fer'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_dst_fer_fer_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_bdd_dst_nfy':
            # FeDowning
            # BeDowning
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst_nfy'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_fdd_bdd_nfy_dst':
            # FeDowning
            # BeDowning
            # SendBeIdle
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'nfy':
                self.__state = 'i_fdu_fdd_dst'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy_dst'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd_fic_dst_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_nfy'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_bdd_fic_nfy_dst':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_fic_dst'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_dst_fic_nfy_fer':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanOpen
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd'
                self.__close_event_channel()
                self.__notify_devices('down')
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy_dst_fic_nfy':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanOpen
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic_dst'
                self.__close_event_channel()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_nfy_dst_fic_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy_dst_fic_fer':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanOpen
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd'
                self.__close_event_channel()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_ber_nfy_dst_fic_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer_fer_bdu_ber_bdd':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_fer'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_dst_fer_fer_bdu_ber_bdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer_fer_bdu_ber_add':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUping
            # AddDirty
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_dst_fer_fer_bdu_ber_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_dst_fer'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer_fer_bdu_ber_nfy':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd'
                self.__complete_fe_down()
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_dst_fer_fer_bdu_ber_nfy_bdd'
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_dst_fer'
                self.__notify_devices('down')
            elif stimulus == 'add':
                self.__state = 'i_fdu_dst_fer_fer_bdu_ber'
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_bdd_dst_fer_ber_bdu':
            # FeDown
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_dst'
                self.__send_be_create()
                self.__send_fe_is_disconnected()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer_ber'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fdd_fer_ber_fdu_fic_fer':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeIdle
            # BeIfDestroying
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_fer_ber'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic_fer'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_dst_fer_ber'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer'
                self.__send_be_create()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_bdd_dst_fer_fer_ber_bdu':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer_ber_bdu'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_dst_fer_fer_ber'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_ber_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst':
            # FeDown
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy':
            # FeDown
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy'
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_dst_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_dst'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd':
            # FeDown
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd_fdu'
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fdd_bdd_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_nfy_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd'
                self.__notify_devices('down')
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_ber'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_bdd_dst_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDestroyed
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_bdd_dst_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_bdd_nfy_dst':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDestroyed
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_bdd_dst_fer'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy_dst_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst_nfy'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_bdd_nfy_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_bdd_ber':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_bdd_nfy_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_nfy_ber':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd'
                self.__notify_devices('down')
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_ber_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_ber_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfConnectable
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_dst'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_bdd_ber'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_ber_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_add_ber_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfConnected
            # FeIfConnected
            # DevicesUping
            # AddDirty
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd'
                self.__close_event_channel()
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_fer'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_dst_ber':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_ber_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeISConnected
            # BeIfDisconnecting
            # FeIfConnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfConnected
            # FeIfConnected
            # DevicesUp
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd'
                self.__close_event_channel()
                self.__notify_devices('down')
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_fer'
                self.__close_event_channel()
                self.__notify_devices('down')
                self.__send_be_disconnect()
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_fer'
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_bdd_dst':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_bdd_dst'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_dst_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_dst_nfy':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeISConnected
            # BeIfDisconnecting
            # FeIfConnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_dst_nfy_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_nfy_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_dst_fer':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfConnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_dst'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_dst_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDisconnected
            # FeIfConnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_dst_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_dst'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_bdd_dst_nfy':
            # FeDowning
            # BeDowning
            # SendBeDontCare
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst_nfy'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst':
            # FeDowning
            # BeDowning
            # SendBeDontCare
            # SendFeDontCare
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_dst_nfy_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeDontCare
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_dst_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeDontCare
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_nfy_dst_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeDontCare
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_dst_nfy_ber_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_dst_nfy'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_nfy_dst_nfy_ber':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeISDisconnected
            # BeIfDisconnecting
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_dst_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_nfy_dst_ber_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_bdd_fic_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_dst_ber'
                self.__send_be_destroy()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_dst'
                self.__send_fe_is_destroyed()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_bdd_dst_nfy':
            # FeDown
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_bdd_dst_nfy'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_fdd_bdd_nfy_dst':
            # FeDown
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer_ber'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd_dst_nfy_fer':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst_nfy'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_bdd_nfy_dst_fer':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_bdd_dst_fer_fer_ber'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_bdd'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst'
                self.__close_event_channel()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_nfy':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_ber'
                self.__send_be_destroy()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu'
                self.__complete_be_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_dst_ber'
                self.__send_be_destroy()
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_nfy_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fer'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_nfy_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_ber_nfy_dst_fic_nfy_fer':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanOpen
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy'
                self.__close_event_channel()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer_fer_bdu_ber_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst_fer'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_dst_fer_fer_bdu_ber_nfy_bdd':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst_fer'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_dst_fer_fer'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_dst':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd'
                self.__complete_fe_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst_nfy':
            # FeDown
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_dst_nfy'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst':
            # FeDown
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdu':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst'
                self.__notify_devices('down')
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_dst_nfy'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd_fdu':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_bdd'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_bdd'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd_fdu_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_dst'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_bdd_dst_nfy_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_dst_nfy'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_dst_nfy_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_bdd_nfy_dst_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_bdd_dst_fer_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy_dst_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_bdd_nfy_ber':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_dst_nfy'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy_ber'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_ber':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_bdd_nfy_dst'
                self.__complete_fe_down()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy_ber'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_nfy_ber_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfConnectable
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_dst'
                self.__complete_fe_down()
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_ber'
                self.__notify_devices('down')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy_ber':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy_ber_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_dst_ber_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_fdd_nfy_dst'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_ber'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy_ber_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst_nfy'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy_dst'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd'
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu'
                self.__send_be_create()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst_nfy'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeISConnected
            # BeIfDisconnecting
            # FeIfConnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_fer':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfConnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_nfy_dst'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDisconnected
            # FeIfConnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy_ber'
                self.__send_be_destroy()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_nfy_dst'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_bdd_dst_nfy':
            # FeUp
            # BeDowning
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_bdd_dst_nfy'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_dst_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_dst_nfy'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd_dst_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_dst_nfy_fer':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfConnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_dst_nfy_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_dst'
                self.__notify_devices('down')
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_dst_nfy'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_ber_dst_nfy_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDisconnected
            # FeIfConnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_dst_nfy'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeDontCare
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeDontCare
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeISDisconnected
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_nfy'
                self.__close_event_channel()
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fer':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_fer'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_bdd'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst'
                self.__close_event_channel()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fer_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfConnected
            # FeIfConnectable
            # DevicesDowning
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd_fic'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fer_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_nfy_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDisconnected
            # FeIfDisconnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_fer_ber'
                self.__complete_fe_down()
                self.__send_be_destroy()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_fer'
                self.__complete_be_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer'
                self.__open_event_channel()
                self.__send_be_connect()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_dst_fer_ber'
                self.__send_be_destroy()
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_dst_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_dst_nfy'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_dst_nfy'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd_dst_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_dst':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISDisconnected
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_bdd_nfy_dst'
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_ber_bdd_nfy_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_bdd_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_bdd_nfy_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_bdd_fdu_nfy_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_bdd_nfy'
                self.__complete_fe_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy_dst'
                self.__notify_devices('down')
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_bdd_nfy'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd_fdu_fer':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeIdle
            # BeIfDestroyed
            # FeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd'
                self.__complete_fe_down()
            elif stimulus == 'fic':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_bdd_nfy_dst'
                self.__send_fe_is_destroyed()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_bdd_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_bdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer'
                self.__complete_fe_down()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fdd_dst_fer_ber'
                self.__complete_fe_down()
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy_ber':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer_ber'
                self.__complete_fe_down()
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy_ber_bdu'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy_ber_ber':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeIdle
            # BeIfDestroying
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst_fer_ber'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy_ber_bdu'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst_nfy':
            # FeUp
            # BeDown
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd_dst'
            elif stimulus == 'bdu':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst_nfy_bdu'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_dst'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeISConnected
            # BeIfCreating
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber'
                self.__notify_devices('up')
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd_nfy':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_bdd_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst_nfy'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy_fer':
            # FeUp
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfConnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_dst_ber'
                self.__send_be_destroy()
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy_ber':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeISConnected
            # BeIfDestroying
            # FeIfConnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_dst_ber'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy_ber_ber'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeDontCare
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_fer':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_fer'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_fer_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeDontCare
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_fer_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_ber_nfy_fdd'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer':
            # FeUp
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_fer'
                self.__close_event_channel()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_bdd_fic_fer'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_dst_nfy'
                self.__close_event_channel()
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISDisconnected
            # BeIfConnected
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_fic'
                self.__close_event_channel()
                self.__complete_be_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_nfy_dst_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer_ber'
                self.__send_fe_is_connected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fer_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfConnected
            # FeIfConnected
            # DevicesDowning
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd'
                self.__close_event_channel()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer_ber'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fer_ber_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy_ber_bdu':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfDestroyed
            # FeIfConnectable
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_dst_fer_ber_bdu'
                self.__complete_fe_down()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_dst_nfy_bdd_nfy_ber'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst_nfy_bdu':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd_bdu_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst_nfy'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_dst_bdu'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_bdd':
            # FeUp
            # BeDowning
            # SendBeDontCare
            # SendFeISConnected
            # BeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd_nfy'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_bdd'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy'
                self.__complete_be_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_dst':
            # FeUp
            # BeUp
            # SendBeCreate
            # SendFeISConnected
            # BeIfCreating
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_dst'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_dst_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_dst'
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_add'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy_ber_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDestroyed
            # FeIfConnected
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_fdd_bdu_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst_nfy'
                self.__complete_be_down()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_dst_bdu'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_fer':
            # FeDowning
            # BeUp
            # SendBeConnect
            # SendFeIdle
            # BeIfConnecting
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy'
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_fer_ber'
                self.__send_be_disconnect()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeDontCare
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_ber_fdd_nfy_dst_nfy_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_fer_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fdd_ber'
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_fer_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_bdd'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_fer_ber'
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd'
                self.__complete_fe_down()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer_ber':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfConnected
            # FeIfConnected
            # DevicesDown
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy'
                self.__close_event_channel()
                self.__complete_be_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                pass
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer_ber_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fer_ber_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfConnected
            # FeIfConnected
            # DevicesDowning
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_fer_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_ber_nfy_bdd'
                self.__close_event_channel()
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_fer'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer_ber_fer'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_dst_ber':
            # FeUp
            # BeUp
            # SendBeDestroy
            # SendFeISConnected
            # BeIfDestroying
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fdd_dst_ber'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_bdd_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_dst_ber'
                self.__send_fe_is_disconnected()
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst_nfy_bdu'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_dst':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDisconnected
            # DevicesUping
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_dst_nfy'
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_dst'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_add':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDisconnected
            # DevicesUping
            # AddDirty
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_ber_bdd'
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_dst'
            elif stimulus == 'add':
                pass
            elif stimulus == 'nfy':
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_add'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_nfy':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_nfy_fdd'
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd'
                self.__notify_devices('down')
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'add':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber'
                self.__notify_devices('up')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_nfy'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_fer_ber':
            # FeDowning
            # BeUp
            # SendBeDisconnect
            # SendFeIdle
            # BeIfDisconnecting
            # FeIfDestroyed
            # DevicesDown
            # AddClean
            # EvtChanClosed
            # DestroyIdle
            if stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fdd_bdd_fer'
                self.__complete_fe_down()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_fdd_nfy_dst_nfy_ber'
            elif stimulus == 'add':
                pass
            elif stimulus == 'ber':
                self.__state = 'i_fdu_bdu_fdd_fer_ber'
                self.__complete_fe_down()
                self.__send_be_destroy()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_nfy_fer_ber_fer':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeIdle
            # BeIfConnected
            # FeIfConnected
            # DevicesDown
            # AddClean
            # EvtChanOpen
            # DestroyIdle
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_fer_ber_nfy_fdd_fdu_fic_fdd_nfy_fer_ber'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'bdd':
                self.__state = 'i_fdu'
                self.__close_event_channel()
                self.__complete_be_down()
                self.__send_fe_is_disconnected()
            elif stimulus == 'dst':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_dst_nfy_fer'
                self.__close_event_channel()
                self.__send_be_disconnect()
            elif stimulus == 'add':
                pass
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_dst_nfy':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDisconnected
            # DevicesUp
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
                self.__notify_devices('down')
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst'
                self.__notify_devices('down')
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_dst_nfy'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_ber_nfy_dst':
            # FeUp
            # BeUp
            # SendBeIdle
            # SendFeISConnected
            # BeIfDisconnected
            # DevicesDowning
            # AddClean
            # EvtChanClosed
            # DestroyOutstanding
            if stimulus == 'fdd':
                self.__state = 'i_fdu_bdu_ber_fdd_nfy_dst'
            elif stimulus == 'bdd':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_dst'
            elif stimulus == 'nfy':
                self.__state = 'i_fdu_bdu_fic_fer_ber_nfy_ber_bdd_nfy_bdu_dst_ber'
                self.__send_be_destroy()
            elif stimulus == 'fer':
                self.__state = 'i_fdu_bdu_ber_nfy_dst'
                self.__send_fe_is_disconnected()
            else:
                self.__stimulus_error(stimulus)
        else:
            self.__stimulus_error(stimulus)

    # FIXME: either need to serialise notifications or code BE to handle
    # concurrent claim port messages

    def __notify_devices(self, function):
        log.debug("UsbifBackendInterface.__notify_devices")
        self.__notify_outstanding = 1
        for device in self.__devices.values():
            self.__notify_outstanding = self.__notify_outstanding + 1
            deferred = defer.Deferred()
            deferred.addCallback(self.__notify_devices_1)
            if function == 'up':
                device.be_up(deferred)
            elif function == 'down':
                device.be_down(deferred)
            else:
                raise NotImplementedError()
        self.__notify_devices_1()

    def __notify_devices_1(self, ignored=0):
        self.__notify_outstanding = self.__notify_outstanding - 1
        if self.__notify_outstanding == 0:
            self.__stimulus('nfy')

    def __complete_fe_down(self):
        log.debug("UsbifBackendInterface.__complete_fe_down")
        deferred = self.__fe_deferred
        self.__fe_deferred = None
        deferred.callback(None)

    def __complete_be_down(self):
        log.debug("UsbifBackendInterface.__complete_be_down")
        deferred = self.__be_deferred
        self.__be_deferred = None
        deferred.callback(None)

    def __complete_destroy(self):
        log.debug("UsbifBackendInterface.__complete_destroy")
        deferred = self.__destroy_deferred
        self.__destroy_deferred = None
        deferred.callback(None)
        self.close()

    def __open_event_channel(self):
        log.debug("UsbifBackendInterface.__open_event_channel")
        self.__evtchn = channel.eventChannel(self.dom, self.controller.dom)

    def __close_event_channel(self):
        log.debug("UsbifBackendInterface.__close_event_channel")
        channel.eventChannelClose(self.__evtchn)

    def __send_be_create(self):
        log.debug("UsbifBackendInterface.__send_be_create")
        msg = packMsg(
            'usbif_be_create_t',
            { 'domid'  : self.controller.dom,
              'handle' : self.handle
            })
        self.__send_be_message(msg)

    def __send_be_destroy(self):
        log.debug("UsbifBackendInterface.__send_be_destroy")
        msg = packMsg(
            'usbif_be_destroy_t',
            { 'domid'  : self.controller.dom,
              'handle' : self.handle
            })
        self.__send_be_message(msg)

    def __send_be_connect(self):
        log.debug("UsbifBackendInterface.__send_be_connect")
        msg = packMsg(
            'usbif_be_connect_t',
            {   'domid'  : self.controller.dom,
                'handle' : self.handle,
                'shmem_frame'   : self.__shmem_frame,
                'evtchn'        : self.__evtchn['port1'],
                'bandwidth'     : 500 # FIXME
            })
        self.__send_be_message(msg)

    def __send_be_disconnect(self):
        log.debug("UsbifBackendInterface.__send_be_disconnect")
        msg = packMsg(
            'usbif_be_disconnect_t',
            {   'domid'  : self.controller.dom,
                'handle' : self.handle
            })
        self.__send_be_message(msg)

    def __send_fe_is_disconnected(self):
        log.debug("UsbifBackendInterface.__send_fe_is_disconnected")
        msg = packMsg(
            'usbif_fe_interface_status_changed_t',
            {   'handle' : self.handle,
                'status' : USBIF_INTERFACE_STATUS_DISCONNECTED
            })
        self.__send_fe_message(msg)

    def __send_fe_is_connected(self):
        log.debug("UsbifBackendInterface.__send_fe_is_connected")
        msg = packMsg(
            'usbif_fe_interface_status_changed_t',
            {   'handle'    : self.handle,
                'status'    : USBIF_INTERFACE_STATUS_CONNECTED,
                'evtchn'    : self.__evtchn['port2'],
                'bandwidth' : 500, # FIXME
                'num_ports' : 1 # FIXME
            })
        self.__send_fe_message(msg)

    def __send_fe_is_destroyed(self):
        log.debug("UsbifBackendInterface.__send_fe_is_destroyed")
        msg = packMsg(
            'usbif_fe_interface_status_changed_t',
            {   'handle' : self.handle,
                'status' : USBIF_INTERFACE_STATUS_DESTROYED
            })
        self.__send_fe_message(msg)

    def __send_fe_message(self, msg):
        log.debug("UsbifBackendInterface.__send_fe_message")
        d = defer.Deferred()
        d.addCallback(self.__fe_message_response)
        self.controller.writeRequest(msg, response=d)

    def __send_be_message(self, msg):
        log.debug("UsbifBackendInterface.__send_be_message")
        d = defer.Deferred()
        d.addCallback(self.__be_message_response)
        self.writeRequest(msg, response=d)

    def __fe_message_response(self, msg):
        log.debug("UsbifBackendInterface.__fe_message_response")
        self.__stimulus('fer')

    def __be_message_response(self, msg):
        log.debug("UsbifBackendInterface.__be_message_response")
        self.__stimulus('ber')

class UsbDev(controller.SplitDev):

    def __init__(self, idx, ctrl, config):
        log.debug("UsbDev")
        controller.SplitDev.__init__(self, idx, ctrl)
        self.__state = 'i'
        self.__idx = idx
        self.__path = sxp.child_value(config, 'path')
        if not self.__path:
            raise VmError('usb: missing path')
        self.__backend = sxp.child_value(config, 'backend', '0')
        try:
            xd = get_component('xen.xend.XendDomain')
            self.backendDomain = int(xd.domain_lookup(self.__backend).id)
        except:
            raise VmError('usb: invalid backend domain')
        try:
            self.__port = int(sxp.child_value(config, 'port'))
        except:
            raise VmError('usb: invalid frontend port')
        self.getBackendInterface().add_device(self,idx)

    def be_up(self, deferred):
        log.debug("UsbDev.be_up")
        self.__be_deferred = deferred
        self.__stimulus('bdu')

    def be_down(self, deferred):
        log.debug("UsbDev.be_down")
        self.__be_deferred = deferred
        self.__stimulus('bdd')

    def destroy(self, deferred):
        log.debug("UsbDev.destroy")
        self.__destroy_deferred = deferred
        self.__stimulus('dst')

    def sxpr(self):
        log.debug("UsbDev.sxpr")
        val = \
        [
            'usb',
            [ 'idx',     self.idx       ],
            [ 'path',    self.__path    ],
            [ 'backend', self.__backend ],
            [ 'port',    self.__port    ]
        ]
        return val

    # bdu - backend domain up
    # bdd - backend domain down
    # dst - destroy device
    # __send_be_claim_port   : rsp
    # __send_be_release_port : rsp
    # __be_complete
    # __destroy_complete

    def __stimulus(self, stimulus):
        log.debug("UsbDev.__stimulus")
        log.debug("state:%s stimulus:%s", self.__state, stimulus)
        if self.__state == 'i':
            # BE down
            if stimulus == 'bdu':
                self.__state = 'i_bdu'
                self.__send_be_claim_port()
            elif stimulus == 'bdd':
                self.__be_complete()
            elif stimulus == 'dst':
                self.__destroy_complete()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_bdu':
            # BE up
            # sending be claim port
            # be_up outstanding
            if stimulus == 'rsp':
                self.__state = 'i_bdu_rsp'
                self.__be_complete()
            elif stimulus == 'dst':
                self.__state = 'i_bdu_dst'
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_bdu_rsp':
            # BE up
            # port claimed
            if stimulus == 'bdu':
                self.__be_complete()
            elif stimulus == 'bdd':
                self.__state = 'i'
                self.__be_complete()
            elif stimulus == 'dst':
                self.__state = 'i_bdu_rsp_dst'
                self.__send_be_release_port()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_bdu_dst':
            # BE up
            # sending be claim port
            # be_up outstanding
            # destroy outstanding
            if stimulus == 'rsp':
                self.__state = 'i_bdu_rsp_dst'
                self.__be_complete()
                self.__send_be_release_port()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_bdu_rsp_dst':
            # BE up
            # sending be release port
            # destroy outstanding
            if stimulus == 'bdu':
                self.__be_complete()
            elif stimulus == 'bdd':
                self.__state = 'i_bdu_rsp_dst_bdd'
            elif stimulus == 'rsp':
                self.__state = 'i_bdu_rsp_dst_rsp'
                self.__destroy_complete()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_bdu_rsp_dst_bdd':
            # BE down
            # sending be release port
            # be_down outstanding
            # destroy outstanding
            if stimulus == 'rsp':
                self.__state = 'i_bdu_rsp_dst_rsp'
                self.__be_complete()
                self.__destroy_complete()
            else:
                self.__stimulus_error(stimulus)
        elif self.__state == 'i_bdu_rsp_dst_rsp':
            # BE up/down
            # destroyed
            self.__stimulus_error(stimulus)
        else:
            self.__stimulus_error(stimulus)

    def __stimulus_error(self, stimulus):
        log.debug(
            "USB device in state %s received unexpected stimulus %s",
            self.__state,
            stimulus
        )

    def __send_be_claim_port(self):
        log.debug("UsbDev.__send_be_claim_port")
        msg = packMsg(
            'usbif_be_claim_port_t',
            {   'domid'  : self.controller.dom,
                'handle' : self.getBackendInterface().handle,
                'port'   : self.__port,
                'path'   : self.__path
            })
        self.__send_be_message(msg)

    def __send_be_release_port(self):
        log.debug("UsbDev.__send_be_release_port")
        msg = packMsg(
            'usbif_be_release_port_t',
            {   'domid'  : self.controller.dom,
                'handle' : self.getBackendInterface().handle,
                'port'   : self.__port
            })
        self.__send_be_message(msg)

    def __send_be_message(self, msg):
        log.debug("UsbDev.__send_be_message")
        d = defer.Deferred()
        d.addCallback(self.__message_response)
        self.getBackendInterface().writeRequest(msg, response=d)

    def __message_response(self, msg):
        log.debug("UsbDev.__message_response")
        self.__stimulus('rsp')

    def __be_complete(self):
        log.debug("UsbDev.__be_complete")
        deferred = self.__be_deferred
        self.__be_deferred = None
        deferred.callback(None)

    def __destroy_complete(self):
        log.debug("UsbDev.__destroy_complete")
        self.getBackendInterface().remove_device(self,self.__idx)
        self.controller.delDevice(self.__idx)
        deferred = self.__destroy_deferred
        self.__destroy_deferred = None
        deferred.callback(None)
